/*
 * Copyright (C) 2005, Laszlo Pere <pipas@linux.pte.hu>
 *
 * $Id: background.c,v 1.1 2006/02/24 17:27:03 pipas Exp pipas $
 *
 * This is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Library General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Log: background.c,v $
 * Revision 1.1  2006/02/24 17:27:03  pipas
 * Initial revision
 *
 */
#include "support.h"
#include "interface.h"
#include "cuiterm.h"
#include "config.h"
#include "background.h"
#include "terminal.h"
#include "widgets.h"
#include "cuiwindow.h"

static void
background_swap_widgets(Background *bg)
{
	GtkTextView      *text_view;

	text_view = bg->text_view; 
	bg->text_view = bg->secondary_text_view;
	bg->secondary_text_view = text_view;
	/*
	 * FIXME: Somehow this is NULL at start. 
	 */
	bg->secondary_window = gtk_text_view_get_window(
			bg->secondary_text_view, 
			GTK_TEXT_WINDOW_TEXT);
	
	bg->widget = GTK_WIDGET(bg->text_view);
	bg->window = gtk_text_view_get_window(bg->text_view, 
			GTK_TEXT_WINDOW_TEXT);
	
}

/**
 * cui_background_append_widget:
 * @bg: the #Background which will contain the widget
 * @widget: a #CuiWidget
 * @returns: 0 if everything is fine
 *
 * Adds the widget to the background so it can be appear on the screen and the
 * user can use it to control the application.
 */
gint
cui_background_append_widget(Background *bg, gpointer widget)
{
	g_return_val_if_fail(bg != NULL, -1);
	g_return_val_if_fail(widget != NULL, -1);
	/*
	 * Let's check if the widget is already in the list. 
	 */
	if (g_list_find(bg->bg_widgets, widget)){
		g_warning("%s(): The widget is already in the list: %p", 
				__func__, widget);
		return -1;
	}

	bg->bg_widgets = g_list_append(bg->bg_widgets, widget);
	return 0;
}

/**
 * cui_background_erase:
 *
 *
 */
gint
cui_background_erase(Background *bg)
{
	GList *l;
	gint n = 0;
	
	g_return_val_if_fail(bg != NULL, -1);
	
	l = bg->bg_widgets;
	while (l != NULL){
		g_object_unref(G_OBJECT(l->data));
		l = l->next;
		++n;
	}
	g_list_free(bg->bg_widgets);
	bg->bg_widgets = NULL;
	return n;
}

/**
 * background_new:
 * @widget: a #GtkTextView which will hold the new background
 * @returns: a new background
 *
 * Creates a new background for the given GtkWidget.
 */
Background *
background_new(GtkTextView *widget)
{
	GtkWidget       *app1;
	Background      *bg = g_malloc(sizeof(Background));
	
	g_assert(GTK_IS_TEXT_VIEW(widget));
	/* 
	 * Stored things from other parts of the galaxy.
	 */
	bg->bg_widgets = NULL;
	bg->dnd_target_widget = NULL;
	
	bg->text_view = widget;
	bg->widget = GTK_WIDGET(widget);
	bg->window = gtk_text_view_get_window(widget, GTK_TEXT_WINDOW_TEXT);
	
	bg->secondary_text_view = NULL;
	bg->secondary_window    = NULL;
	bg->text_buffer = gtk_text_view_get_buffer(widget);
	/* 
	 * Things needed to draw 
	 */
	bg->colormap = gdk_drawable_get_colormap(GDK_DRAWABLE(bg->window));
	bg->gc = gdk_gc_new(GDK_DRAWABLE(bg->window));
	/*
	 * Things to handle mouse clicks and double mouse clicks.
	 */
	app1 = lookup_widget(GTK_WIDGET(widget), "app1");
	g_object_get(G_OBJECT(gtk_widget_get_settings(GTK_WIDGET(widget))), 
			"gtk-double-click-time", &bg->double_click_time, 
			NULL);
	bg->last_button = 100;
	bg->last_click_time = 0;
	bg->click_count = 0;
	bg->still_holding = FALSE;	
	/*
	 * Mouse cursors.
	 */
	bg->mouse_cursor_arrow = gdk_cursor_new(GDK_TOP_LEFT_ARROW);
	bg->mouse_cursor_hand = gdk_cursor_new(GDK_HAND2);
	bg->mouse_cursor_text = gdk_cursor_new(GDK_XTERM);
	gdk_window_set_cursor(bg->window, bg->mouse_cursor_arrow);
	/*
	 * View flags.
	 */
	bg->view_active_areas = FALSE;
	bg->view_sizebars = TRUE;

	return bg;
}

/**
 * background_set_secondary_text_view:
 * @bg: a #Background
 * @text_view: a #GtkTextView also shows the given background
 *
 * Sets the secondary text view which uses this background.
 */
void 
background_set_secondary_text_view(Background *bg, 
		GtkTextView *text_view)
{
	gtk_widget_realize(GTK_WIDGET(text_view));
	bg->secondary_text_view = text_view;
	//gtk_widget_set_colormap(GTK_WIDGET(text_view), bg->colormap);
	bg->secondary_window = gtk_text_view_get_window(text_view, 
			GTK_TEXT_WINDOW_TEXT);
	if (bg->secondary_window != NULL)
		gdk_window_set_cursor(bg->secondary_window, 
				bg->mouse_cursor_arrow);
}

/**
 * background_set_method:
 * @bg: a background to set
 * @method: the usage method of the background
 *
 * Sets the usage method of the background.
 */
void
background_set_method(Background *bg, BackgroundMethod method)
{
	g_return_if_fail(bg != NULL);
	bg->method = method;
	switch (method){
		case MethodGraphics:
			gdk_window_set_cursor(bg->window, 
					bg->mouse_cursor_arrow);
			if (bg->secondary_window != NULL)
				gdk_window_set_cursor(bg->secondary_window, 
						bg->mouse_cursor_arrow);
			break;
			
		case MethodSelection:
			gdk_window_set_cursor(bg->window, 
					bg->mouse_cursor_hand);
			if (bg->secondary_window != NULL)
				gdk_window_set_cursor(bg->secondary_window,
						bg->mouse_cursor_hand);
			break;
			
		case MethodEditor:
			gdk_window_set_cursor(bg->window, 
					bg->mouse_cursor_text);
			if (bg->secondary_window != NULL)
				gdk_window_set_cursor(bg->secondary_window,
						bg->mouse_cursor_text);
			break;

		default:
			g_warning("%s(): Ivalid method: %d", __func__, 
					method);
	}
}

/**
 * background_toggle_view_flag:
 * @bg: a #background
 * @option: the view flag to be changed
 *
 * Switches a view flag to its opposite state.
 */

void
background_toggle_view_flag(Background *bg, 
		BackgroundViewOption option)
{
	switch (option){
		case ActiveAreas:
			bg->view_active_areas = !bg->view_active_areas;
			break;
		case Sizebars:
			bg->view_sizebars = !bg->view_sizebars;
			break;
		default:
			g_warning("%s(): Invalid view flag option: %d", __func__,
					option);
	}
}

/**
 * background_store_widget:
 * @bg: a #background
 * @widget: a #CuiWidget to be stored
 *
 * Stores a #CuiWidget related to an action to be used later.
 */

void
background_store_widget(Background *bg, gpointer widget)
{
	g_return_if_fail(bg != NULL);
	bg->stored_widget = widget;
}

/**
 * background_store_command:
 * @bg: a #background
 * @command: a command to be stored
 *
 * Stores a command and clears the stored widget.
 */

void
background_store_command(Background *bg, StoredCommand command)
{
	g_return_if_fail(bg != NULL);
	bg->stored_command = command;
	bg->stored_widget = NULL;
}

/**
 * background_get_last_dnd_target:
 * @background: a #background
 * @returns: the last #CuiWidget used as a DnD target.
 *
 * Returns the #CuiWidget used as a Drag&Drop target. The return value can be
 * NULL if there were no widget under the mouse pointer.
 */
gpointer *
background_get_last_dnd_target(Background *background)
{
	g_return_val_if_fail(background != NULL, NULL);
	return background->dnd_target_widget;
}

/**
 * background_get_last_dnd_source:
 * @background: a #background
 * @returns: the last #CuiWidget used as a DnD source.
 *
 * Returns the #CuiWidget used as a Drag&Drop source. The return value can be
 * NULL if there were no widget under the mouse pointer.
 */
gpointer *
background_get_last_dnd_source(Background *background)
{
	g_return_val_if_fail(background != NULL, NULL);
	return background->dnd_source_widget;
}

/**
 * background_get_stored_widget:
 * @background: a #background
 * @returns: the last #CuiWidget that has been stored.
 *
 * Returns the #CuiWidget that is currently stored. The return value can be
 * NULL if there was no widget currently stored.
 */

gpointer *
background_get_stored_widget(Background *background)
{
	g_return_val_if_fail(background != NULL, NULL);
	return background->stored_widget;
}

/**
 * background_get_view_flag:
 * @bg: a #background
 * @option: a view flag
 * @returns: the current state of the given view flag.
 */

gboolean
background_get_view_flag(Background *bg, 
		BackgroundViewOption option)
{
	switch (option){
		case ActiveAreas:
			return bg->view_active_areas;
			break;
		case Sizebars:
			return bg->view_sizebars;
			break;
		default:
			g_warning("%s(): Invalid view flag option: %d", __func__,
					option);
	}
}

/**
 * background_check_stored_command:
 * @bg: a #background
 * @command: a command that may have been stored
 * @returns: TRUE if the actually stored value is the same as the given.
 */

gboolean
background_check_stored_command(Background *bg, StoredCommand command)
{
	g_return_val_if_fail(bg != NULL, FALSE);
	if (bg->stored_command == command) return TRUE;
	return FALSE;
}

/*****************************************************************************
 * These functions and variables handling the mouse events such as button    *
 * clicks, double button clicks and long button clicks.                      *
 *                                                                           *
 ****************************************************************************/
static gboolean 
longclick_timeout_function(gpointer data)
{	
	GtkWidget  *widget = (GtkWidget *)data;
	Terminal   *t = g_object_get_data(G_OBJECT(widget), "terminal");
	Background *bg;
	
	g_assert(t != NULL);
	bg = t->background;
	g_assert(bg != NULL);
	
	/*
	 * If the user released the button in the meantime, we simple return
	 * from this function.
	 */
	if (!bg->still_holding)
		return FALSE;

	bg->still_holding = FALSE;
#ifdef DEBUG_BG	
	g_message("%s(): Long button click at: %3.0f, %3.0f", 
			__func__, bg->last_mouse_x, bg->last_mouse_y);
#endif
	
	background_do_entry(t, bg, widget, bg->last_mouse_x, bg->last_mouse_y);
	
	return FALSE;
}

/**
 * background_single_click_event:
 *
 * This is where we handle single clicks with all of the mouse buttons.
 */
static gboolean
background_single_click_event(
		Background *bg,
		GtkWidget *widget, 
		GdkEventButton *event, 
		gpointer user_data)
{
	GList        *list_item;
	CuiWidget    *cuiwidget;
	CuiWidget    *menu = NULL;
	CuiWidget    *workspace_menu = NULL;
	gint          menusize = -1;
	gchar        *widget_name;
	gint          x, y;

	
	/*
	 * This is where we handle the CuiWidget objects.
	 */
	if (event->button != 3)
		return TRUE;
	
	gtk_text_view_window_to_buffer_coords(GTK_TEXT_VIEW(widget), 
			GTK_TEXT_WINDOW_TEXT,
			(gint)event->x, (gint)event->y, &x, &y);
	
	list_item = bg->bg_widgets;
	while (list_item != NULL){
		cuiwidget = list_item->data;
		/*
		 * If the widget is a popup menu and its name is this, then we
		 * store a pointer and popup this menu if nothing else can be
		 * found at the specifiad location.
		 */
		widget_name = cui_widget_get_name(cuiwidget);
		if (CUI_IS_POPUP_MENU(cuiwidget) && 
				widget_name != NULL &&
				strcmp(widget_name, "WorkspaceMenu") == 0){
			workspace_menu = cuiwidget;
			goto next_item;
		}
		/*
		 * If it is not a menu or the mouse pointer is not over the 
		 * area we can go to the next item.
		 */
		if (!CUI_IS_POPUP_MENU(cuiwidget) || 
				!cui_widget_is_mouse_over(
					CUI_WIDGET(cuiwidget), x, y))
			goto next_item;
		/*
		 * We choose the menu with the smallest area.
		 */
		if (menu == NULL || cui_widget_get_size(cuiwidget) < menusize){
			menu = cuiwidget;
			menusize = cui_widget_get_size(cuiwidget);
		}
next_item:
		list_item = list_item->next;
	}
	/*
	 * We show the pop-up menu. FIXME: this should be much more simple if we
	 * use the GObjects better.
	 */
	if (CUI_IS_FILE_MENU(menu)){
		cui_file_menu_popup(CUI_FILE_MENU(menu), widget, event);
		return TRUE;
	}
	if (CUI_IS_MODE_MENU(menu)){
		cui_mode_menu_popup(CUI_MODE_MENU(menu), widget, event);
		return TRUE;
	}
	if (CUI_IS_USER_MENU(menu)){
		cui_user_menu_popup(CUI_USER_MENU(menu), widget, event);
		return TRUE;
	}
	if (CUI_IS_POPUP_MENU(menu)){
		cui_popup_menu_popup(CUI_POPUP_MENU(menu), widget, event);
		return TRUE;
	}
	/*
	 * If nothing could be found we start up the workspace menu.
	 */
	if (workspace_menu != NULL)
		cui_popup_menu_popup(CUI_POPUP_MENU(workspace_menu), 
				widget, event);
	return TRUE;
}

/**
 * background_button_
 */
static gboolean
background_double_click_event(
		Background     *bg,
		GtkWidget      *widget, 
		GdkEventButton *event, 
		gpointer        user_data)
{
	Terminal *terminal;
	
	terminal = g_object_get_data(G_OBJECT(widget), "terminal");
	g_assert(terminal != NULL);
	g_assert(bg != NULL);

	if (event->button == 1){
		bg->still_holding = FALSE;
		background_do_action(terminal, bg, widget, event);
	}
	return TRUE;
}


/**
 * background_button_press_event:
 * @bg: a #Background
 * @widget: the #GtkWidget the #Background pertain to
 * @event: the mouse event we try to handle
 * @user_data: the pointer given for the signal connection (may be NULL)
 * @returns: TRUE if the event handled
 *
 * Called when the any of the mouse buttons are pressed on a #Background. This
 * function will decide what to do and call the appropriate function to handle
 * the event.
 */
gboolean
background_button_press_event(
		Background *bg,
		GtkWidget *widget, 
		GdkEventButton *event, 
		gpointer user_data)
{
	guint32   current_time;
	Terminal  *t = g_object_get_data(G_OBJECT(widget), "terminal"); 
	/*
	 * If this is the same button and not passed much time we increment 
	 * the click count to detect double and triple clicks. Otherwise this
	 * must be a first click.
	 */
	current_time = gtk_get_current_event_time();
	if (bg->last_button == event->button && 
			current_time - bg->last_click_time < 
				bg->double_click_time) 
		bg->click_count++;
	else 
		bg->click_count = 0;
	/*
	 * If this is the button one we store the coordinates and we fire up a
	 * timeout to detect longclicks.
	 */
	if (event->button == 1){
		bg->last_mouse_x = event->x;
		bg->last_mouse_y = event->y;
		bg->still_holding = TRUE;
		g_timeout_add(400, longclick_timeout_function, 
				(gpointer) widget);
	}
	/*
	 * Storing important info about this event.
	 */
	bg->last_button = event->button;
	bg->last_click_time = current_time;
	/*
	 * We handle single, double a triple mouse clicks here.
	 */
	switch(bg->click_count){
		case 0:
			background_single_click_event(bg, widget, 
					event, user_data);
			break;
		case 1:
			background_double_click_event(bg, widget, 
					event, user_data);
			break;
		case 2:
			/* Nothing happens */
			break;
		default:
			bg->click_count = 0;
	}
	return TRUE;
}

gboolean
background_button_release_event(
		Background      *bg,
		GtkWidget       *widget, 
		GdkEventButton  *event, 
		gpointer         user_data)
{
	bg->still_holding = FALSE;
	return TRUE;
}

/**
 * background_dnd_motion_notify:
 * @background: a #Background 
 * @drag_context: the context 
 * @x: the mouse x coordinate
 * @y: the mouse y coordinate
 * @returns: TRUE if a DnD target has been found
 *
 * Called when the mouse is moving during the DnD operation. This function will
 * try to find a matching DnD target, selecting a DnD target type, and highlight
 * the DnD target if a match found.
 */
gboolean 
background_dnd_motion_notify(
		GtkWidget      *widget,
		Background     *background, 
		GdkDragContext *drag_context, 
		gint            x, 
		gint            y)
{
	gboolean 	 retval;
	gboolean         secondary = FALSE;
	static CuiWidget *old_widget = NULL;
	CuiWidget        *cuiwidget;
	GtkTargetList    *target_list;
	GList            *list_item;
	gint              wx, wy;
	
	g_return_val_if_fail(background != NULL && drag_context != NULL, FALSE);
	
	/*
	 * If this is the secondary widget we swap the widgets. FIXME: a better
	 * solution is needed.
	 */
	if (background->widget != widget){
		secondary = TRUE;
		background_swap_widgets(background);
	}

	/*
	 * If at least one of the coordinates is negative we has been called to
	 * remove the highlight.
	 */
	if ((x < -1 || y < -1) && old_widget != NULL){
		cui_rectangle_set_highlighted(CUI_RECTANGLE(old_widget), FALSE);
		old_widget = NULL;
	}

	gtk_text_view_window_to_buffer_coords(
			GTK_TEXT_VIEW(background->widget), 
			GTK_TEXT_WINDOW_TEXT, x, y, &wx, &wy);

	list_item = background->bg_widgets;
	while (list_item != NULL){
		cuiwidget = list_item->data;
		/*
		 * FIXME: Currently only the CuiRectangle objects can be DnD
		 * target. It is possible it was not wise decision.
		 * 
		 */
		if (!CUI_IS_RECTANGLE(cuiwidget))
			goto next_item;
		if (!cui_widget_is_mouse_over(cuiwidget, wx, wy))
			goto next_item;
		if (cuiwidget == old_widget){
			retval = TRUE;
			goto end;
		}
		if (!cui_rectangle_is_dnd_target(CUI_RECTANGLE(cuiwidget)))
			goto next_item;
		
		target_list = cui_rectangle_get_target_list(CUI_RECTANGLE(cuiwidget));
		background->match = gtk_drag_dest_find_target(background->widget, drag_context, target_list);
		if (background->match == GDK_NONE)
			goto next_item;
		
		/*
		 * Now we found a rectangle under the cursor which can be a DnD
		 * target. Now we can check if this rectangle can accept the
		 * dragged type. 
		 */
		if (old_widget != NULL){
			cui_rectangle_set_highlighted(CUI_RECTANGLE(old_widget), FALSE);
			old_widget = NULL;
		}
		
		cui_rectangle_set_highlighted(CUI_RECTANGLE(cuiwidget), TRUE);
		background->dnd_target_widget = cuiwidget;
		old_widget = cuiwidget;
		retval = TRUE;
		goto end;
next_item:
		list_item = list_item->next;
	}
	/*
	 * We found nothing so we remove the highlight status if there is a
	 * highlighted area.
	 */
	if (old_widget != NULL){
		cui_rectangle_set_highlighted(CUI_RECTANGLE(old_widget), FALSE);
		old_widget = NULL;
	}
	/*
	 * We could not find DnD target at this point to handle this drag.
	 */
	retval = FALSE;
end:
	if (secondary)
		background_swap_widgets(background);
	return retval;
}


gboolean
background_motion_notify_event(
		Background     *bg,
		GtkWidget      *widget, 
		GdkEventMotion *event, 
		gpointer       user_data)
{
	Terminal         *t;
	CuiWidget        *cuiwidget;
	static CuiWidget *old_widget = NULL;
	static CuiWidget *old_window = NULL;
	GList            *list_item;
	gint              x, y;
	
	g_assert(bg != NULL);
#ifdef DEBUG	
	g_message("%s(): event->window: %p", __func__, event->window);
	  g_message("%s(): event->x: %f", __func__, event->x);
	  g_message("%s(): event->y: %f", __func__, event->y);
#endif
		
	// The user moved the mouse why holding the button down.
	if (bg->still_holding){
		bg->still_holding = FALSE;
		dnd_drag_begin(event, bg->last_button, widget);
		return TRUE;
	}
	// The user moved the mouse, so not still any more.
	bg->still_holding = FALSE;
	/*
	 * We handle the leave event here.
	 */
	if (event->type == GDK_LEAVE_NOTIFY) {
		if (old_window != NULL) {
			cui_widget_mouse_leave(old_window);
			old_window = NULL;
		}
		if (old_widget != NULL) {
			cui_widget_mouse_leave(old_widget);
			old_widget = NULL;
		}
		return FALSE;
	}

	gtk_text_view_window_to_buffer_coords(GTK_TEXT_VIEW(widget), 
			GTK_TEXT_WINDOW_TEXT,
			(gint)event->x, (gint)event->y, &x, &y);
	/*
	 * First we try to find a CuiWindow the mouse is entered.
	 */
	list_item = bg->bg_widgets;
	while (list_item != NULL) {
		cuiwidget = list_item->data;
		// This is not a window.
		if (!CUI_IS_WINDOW(cuiwidget))
			goto next_item1;
		// This window is far from us.
		if (!cui_widget_is_mouse_over(cuiwidget, x, y))
			goto next_item1;
		// This is the same window we saw last time.
		if (cuiwidget == old_window)
			goto ready_with_windows;
		// Now we have a new window where the mouse entered.
		if (old_window != NULL)
			cui_widget_mouse_leave(old_window);
		cui_widget_mouse_enter(cuiwidget);
		old_window = cuiwidget;
		goto ready_with_windows;
next_item1:
		list_item = list_item->next;
	}
	// Now we could not find a window, maybe the mouse just left one
	if (old_window != NULL)
		cui_widget_mouse_leave(old_window);
	old_window = NULL;
		
		
ready_with_windows:
	/*
	 * Now the same with the general CuiWidgets. It is important, that we
	 * send the entered signal to the oldest widget, we find first in the 
	 * list.
	 */
	list_item = bg->bg_widgets;
	while (list_item != NULL){
		cuiwidget = list_item->data;
		// This is a window
		if (CUI_IS_WINDOW(cuiwidget))
			goto next_item2;
		// This widget is not here.
		if (!cui_widget_is_mouse_over(cuiwidget, x, y))
			goto next_item2;
		// This is the same widget we saw before.
		if (cuiwidget == old_widget) 
			goto ready_with_widgets;
		// Now, we found a new widget under the mouse.
		if (old_widget != NULL)
			cui_widget_mouse_leave(old_widget);
		cui_widget_mouse_enter(cuiwidget);
		old_widget = cuiwidget;
		goto ready_with_widgets;
next_item2:
		list_item = list_item->next;
	}
	// Now there is no widget under the mouse, maybe we left one.
	if (old_widget != NULL)
		cui_widget_mouse_leave(old_widget);
	old_widget = NULL;

ready_with_widgets:
	return FALSE;
}


gboolean
background_leave_notify_event(
		Background       *bg,
		GtkWidget        *widget, 
		GdkEventCrossing *event, 
		gpointer          user_data)
{
	/*
	 * We handle the motion and leave events in the same function.
	 */
	background_motion_notify_event(bg, widget, (GdkEventMotion *)event, user_data);	
	return FALSE;
}
/*
 * I think the drawing of the CUI widgets is a rather efficient process.
 */
#undef PROFILE
#ifdef PROFILE
#   include <time.h>
#endif

gboolean
background_expose_event (
		GtkWidget          *widget, 
		GdkEventExpose     *event, 
		gpointer            user_data)
{
	gboolean secondary = FALSE;
	Terminal *terminal;
	GdkRectangle rec;
	CuiWidget *cuiwidget;
	Background *bg;
	GList *list_item;
	gint n = 0;
#ifdef PROFILE
	clock_t start_time, end_time;
#endif

	terminal = g_object_get_data(G_OBJECT(widget), "terminal");
	g_assert(terminal != NULL);
	bg = terminal->background;
	g_assert(bg != NULL);

	/*
	 * If this is the secondary widget we sap the widgets. FIXME: a better
	 * solution is needed.
	 */
	if (bg->widget != widget){
		secondary = TRUE;
		background_swap_widgets(bg);
	}
	
	gdk_region_get_clipbox(event->region, &rec);
	gdk_gc_set_clip_region(bg->gc, event->region);
	/*
	 * This is where we redraw the screen widget objects. 
	 */
	n = 0;
	list_item = bg->bg_widgets;
#ifdef PROFILE
	start_time = clock();
#endif
	while (list_item != NULL){
		cuiwidget = list_item->data;
		
		if (CUI_IS_ACTION_AREA(cuiwidget))
			cui_action_area_redraw(CUI_ACTION_AREA(cuiwidget), 1);
		
		if (CUI_IS_PIXMAP(cuiwidget))
			cui_pixmap_redraw(CUI_PIXMAP(cuiwidget), 1);
		
		if (CUI_IS_RECTANGLE(cuiwidget))
			cui_rectangle_redraw(CUI_RECTANGLE(cuiwidget), 1);

		list_item = list_item->next;
		++n;
	}
#ifdef PROFILE
	end_time = clock();
	g_message("%s(): Checked %d item in: %f\n", 
			__func__,
			n,
			((double)(end_time-start_time))/CLOCKS_PER_SEC);
#endif

	if (secondary)
		background_swap_widgets(bg);
	
  	return FALSE;
}


/**
 * background_do_entry:
 *
 *
 */
gboolean 
background_do_entry(Terminal *terminal,
		Background *bg,
		GtkWidget  *widget,
		gint        event_x,
		gint        event_y)
{
	gboolean         secondary = FALSE;
	GList           *list_item;
	CuiWidget       *cuiwidget;
	gint             x, y;

#ifdef DEBUG_BG	
	g_message("%s(): starting...", __func__);
#endif

	g_return_val_if_fail(terminal != NULL && bg != NULL, TRUE);
	/*
	 * If this is the secondary widget we swap the widgets. FIXME: a better
	 * solution is needed.
	 */
	if (bg->widget != widget){
		secondary = TRUE;
		background_swap_widgets(bg);
	}
	/*
	 * This is where we handle the CuiWidget objects.
	 */
	list_item = bg->bg_widgets;
	gtk_text_view_window_to_buffer_coords(GTK_TEXT_VIEW(widget), 
			GTK_TEXT_WINDOW_TEXT,
			event_x, event_y, &x, &y);

	while (list_item != NULL){
		cuiwidget = list_item->data;
		if (!cui_widget_is_mouse_over(CUI_WIDGET(cuiwidget), x, y))
			goto next_item;
		if (CUI_IS_ENTRY(cuiwidget)){
			cui_entry_realize(CUI_ENTRY(cuiwidget));
			break;
		}
		if (CUI_IS_CALENDAR(cuiwidget)){
			cui_calendar_realize(CUI_CALENDAR(cuiwidget));
			break;
		}
next_item:
		list_item = list_item->next;
	}
	
end:	
	if (secondary)
		background_swap_widgets(bg);
	return TRUE;
}

/**
 * background_do_action:
 *
 */
int 
background_do_action(Terminal *terminal,
		Background *bg,
		GtkWidget *widget, 
		GdkEventButton *event)
{
	gint x, y;
	CuiWidget *cuiwidget;
	GList *list_item;
	
	g_return_val_if_fail(terminal != NULL && bg != NULL, TRUE);
	g_return_val_if_fail(widget != NULL && event != NULL, TRUE);
				
	/*
	 * This is where we handle the CuiWidget objects.
	 */
	list_item = bg->bg_widgets;
	gtk_text_view_window_to_buffer_coords(GTK_TEXT_VIEW(widget), 
			GTK_TEXT_WINDOW_TEXT,
			(gint)event->x, (gint)event->y, &x, &y);

	while (list_item != NULL){
		cuiwidget = list_item->data;
		/*
		 * As of now CuiWindows will not get double click events.
		 */
		if (CUI_IS_WINDOW(cuiwidget))
			goto next_item;
		
		if (!cui_widget_is_mouse_over(CUI_WIDGET(cuiwidget), x, y))
			goto next_item;
		cui_widget_activate(cuiwidget);
		return TRUE;
next_item:
		list_item = list_item->next;
	}
	return TRUE;
}

