/*
 * Copyright (C) 2004 Laszlo Pere
 * $Id: filename.c,v 1.1 2006/02/24 17:18:34 pipas Exp pipas $
 *
 * This is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Library General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Log: filename.c,v $
 * Revision 1.1  2006/02/24 17:18:34  pipas
 * Initial revision
 *
 */
#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif
#ifdef ALLDEBUG
#  define FILENAME_DEBUG
#endif


#include <stdlib.h>
#include <gnome.h>
#include "filename.h"


/*****************************************************************************
 * Instantiable classed type: CuiFilename                                    *
 *                                                                           *
 *****************************************************************************/
#define PROPERTY_FILENAME 1
#define PROPERTY_DIRECTORY 2
#define PROPERTY_MODE_STRING 3
#define PROPERTY_OWNER 4
#define PROPERTY_GROUP_OWNER 5
#define PROPERTY_SIZE 6
#define PROPERTY_LAST_USED 7
#define PROPERTY_LAST_MODIFIED 8
#define PROPERTY_LAST_INODE_OP 9

static void
cui_filename_set_property(GObject *object, 
		guint           property_id, 
		const GValue    *value, 
		GParamSpec      *pspec);
static void
cui_filename_get_property(GObject *object, 
		guint            property_id, 
		GValue           *value, 
		GParamSpec       *pspec);

/*
 * This is a simple locally used type for the file information pool, not a
 * derived class of the CuiFilename.
 */
typedef struct CuiDir {
	gchar *dirname;
	GTree *tree;
} CuiDir;


static gint 
cui_filename_compare(CuiFilename *a, CuiFilename *b)
{
	g_assert(CUI_IS_FILENAME(a) && CUI_IS_FILENAME(b));
	return strcmp(a->filename, b->filename);
}


/*
 * Try to find the directory with the given path in the file pool. If the name
 * could not be found the function will create and store a new item.
 */
static CuiDir *
directory_find(CuiFilenameClass *klass, const gchar *dirname)
{
	CuiDir *retval;

	retval = g_tree_lookup(klass->dir_repository, dirname);
	if (retval == NULL){
		retval = g_new(CuiDir, 1);
		retval->dirname = g_strdup(dirname);
		retval->tree = g_tree_new((GCompareFunc)strcmp);
		g_tree_insert(klass->dir_repository,
				retval->dirname,
				retval);
	}
	return retval;
}

static CuiFilename *
cui_filename_from_pool(CuiFilenameClass *klass, 
		const char *dirname,
		const char *filename)
{
	CuiDir *dir;
	CuiFilename *file;
	g_return_val_if_fail(CUI_IS_FILENAME_CLASS(klass), NULL);
	g_assert(dirname != NULL && filename != NULL);
	
	dir = directory_find(klass, dirname);
	file = g_tree_lookup(dir->tree, filename);

	return file;
}

static void 
cui_filename_class_init(
		gpointer g_class, 
		gpointer g_class_data) 
{
	GObjectClass *gobject_class = G_OBJECT_CLASS(g_class);
	CuiFilenameClass *klass = CUI_FILENAME_CLASS(g_class);
	GParamSpec *pspec;
	
	/*
	 * The directory name is at the very begin of the CuiDir struct, so we
	 * can simply compare the directory names.
	 */
	klass->dir_repository = g_tree_new((GCompareFunc)strcmp);
	
	gobject_class->set_property = cui_filename_set_property;
  	gobject_class->get_property = cui_filename_get_property;

	/*
	 * Properties.
	 */
	pspec = g_param_spec_string(
			"filename", "file-name",
                        "The basename of the file.",
                        "",
                        G_PARAM_READWRITE);
	g_object_class_install_property(gobject_class, 
			PROPERTY_FILENAME, 
			pspec);
	
	pspec = g_param_spec_string(
			"directory", "directory-name",
                        "The directory in which we found the file.",
                        "",
                        G_PARAM_READWRITE);
	g_object_class_install_property(gobject_class, 
			PROPERTY_DIRECTORY, 
			pspec);
	
	pspec = g_param_spec_string( 
			"mode_string", "mode-string", 
			"The mode string as the '-l' option of the 'ls' "
			"used to write.", 
			"", 
			G_PARAM_READWRITE);
	g_object_class_install_property(gobject_class, 
			PROPERTY_MODE_STRING, 
			pspec);
	
	pspec = g_param_spec_string( 
			"owner", "file-owner", 
			"The user name of the file owner.",
			"", 
			G_PARAM_READWRITE);
	g_object_class_install_property(gobject_class, 
			PROPERTY_OWNER, 
			pspec);
	
	pspec = g_param_spec_string( 
			"group", "group-owner", 
			"The name of the file group owner.",
			"", 
			G_PARAM_READWRITE);
	g_object_class_install_property(gobject_class, 
			PROPERTY_GROUP_OWNER, 
			pspec);
	
	pspec = g_param_spec_int64(
			"size", "file-size",
			"File size in bytes.",
			G_MININT64, G_MAXINT64, -1, 
			G_PARAM_READWRITE);
	g_object_class_install_property(gobject_class, 
			PROPERTY_SIZE, 
			pspec);
	
	pspec = g_param_spec_pointer(
			"last-used", "last_used",
                        "The time the file last used.",
                        G_PARAM_READWRITE);
	g_object_class_install_property(gobject_class, 
			PROPERTY_LAST_USED, 
			pspec);

	pspec = g_param_spec_pointer(
			"last-modified", "last_modified",
                        "The time the file last modified.",
                        G_PARAM_READWRITE);
	g_object_class_install_property(gobject_class, 
			PROPERTY_LAST_MODIFIED, 
			pspec);

	pspec = g_param_spec_pointer(
			"inode-modified", "inode_last_modified",
                        "The time the file inode last used.",
                        G_PARAM_READWRITE);
	g_object_class_install_property(gobject_class, 
			PROPERTY_LAST_INODE_OP, 
			pspec);
}

static void
cui_filename_base_class_init(CuiFilenameClass *klass)
{
	
}

static void
cui_filename_instance_init(
		GTypeInstance   *instance, 
		gpointer         g_class)
{
	CuiFilename *self = (CuiFilename *)instance;
	CuiFilenameClass *klass = (CuiFilenameClass *)g_class;

	self->file_type = FILETYPE_UNKNOWN;
	
	self->user_read = MODE_UNKNOWN;
	self->user_write = MODE_UNKNOWN;
	self->user_execute = MODE_UNKNOWN;
	self->group_read = MODE_UNKNOWN;
	self->group_write = MODE_UNKNOWN;
	self->group_execute = MODE_UNKNOWN;
	self->others_read = MODE_UNKNOWN;
	self->others_write = MODE_UNKNOWN;
	self->others_execute = MODE_UNKNOWN;
	
	self->filename = NULL;
	self->directory = NULL;
	self->owner = NULL;
	self->group_owner = NULL;
	
	self->size = -1;                    /* This mean of course: unknown size */
	self->sizechange = FileSizeUnknown; 

	self->last_used = NULL;
	self->last_modified = NULL;
	self->last_inode_op = NULL;
}

static void
cui_filename_set_property(GObject *object, 
		guint           property_id, 
		const GValue    *value, 
		GParamSpec      *pspec)
{
	CuiFilename    *self = (CuiFilename *)object;
	gint64 	        newsize;
		
	switch(property_id){
		case PROPERTY_FILENAME:
			if (self->filename != NULL)
				g_warning("The filename property can not "
						"be modified.");
			else
				self->filename = g_value_dup_string(value);
			break;
		case PROPERTY_DIRECTORY:
			if (self->directory != NULL)
				g_warning("The directory property can not "
						"be modified.");
			else
				self->directory = g_value_dup_string(value);
			break;

		case PROPERTY_MODE_STRING:
			g_free(self->mode_string);
			self->mode_string = g_value_dup_string(value);
			/*
			 * If the mode string is shorter than 10 characters we
			 * can not disassemble the meaning of it.
			 */
			if (self->mode_string == NULL ||
					strlen(self->mode_string) < 10)
				return;
			/*
			 * All right, this is not pretty, but fast (I hope...)
			 */
			switch (self->mode_string[0]){
				case '-':
					self->file_type = FILETYPE_FILE;
					break;
				case 'd':
					self->file_type = FILETYPE_DIRECTORY;
					break;
				case 'b':
					self->file_type = FILETYPE_BLOCK;
					break;
				case 'c':
					self->file_type = FILETYPE_CHAR;
					break;
				case 'p':
					self->file_type = FILETYPE_FIFO;
					break;
				case 'l':
					self->file_type = FILETYPE_LINK;
					break;
				case 's':
					self->file_type = FILETYPE_SOCKET;
					break;
				default:
					self->file_type = FILETYPE_UNKNOWN;
			}
			/* 
			 * The rights of the user (I mean the owner). 
			 */
			switch (self->mode_string[1]){
				case 'r':
					self->user_read = MODE_YES;
					break;
				case '-':
					self->user_read = MODE_NO;
					break;
				default:
					self->user_read = MODE_NO;
			}
			switch (self->mode_string[2]){
				case 'w':
					self->user_write = MODE_YES;
					break;
				case '-':
					self->user_write = MODE_NO;
					break;
				default:
					self->user_write = MODE_UNKNOWN;
			}
			switch (self->mode_string[3]){
				case 'x':
					self->user_execute = MODE_YES;
					break;
				case '-':
					self->user_execute = MODE_NO;
					break;
				default:
					self->user_execute = MODE_UNKNOWN;
			}
			/* 
			 * The rights of the group owner. 
			 */
			switch (self->mode_string[4]){
				case 'r':
					self->group_read = MODE_YES;
					break;
				case '-':
					self->group_read = MODE_NO;
					break;
				default:
					self->group_read = MODE_NO;
			}
			switch (self->mode_string[5]){
				case 'w':
					self->group_write = MODE_YES;
					break;
				case '-':
					self->group_write = MODE_NO;
					break;
				default:
					self->group_write = MODE_UNKNOWN;
			}
			switch (self->mode_string[6]){
				case 'x':
					self->group_execute = MODE_YES;
					break;
				case '-':
					self->group_execute = MODE_NO;
					break;
				default:
					self->group_execute = MODE_UNKNOWN;
			}
			/* 
			 * The rights of the others 
			 */
			switch (self->mode_string[7]){
				case 'r':
					self->others_read = MODE_YES;
					break;
				case '-':
					self->others_read = MODE_NO;
					break;
				default:
					self->others_read = MODE_NO;
			}
			switch (self->mode_string[8]){
				case 'w':
					self->others_write = MODE_YES;
					break;
				case '-':
					self->others_write = MODE_NO;
					break;
				default:
					self->others_write = MODE_UNKNOWN;
			}
			switch (self->mode_string[9]){
				case 'x':
					self->others_execute = MODE_YES;
					break;
				case '-':
					self->others_execute = MODE_NO;
					break;
				default:
					self->others_execute = MODE_UNKNOWN;
			}
			break;
			
		case PROPERTY_OWNER:
			g_free(self->owner);
			self->owner = g_value_dup_string(value);
			break;

		case PROPERTY_GROUP_OWNER:
			g_free(self->group_owner);
			self->group_owner = g_value_dup_string(value);
			break;
		
		case PROPERTY_SIZE:
			newsize = g_value_get_int64(value);
			/* We want to know how the size is changed. */
			if (self->sizechange == FileSizeUnknown)
				self->sizechange = FileSizeSame;
			else if (newsize == self->size)
				self->sizechange = FileSizeSame;
			else if (newsize < self->size)
				self->sizechange = FileSizeDecreasing;
			else
				self->sizechange = FileSizeIncreasing;
			
			self->size = newsize;
			break;
			
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID(
					object, property_id, pspec);
	}
}

static void
cui_filename_get_property(GObject *object, 
		guint            property_id, 
		GValue           *value, 
		GParamSpec       *pspec)
{
	CuiFilename *self = (CuiFilename *) object;
	
	switch(property_id){
		case PROPERTY_FILENAME:
			g_value_set_string(value, self->filename);
			break;
			
		case PROPERTY_DIRECTORY:
			g_value_set_string(value, self->directory);
			break;
			
		case PROPERTY_OWNER:
			g_value_set_string(value, self->owner);
			break;
			
		case PROPERTY_GROUP_OWNER:
			g_value_set_string(value, self->group_owner);
			break;

		case PROPERTY_SIZE:
			g_value_set_int64(value, self->size);
			break;
			
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID(
					object, property_id, pspec);
	}
}

GType 
cui_filename_get_type(void) 
{
	static GType type = 0;
	
	if (type == 0) {
		static const GTypeInfo info = {
			sizeof(CuiFilenameClass),
			(GBaseInitFunc) cui_filename_base_class_init, 
			(GBaseFinalizeFunc) NULL, 
			(GClassInitFunc) cui_filename_class_init, 
			NULL,
			NULL,
			sizeof(CuiFilename),
			0,
			(GInstanceInitFunc) cui_filename_instance_init 
		};
		type = g_type_register_static(G_TYPE_OBJECT, 
				"CuiFilename", &info, 0);
	}
	return type;
}

gchar *
cui_remove_tilde(const gchar *directory)
{
	if (*directory != '~')
		return g_strdup(directory);
	else
		return g_build_path(G_DIR_SEPARATOR_S, g_getenv("HOME"), 
				directory + 1, NULL);
}


/**
 * cui_filename_new:
 * @file: the basename of the file
 * @directory: the path of the directory holds this file
 * @returns: a new #CuiFilename object 
 *
 * Creates a new #CuiFilename which holds infomation about files and
 * directories.
 *
 * The function will unquote the filename with the #g_shell_unquote function of
 * the G library.
 */
CuiFilename *
cui_filename_new(const char *file, const char *directory)
{
	CuiFilenameClass *klass;
	gchar *plain_name;
	CuiFilename *filename;
	CuiDir *dir;
	GValue value = {0,};
	
	g_return_val_if_fail(file != NULL && directory != NULL, NULL);
	plain_name = g_shell_unquote(file, NULL);
	if (plain_name == NULL)
		plain_name = g_strdup(file);
	/*
	 * First we try to find the filename in the pool.
	 */
	klass = (CuiFilenameClass *)g_type_class_peek(CUI_FILENAME_TYPE);
	if (klass != NULL)
		filename = cui_filename_from_pool(klass, directory, plain_name);
	else
		filename = NULL;
	if (filename != NULL){
#ifdef DEBUG
		g_message("%s(): File in '%s' called '%s' found in pool.",
				__func__, file, directory);
#endif
		g_object_ref(G_OBJECT(filename));
		return filename;
	}
	/*
	 * We create.
	 */
	g_value_init(&value, G_TYPE_STRING);
	filename = g_object_new(CUI_FILENAME_TYPE, NULL);
	g_value_set_string(&value, plain_name);
	g_object_set_property(G_OBJECT(filename), "filename", &value);
	g_value_set_string(&value, directory);
	g_object_set_property(G_OBJECT(filename), "directory", &value);
	/*
	 * We store
	 */
	if (klass == NULL)
		klass = (CuiFilenameClass *)
			g_type_class_peek(CUI_FILENAME_TYPE);
	dir = directory_find(klass, directory);
	g_tree_insert(dir->tree, filename->filename, filename);
	return filename;
}

/**
 * cui_filename_get_filename:
 * @file: a #CuiFilename
 * @returns: the basename of this file 
 *
 * Returns the basename of the file. The return value is a newly allocated
 * memory should be freed by the caller.
 */
gchar *
cui_filename_get_filename(CuiFilename *file){
	gchar *filename;
	g_return_if_fail(CUI_IS_FILENAME(file));
	g_object_get(G_OBJECT(file), "filename", &filename, NULL);
	return g_strdup(filename);
}

/*
 * The stupid G library puts the filename into single quotes, so the shell will
 * not handle the ~ character as the user's home directory.
 */
static gchar *
shell_quote(gchar *filename)
{
	GString *quoted = g_string_new("");
	gunichar c;
	g_return_val_if_fail(filename != NULL, NULL);
	
	do{
		c = g_utf8_get_char(filename);
		filename = g_utf8_next_char(filename);
		if (c == ' ' || c == '(' || c ==  ')' || c == '$' ||
				c == '\'' || c == ';')
			quoted = g_string_append_unichar(quoted, '\\');
		quoted = g_string_append_unichar(quoted, c);
	} while (c != '\0');
	
	return g_string_free(quoted, FALSE);
}

/**
 * cui_filename_get_canonical_filename:
 * @file: a #CuiFilename
 * @working_directory: the current working directory
 * @returns: the basename of this file 
 *
 * Returns the simplest usable name for the file. The return value is newly 
 * allocated should be freed by the caller.
 */
gchar *
cui_filename_get_canonical_filename(CuiFilename *file, 
		gchar *working_directory)
{
	gchar *filename, *tmp;
	g_return_val_if_fail(CUI_IS_FILENAME(file), NULL);

	/*
	 * The root directory.
	 */
	if (file->filename[0] == '/')
		return g_strdup(file->filename);

	/*
	 *
	 */
	if (strcmp(file->directory, working_directory) != 0){
		tmp = g_build_filename(file->directory, file->filename, NULL);
		filename = shell_quote(tmp);
		g_free(tmp);
		return filename;
	}
	filename = shell_quote(file->filename);
	return filename;
}

/**
 * cui_filename_get_canonical_for_newname:
 * @file: a #CuiFilename:
 * @working_directory: the current working directory
 * @newname: the new name of the file 
 * @returns: the shortest usable new name for the file
 *
 * Returns what will be the canonical name of the given file if we change its
 * name to the new name. This file is used when generating commands like mv,
 * shich must be used like this:
 *
 *   mv ~/directory/filename ~/directory/newfilename
 *
 * The function returns a newly allocated string which should be freed.
 */
gchar *
cui_filename_get_canonical_for_newname(
		CuiFilename *file,
		gchar *working_directory,
		gchar *newname)
{
	gchar *filename, *tmp;
	g_return_val_if_fail(CUI_IS_FILENAME(file), NULL);

	if (strcmp(file->directory, working_directory) != 0){
		tmp = g_build_filename(file->directory, newname, NULL);
		filename = shell_quote(tmp);
		g_free(tmp);
		return filename;
	}
	filename = shell_quote(newname);
	return filename;
}
		

/**
 * cui_filename_get_owner:
 * @file: a #CuiFilename
 * @returns: the name of the owner of this file 
 *
 */
gchar *
cui_filename_get_owner(CuiFilename *file){
	gchar *owner;
	g_return_if_fail(CUI_IS_FILENAME(file));
	g_object_get(G_OBJECT(file), "owner", &owner, NULL);
	return g_strdup(owner);
}

/**
 * cui_filename_get_group:
 * @file: a #CuiFilename
 * @returns: the name of the group owner of this file 
 *
 */
gchar *
cui_filename_get_group(CuiFilename *file){
	gchar *owner;
	g_return_if_fail(CUI_IS_FILENAME(file));
	g_object_get(G_OBJECT(file), "group", &owner, NULL);
	return g_strdup(owner);
}

/**
 * cui_filename_set_mode_string:
 * @filename: a #CuiFilename
 * @mode_string: the mode string (first column of the output of 'ls -l')
 *
 * Sets the mode bits stored for this file by disassembling the mode string. The
 * mode string is looks like this: 'drwxr-----' as the ls program used to print
 * when the -l option is in use.
 *
 */
void
cui_filename_set_mode_string(CuiFilename *filename, 
		const gchar *mode_string)
{
	GValue val = {0,};
	g_return_if_fail(CUI_IS_FILENAME(filename));
	g_value_init(&val, G_TYPE_STRING);
	g_value_set_string (&val, mode_string);
	g_object_set_property(G_OBJECT(filename), "mode_string", &val);
}

/**
 * cui_filename_set_mode_string_owner:
 * @filename: a #CuiFilename
 * @string: the owner rights in the mode string (eg. "rwx")
 *
 * Sets the mode bits stored for this file by disassembling the mode string.
 * This function will change the stored mode string if it is already set.
 */
void
cui_filename_set_mode_string_owner(CuiFilename *filename, 
		const gchar *string)
{
	g_return_if_fail(CUI_IS_FILENAME(filename));
	g_return_if_fail(string != NULL);
	
	if( filename->mode_string != NULL) {
		filename->mode_string[1] = string[0];
		filename->mode_string[2] = string[1];
		filename->mode_string[3] = string[2];
	}
	/* 
	 * The rights of the user (I mean the owner). 
	 */
	switch (string[0]){
		case 'r':
			filename->user_read = MODE_YES;
			break;
		case '-':
			filename->user_read = MODE_NO;
			break;
		default:
			filename->user_read = MODE_NO;
	}
	switch (string[1]){
		case 'w':
			filename->user_write = MODE_YES;
			break;
		case '-':
			filename->user_write = MODE_NO;
			break;
		default:
			filename->user_write = MODE_UNKNOWN;
	}
	switch (string[2]){
		case 'x':
			filename->user_execute = MODE_YES;
			break;
		case '-':
			filename->user_execute = MODE_NO;
			break;
		default:
			filename->user_execute = MODE_UNKNOWN;
	}
}

/**
 * cui_filename_set_mode_string_group:
 * @filename: a #CuiFilename
 * @string: the group rights in the mode string (eg. "rwx")
 *
 * Sets the mode bits stored for this file by disassembling the mode string.
 * This function will change the stored mode string if it is already set.
 */
void
cui_filename_set_mode_string_group(CuiFilename *filename, 
		const gchar *string)
{
	g_return_if_fail(CUI_IS_FILENAME(filename));
	g_return_if_fail(string != NULL);
	
	if( filename->mode_string != NULL) {
		filename->mode_string[4] = string[0];
		filename->mode_string[5] = string[1];
		filename->mode_string[6] = string[2];
	}
	/* 
	 * The rights of the group owner. 
	 */
	switch (string[0]){
		case 'r':
			filename->group_read = MODE_YES;
			break;
		case '-':
			filename->group_read = MODE_NO;
			break;
		default:
			filename->group_read = MODE_NO;
	}
	switch (string[1]){
		case 'w':
			filename->group_write = MODE_YES;
			break;
		case '-':
			filename->group_write = MODE_NO;
			break;
		default:
			filename->group_write = MODE_UNKNOWN;
	}
	switch (string[2]){
		case 'x':
			filename->group_execute = MODE_YES;
			break;
		case '-':
			filename->group_execute = MODE_NO;
			break;
		default:
			filename->group_execute = MODE_UNKNOWN;
	}
}

/**
 * cui_filename_set_mode_string_others:
 * @filename: a #CuiFilename
 * @string: the others rights in the mode string (eg. "rwx")
 *
 * Sets the mode bits stored for this file by disassembling the mode string.
 * This function will change the stored mode string if it is already set.
 */
void
cui_filename_set_mode_string_others(CuiFilename *filename, 
		const gchar *string)
{
	g_return_if_fail(CUI_IS_FILENAME(filename));
	g_return_if_fail(string != NULL);
	
	if( filename->mode_string != NULL) {
		filename->mode_string[7] = string[0];
		filename->mode_string[8] = string[1];
		filename->mode_string[9] = string[2];
	}
	/* 
	 * The rights of the others 
	 */
	switch (string[0]){
		case 'r':
			filename->others_read = MODE_YES;
			break;
		case '-':
			filename->others_read = MODE_NO;
			break;
		default:
			filename->others_read = MODE_NO;
	}
	switch (string[1]){
		case 'w':
			filename->others_write = MODE_YES;
			break;
		case '-':
			filename->others_write = MODE_NO;
			break;
		default:
			filename->others_write = MODE_UNKNOWN;
	}
	switch (string[2]){
		case 'x':
			filename->others_execute = MODE_YES;
			break;
		case '-':
			filename->others_execute = MODE_NO;
			break;
		default:
			filename->others_execute = MODE_UNKNOWN;
	}
}

/**
 * cui_filename_set_owner:
 * @filename: a #CuiFilename
 * @owner: the user name of the owner
 *
 * Sets the file owner's name for the filename.
 */
void
cui_filename_set_owner(CuiFilename *filename, const gchar *owner)
{
	GValue val = {0,};
	g_return_if_fail(CUI_IS_FILENAME(filename));
	g_value_init(&val, G_TYPE_STRING);
	g_value_set_string (&val, owner);
	g_object_set_property(G_OBJECT(filename), "owner", &val);
}

/**
 * cui_filename_set_group:
 * @filename: a #CuiFilename
 * @group: the group name
 *
 * Sets the group owner's name for the filename.
 */
void
cui_filename_set_group(CuiFilename *filename, const gchar *group)
{
	GValue val = {0,};
	g_return_if_fail(CUI_IS_FILENAME(filename));
	g_value_init(&val, G_TYPE_STRING);
	g_value_set_string (&val, group);
	g_object_set_property(G_OBJECT(filename), "group", &val);
}


/**
 * cui_filename_set_owner_and_group:
 * @filename: a #CuiFilename
 * @owner: the user name of the owner
 * @group: the group name
 *
 * Sets the name of the file owner and group owner for the filename.
 */
void
cui_filename_set_owner_and_group(CuiFilename *filename, 
		const gchar *owner,
		const gchar *group)
{
	GValue val = {0,};
	g_return_if_fail(CUI_IS_FILENAME(filename));
	g_value_init(&val, G_TYPE_STRING);
	g_value_set_string (&val, owner);
	g_object_set_property(G_OBJECT(filename), "owner", &val);
	g_value_set_string (&val, group);
	g_object_set_property(G_OBJECT(filename), "group", &val);
}

/**
 * cui_filename_get_filetype:
 * @filename: a #Filename object pointer
 * @returns: the type of the file
 *
 * Returns the type of the file (such as regular file, directory, block special
 * file, etc.).
 *
 */
CuiFileType
cui_filename_get_filetype(CuiFilename *filename)
{
	return filename->file_type;
}

/**
 * cui_filename_get_typename:
 * @filename: a #Filename object pointer
 * @returns: the type of the file
 *
 * Returns the type of the file as a string (such as "regular file", 
 * "directory", "block special file", etc.). The string should not modified oor
 * freed.
 *
 */
const gchar *
cui_filename_get_typename(CuiFilename *filename)
{
	switch (filename->file_type){
		case FILETYPE_FILE:
			return _("regular file");
			
		case FILETYPE_DIRECTORY:
			return _("directory");
			
		case FILETYPE_BLOCK:
			return _("block special file");
			
		case FILETYPE_CHAR:
			return _("character special file");
			
		case FILETYPE_FIFO:
			return _("named pipe");
			
		case FILETYPE_LINK:
			return _("soft link");
			
		case FILETYPE_SOCKET:
			return _("socket");
			
		default:
			return _("unknown file");
	}
}

/**
 * cui_filename_is_compressed:
 * @filename: a #CuiFilename
 * @returns: TRUE if the file is compressed
 *
 * Returns TRUE if the file is a compressed file. We return FALSE for a tar
 * archive (if not compressed) since the user may want to compress these files
 * with some other program. 
 */
gboolean
cui_filename_is_compressed(CuiFilename *filename)
{
	gint n;
	gchar *filters[] = {
		"*.gz", "*.bz2", "*.zip", "*.tgz", "*.rar", 
		"*.lha", "*.arj", NULL
	};

	g_return_val_if_fail(filename->filename != NULL, FALSE);
	for (n = 0; filters[n] != NULL; ++n){
		if (fnmatch(filters[n], filename->filename, 0) == 0)
			return TRUE;
	}
	return FALSE;
}

/**
 * cui_filename_is_dir:
 * @filename: a #CuiFilename 
 *
 * Returns TRUE if the given file is a directory type entry.
 *
 */
gboolean
cui_filename_is_dir(CuiFilename *filename)
{
	g_return_val_if_fail(CUI_IS_FILENAME(filename), FALSE);
	return filename->file_type == FILETYPE_DIRECTORY; 
}

/**
 * cui_filename_is_block_device:
 * @filename: a #CuiFilename 
 *
 * Returns TRUE if the given file is a block device.
 *
 */
gboolean
cui_filename_is_block_device(CuiFilename *filename)
{
	g_return_val_if_fail(CUI_IS_FILENAME(filename), FALSE);
	return filename->file_type == FILETYPE_BLOCK; 
}

/**
 * cui_filename_is_character_device:
 * @filename: a #CuiFilename 
 *
 * Returns TRUE if the given file is a character device.
 *
 */
gboolean
cui_filename_is_character_device(CuiFilename *filename)
{
	g_return_val_if_fail(CUI_IS_FILENAME(filename), FALSE);
	return filename->file_type == FILETYPE_CHAR; 
}

/**
 * cui_filename_is_user_read:
 * @filename: a #CuiFilename 
 *
 * Returns TRUE if the user has the right to read the file.
 *
 */
gboolean
cui_filename_is_user_read(CuiFilename *filename)
{
	g_return_val_if_fail(CUI_IS_FILENAME(filename), FALSE);
	return filename->user_read == MODE_YES; 
}

/**
 * cui_filename_is_user_write:
 * @filename: a #CuiFilename 
 *
 * Returns TRUE if the user has the right to write the file.
 *
 */
gboolean
cui_filename_is_user_write(CuiFilename *filename)
{
	g_return_val_if_fail(CUI_IS_FILENAME(filename), FALSE);
	return filename->user_write == MODE_YES; 
}

/**
 * cui_filename_is_user_execute:
 * @filename: a #CuiFilename 
 *
 * Returns TRUE if the user has the right to execute the file.
 *
 */
gboolean
cui_filename_is_user_execute(CuiFilename *filename)
{
	g_return_val_if_fail(CUI_IS_FILENAME(filename), FALSE);
	return filename->user_execute == MODE_YES; 
}

/**
 * cui_filename_is_group_read:
 * @filename: a #CuiFilename 
 *
 * Returns TRUE if the group has the right to read the file.
 *
 */
gboolean
cui_filename_is_group_read(CuiFilename *filename)
{
	g_return_val_if_fail(CUI_IS_FILENAME(filename), FALSE);
	return filename->group_read == MODE_YES; 
}

/**
 * cui_filename_is_group_write:
 * @filename: a #CuiFilename 
 *
 * Returns TRUE if the group has the right to write the file.
 *
 */
gboolean
cui_filename_is_group_write(CuiFilename *filename)
{
	g_return_val_if_fail(CUI_IS_FILENAME(filename), FALSE);
	return filename->group_write == MODE_YES; 
}

/**
 * cui_filename_is_group_execute:
 * @filename: a #CuiFilename 
 *
 * Returns TRUE if the group has the right to execute the file.
 *
 */
gboolean
cui_filename_is_group_execute(CuiFilename *filename)
{
	g_return_val_if_fail(CUI_IS_FILENAME(filename), FALSE);
	return filename->group_execute == MODE_YES; 
}


/**
 * cui_filename_is_others_read:
 * @filename: a #CuiFilename 
 *
 * Returns TRUE if the others has the right to read the file.
 *
 */
gboolean
cui_filename_is_others_read(CuiFilename *filename)
{
	g_return_val_if_fail(CUI_IS_FILENAME(filename), FALSE);
	return filename->others_read == MODE_YES; 
}

/**
 * cui_filename_is_others_write:
 * @filename: a #CuiFilename 
 *
 * Returns TRUE if the others has the right to write the file.
 *
 */
gboolean
cui_filename_is_others_write(CuiFilename *filename)
{
	g_return_val_if_fail(CUI_IS_FILENAME(filename), FALSE);
	return filename->others_write == MODE_YES; 
}

/**
 * cui_filename_is_others_execute:
 * @filename: a #CuiFilename 
 *
 * Returns TRUE if the others has the right to execute the file.
 *
 */
gboolean
cui_filename_is_others_execute(CuiFilename *filename)
{
	g_return_val_if_fail(CUI_IS_FILENAME(filename), FALSE);
	return filename->others_execute == MODE_YES; 
}

/**
 * cui_filename_set_filetype_dir:
 * @filename: a #CuiFilename
 *
 * Sets the type of the filename to directory type entry.
 */
void
cui_filename_set_filetype_dir(CuiFilename *filename)
{
	g_return_if_fail(CUI_IS_FILENAME(filename));
	filename->file_type = FILETYPE_DIRECTORY;
}

/**
 * cui_filename_set_filetype_block:
 * @filename: a #CuiFilename
 *
 * Sets the type of the filename to block device type entry.
 */
void
cui_filename_set_filetype_block(CuiFilename *filename)
{
	g_return_if_fail(CUI_IS_FILENAME(filename));
	filename->file_type = FILETYPE_BLOCK;
}

/**
 * cui_filename_set_filetype_regular:
 * @filename: a #CuiFilename
 *
 * Sets the type of the filename to regular file.
 */
void
cui_filename_set_filetype_regular(CuiFilename *filename)
{
	g_return_if_fail(CUI_IS_FILENAME(filename));
	filename->file_type = FILETYPE_FILE;
	//g_object_notify(G_OBJECT(widget), "");
}

/**
 * cui_filename_set_filetype_link:
 * @filename: a #CuiFilename
 *
 * Sets the type of the filename to soft link.
 */
void
cui_filename_set_filetype_link(CuiFilename *filename)
{
	g_return_if_fail(CUI_IS_FILENAME(filename));
	filename->file_type = FILETYPE_LINK;
	//g_object_notify(G_OBJECT(widget), "");
}

void
cui_filename_set_file_size(CuiFilename *filename, 
		gint64 size)
{
	g_return_if_fail(CUI_IS_FILENAME(filename));
	g_object_set(G_OBJECT(filename), 
			"size", size,
			NULL);
}

CuiFileSizeChange
cui_filename_get_size_change(CuiFilename *filename)
{
	g_return_val_if_fail(CUI_IS_FILENAME(filename), FileSizeUnknown);
	return filename->sizechange;
}

/**
 * cui_filename_get_editor:
 * @filename: a #CuiFilename
 * @long_name: the full name of file used in the command
 * @returns: a command which will start up an editor program for this file
 *
 * Returns a command line string which can be used to open an editor (not a
 * viewer) for the given file. The string is in a newly allocated memory area
 * which must be freed with the g_free() function.
 */
gchar *
cui_filename_get_editor(CuiFilename *filename, gchar *long_name)
{
	int n;
	gchar *types[] = {
		"*.odt",
		"*.rtf",
		"*.doc",
		"*.xls",
		"*.ppt",
		"*.png",      
		"*.jpg", 
		"*.svg",
		"*.gif",
		"*.glade",
		NULL
	};
	
	gchar *apps[] = {
		"oowriter %s", 
		"oowriter %s", 
		"oowriter %s", 
		"oocalc %s",
		"ooimpress %s",
		"gimp %s",
		"gimp %s",
		"gimp %s",
		"gimp %s",
		"glade-2 %s",
		NULL
	};
	
	g_return_val_if_fail(CUI_IS_FILENAME(filename), NULL);

	n = 0;
	while (types[n] != NULL){
		if (fnmatch(types[n], filename->filename, 0) == 0){
			return g_strdup_printf(apps[n], long_name);
		}
		++n;
	}

	return NULL;
}

/**
 * cui_filename_get_viewer:
 * @filename: a #CuiFilename
 * @long_name: the full name of file used in the command
 * @returns: a command which will start up a viewer program for this file
 *
 * Returns a command line string which can be used to open a viewer (not an
 * editor) for the given file. The string is in a newly allocated memory area
 * which must be freed with the g_free() function.
 */
gchar *
cui_filename_get_viewer(CuiFilename *filename, gchar *long_name)
{
	int n;
	gchar *types[] = {
		"*.pdf",
		"*.dvi",
		"*.avi",
		"*.mpg",
		"*.mp3",
		"*.tar",
		"*.tar.gz",
		"*.tgz",
		"*.tar.Z",
		"*.bz2",
		NULL
	};
	
	gchar *apps[] = {
		"acroread %s", 
		"xdvi %s",
		"mplayer %s",
		"mplayer %s",
		"mplayer %s",
		"tar tf %s",
		"tar tzf %s",
		"tar tzf %s",
		"tar tZf %s",
		"tar tjf %s",
		NULL
	};
	
	g_return_val_if_fail(CUI_IS_FILENAME(filename), NULL);

	n = 0;
	while (types[n] != NULL){
		if (fnmatch(types[n], filename->filename, 0) == 0){
			return g_strdup_printf(apps[n], long_name);
		}
		++n;
	}

	return NULL;
}


/**
 * cui_filename_get_processor:
 * @filename: a #CuiFilename
 * @long_name: the full name of file used in the command
 * @returns: a command which will start up a processor program for this file
 *
 * Returns a command line string which can be used to open a processor (not a
 * viewer) for the given file. The string is in a newly allocated memory area
 * which must be freed with the g_free() function.
 */
gchar *
cui_filename_get_processor(CuiFilename *filename, 
		gchar *long_name)
{
	int n;
	gchar *types[] = {
		"*.sh",
		"install-sh",
		"Makefile",
		"*.make",
		"*.patch",
		"*.tar",
		"*.tar.gz",
		"*.tgz",
		"*.tar.Z",
		"*.bz2",
		NULL
	};
	
	gchar *apps[] = {
		"bash %s",
		"bash %s",
		"make",
		"make -f %s",
		"patch -p 1 <%s",
		"tar xf %s",
		"tar xzf %s",
		"tar xzf %s",
		"tar xZf %s",
		"tar xjf %s",
		NULL
	};
	
	g_return_val_if_fail(CUI_IS_FILENAME(filename), NULL);
	/*
	 * If this file is an executable we simple run it. 
	 * FIXME: We should scrutinize this.
	 */
	if (filename->user_execute == MODE_YES ||
			filename->group_execute == MODE_YES ||
			filename->others_execute == MODE_YES)
		if (*long_name == '/' || *long_name == '~')
			return g_strdup_printf("%s", long_name);
		else
			return g_strdup_printf("./%s", long_name);
				
	/*
	 * Otherwise we try to open it with an appropriate application.
	 */
	n = 0;
	while (types[n] != NULL){
		if (fnmatch(types[n], filename->filename, 0) == 0){
			return g_strdup_printf(apps[n], long_name);
		}
		++n;
	}

	return NULL;
}
