/*
 * Copyright (C) 2004 Laszlo Pere
 * $Id:$
 *
 * This is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Library General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Log:$
 */
#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#define _GNU_SOURCE
#include <sys/types.h>
#include <unistd.h>
#include <gnome.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <stdio.h>

#include <bonobo.h>
#include <gnome.h>

#include "terminal.h"
#include "support.h"
#include "config.h"
#include "bash_parser_c.h"
#include "history.h"

#if DEBUG_ALL
#   define DEBUG
#endif

#define DEBUG

/***************************************************************************
 * CUI Completion is a set of functions for handling the programmable      *
 * command completion of the shell.                                        *
 *                                                                         *
 ***************************************************************************/
static GtkWidget *
create_command_completion_tree(void)
{
	GtkWidget *command_tree;
	GtkTreeStore *tree_store = gtk_tree_store_new(1, 
			G_TYPE_STRING  /* */ 
			);
	GtkTreeIter iter_level0;
	GtkTreeIter iter_level1;
	GtkTreeIter iter_level2;
	GtkTreeIter iter_level3;
	GtkTreeViewColumn *column;
	GtkCellRenderer   *renderer;

  
	gtk_tree_store_append(tree_store, &iter_level0, NULL);
	gtk_tree_store_set(tree_store, &iter_level0, 
	  	0, _("File Management"), 
	  	-1);
	gtk_tree_store_append(tree_store, &iter_level1, &iter_level0);
	gtk_tree_store_set(tree_store, &iter_level1, 
	  	0, _("List Files"), 
	  	-1);
	
	
	command_tree = gtk_tree_view_new();
	gtk_widget_set_size_request(command_tree, 300, 100);

	gtk_tree_view_set_model(command_tree, GTK_TREE_MODEL(tree_store));
  
	renderer = gtk_cell_renderer_text_new ();
  	column = gtk_tree_view_column_new_with_attributes(
		  _("Command"), 
		  renderer, "text", 0, NULL);
	gtk_tree_view_append_column(command_tree, column);

	gtk_tree_view_set_headers_visible(command_tree, FALSE);
	gtk_widget_show(command_tree);
}

gboolean
cui_completion_activate(Terminal *t)
{
	gboolean initialized = FALSE;
	GtkWidget *command_tree;
	
	g_assert(t != NULL);
#ifdef DEBUG
	g_message("%s(): Starting on command: '%s'", __func__, 
			t->current_command);
#endif
	gtk_text_view_set_border_window_size(
			t->text_view, 
			GTK_TEXT_WINDOW_BOTTOM,
			100);
	
	command_tree = create_command_completion_tree();

	gtk_text_view_add_child_in_window(
			t->text_view,
			command_tree,
			GTK_TEXT_WINDOW_BOTTOM,
			0, 0);
	initialized = TRUE;
}

/***************************************************************************
 * GUI Completion is special service used to generate dialog windows for   *
 * the command being edited.                                               *
 *                                                                         *
 ***************************************************************************/
static inline gboolean
is_option(const gchar *option)
{
	g_assert(option != NULL);
	return option[0] == '-' && option[1] != '-';
}

static inline gboolean
is_long_option(const gchar *option)
{
	g_assert(option != NULL);
	return option[0] == '-' && option[1] == '-';
}
	
static gchar * 
create_temporary_program_name(void)
{
	GTimeVal  time;
	pid_t     pid;
	gchar    *name;
	
	pid = getpid();
	g_get_current_time(&time);
	name = g_strdup_printf("/var/tmp/cuiterm-program-%d-%ld.sh",
			pid, time.tv_sec);
	/*
	 * Some check would be nice.
	 */
#ifdef DEBUG
	g_message("%s(): filename: '%s'", __func__, name);
#endif
	return name;
}

gchar *
option_to_variable_name(gchar *option)
{
	gunichar c;
	GString *variable = g_string_new("OPTION");

	do {
		c = g_utf8_get_char(option);
		option = g_utf8_next_char(option);
		switch(c) {
			case ' ':
			case '-':
				g_string_append_unichar(variable, '_');
				break;
			default:
				g_string_append_unichar(variable, c);
		}
	} while (c != '\0');
	
	//g_message("%s(): '%s'", __func__, variable->str);
	return g_string_free(variable, FALSE);
}

static void
print_checkbox_for_option(
		FILE        *outfile,
		Terminal    *t,	
		const gchar *program,
		const gchar *option)
{
	gboolean  is_on;
	gchar    *variable_name;
	
	
	g_assert(outfile != NULL && program != NULL && option != NULL);
#ifdef DEBUG
	g_message("%s(): Starting.", __func__);
#endif
	variable_name = option_to_variable_name(option);

	is_on = parse_tree_check_option(t->parse_tree, program, option);
				
	g_fprintf(outfile,
"    <checkbox>\n"
"      <label>%s</label>\n"
"      <action>refresh:COMMAND</action>\n"
"      <variable>%s</variable>\n"
"      <input>echo %s</input>\n"
"    </checkbox>\n",  
		program_get_option_description(program, option), 
		variable_name,
		is_on ? "true" : "false");
finalize:
	g_free(variable_name);
}
		

static FILE * 
open_temporary_program_for_write(const gchar *filename)
{
	FILE *outfile;
	
	g_assert(filename != NULL);
	
	outfile = g_fopen(filename, "w");
	if (outfile == NULL)
		g_warning("%s(): Unable to open file '%s' for writing: %m",
				__func__, filename);
	return outfile;
				
}

gboolean 
option_islong(gchar *option)
{
	g_assert(option != NULL);
	if (option[0] == '-' && option[1] == '-')
		return TRUE;
	else
		return FALSE;
}

gchar 
option_remove_dash(gchar *option)
{
	g_assert(option != NULL);
	if (option[0] != '-')
		return option[0];
	else
		return option[1];
}

static gboolean
print_first_line(FILE *outfile, const gchar *filename)
{
	GtkWidget *dialog;
	gchar     *basename;;
	gchar     *gtkdialog_full_name;
	
	g_assert(outfile != NULL && filename != NULL);
#ifdef DEBUG
	g_message("%s(): Starting.", __func__);
#endif
	
	gtkdialog_full_name = g_find_program_in_path("gtkdialog");
	if (gtkdialog_full_name == NULL) {
		 dialog = gtk_message_dialog_new (NULL,
                                  GTK_DIALOG_DESTROY_WITH_PARENT,
                                  GTK_MESSAGE_ERROR,
                                  GTK_BUTTONS_CLOSE,
_(      "The GtkDialog utility is needed for GUI completion, but could not "
	"found in path. It can be downloaded from http://linux.pte.hu/~pipas/."
	));
 		gtk_dialog_run (GTK_DIALOG (dialog));
 		gtk_widget_destroy (dialog);
		return FALSE;
	}

	basename = g_path_get_basename(filename);
	/*
	 * Printing the first line containing the interpreter location and some
	 * comments in the case of a leftover file.
	 */
	g_fprintf(outfile, "#! %s -e\n", gtkdialog_full_name);
	g_free(gtkdialog_full_name);

	g_fprintf(outfile, 
_(
"#\n"
"# %s: automaticaly generated by the Cuiterm\n"
"# application (download from: http://linux.pte.hu/~pipas)\n"
"# \n"
"# Copyright: (c) 2006 by Laszlo Pere\n"
"# Email: pipas@linux.pte.hu\n"
"#\n"
"# This program is free software; you can redistribute it and/or modify\n"
"# it under the terms of the GNU General Public License as published by the\n"
"# Free Software Foundation; either version 2 of the License, or (at your \n"
"# option) any later version.\n"
"# \n"
"\n"
), basename);

	g_free(basename);
	return TRUE;	
}

static gboolean
print_command_line_function(FILE *outfile,
		const gchar *producer)
{
	gchar   *variable_name = NULL;
	GList   *list;
	guint    list_length;
	GList   *li;
	
#ifdef DEBUG
	g_message("%s(): Starting.", __func__);
#endif
	list = program_get_option_list(producer, NULL); 

	/*
	 * The beginning of the function.
	 */
	g_fprintf(outfile,
"#\n"
"# This function will assemble a command line using the state of the widgets\n"
"# on the graphical user interface.\n"
"#\n"
"function command_line() {\n"
"  command=\"%s\"\n"
"  options=\"\"\n"
"  long_options=\"\"\n\n", producer);
	
	/* The long options. */
	li = list;
	while (li != NULL) {
		if (!option_islong(li->data))
			goto next_item1;
		variable_name = option_to_variable_name(li->data);	
		g_fprintf(outfile, 
"  [ \"$%s\" == \"true\" ] && long_options=\"${long_options} %s\"\n",
				variable_name, li->data);
		g_free(variable_name);
next_item1:
		li = li->next;
	}
	
	/* The short options. */
	li = list;
	while (li != NULL) {
		if (option_islong(li->data))
			goto next_item2;
		variable_name = option_to_variable_name(li->data);	
		g_fprintf(outfile, 
"  [ \"$%s\" == \"true\" ] && options=\"${options}%c\"\n",
				variable_name, option_remove_dash(li->data));
		g_free(variable_name);
next_item2:
		li = li->next;
	}

	g_list_free(list);
	
	/*
	 * The end of the function.
	 */
	g_fprintf(outfile,
"\n"
"  if [ \"$options\" ]; then\n"
"    command=\"$command -$options\"\n"
"  fi\n"
"  if [ \"$long_options\" ]; then\n"
"    command=\"$command $long_options\"\n"
"  fi\n"
"  command=\"$command $FILE_LIST_ALL\" #FIXME: this is unnecessary\n"
"  echo -n $command | tr -s ' '\n"
"}\n\n");
	
	return TRUE;
}


static gint
print_important_options(
		FILE        *outfile,
		Terminal    *t,
		const gchar *producer)
{
	gint     n;
	gchar   *variable_name = NULL;
	GList   *list;
	guint    list_length;
	GList   *li;
	gint     rank;
	gchar   *restrictions[] = {
			    "not (importance/@prompt   = 'true') "
			"and not (importance/@ignored  = 'true') "
			"and not (importance/@indirect = 'true') "
			"and not (importance/@standard = 'true') "
			,
			    "not (importance/@prompt   = 'true') "
			"and not (importance/@ignored  = 'true') "
			"and not (importance/@indirect = 'true') "
			"and not (importance/@standard = 'true') "
			"and     (description/@length < 75) "
			,
			    "not (importance/@prompt   = 'true') "
			"and not (importance/@ignored  = 'true') "
			"and not (importance/@indirect = 'true') "
			"and not (importance/@standard = 'true') "
			"and     (description/@length < 75) "
			"and     (importance/counter/@in-history > 0)"
			,
			    "not (importance/@prompt   = 'true') "
			"and not (importance/@ignored  = 'true') "
			"and not (importance/@indirect = 'true') "
			"and not (importance/@standard = 'true') "
			"and     (description/@length < 75) "
			"and     (importance/counter/@in-history > 20)"
			,
			NULL};
#ifdef DEBUG
	g_message("%s(): Starting.", __func__);
#endif
	/*
	 * Here we try to create a not too long list of the important options. 
	 * If there are too much we get, we try a more rigorous selection.
	 */
	rank = 0;
	while (restrictions[rank] != NULL) {
		list = program_get_option_list(producer, restrictions[rank]);
		list_length = g_list_length(list);
		g_message("%s(): selection %d, got %d item", 
				__func__, rank, list_length);
		if (list_length < 8)
			break;
step_up:
		++rank;
	}
	
	if (list_length == 0) {
		if (rank == 0)
			return 0;
		/*
		 * Step back to a longer list.
		 */
		rank--;
		list = program_get_option_list(producer, restrictions[rank]);
		list_length = g_list_length(list);
		g_message("%s(): selection %d, got %d item", 
				__func__, rank, list_length);
		if (list_length == 0)
			return 0;
	}
	/* 
	 * The options box and widgets. 
	 */
	fprintf(outfile, _("  <frame Important Options>\n"));
	
	n = 0;
	li = list;
	while (li != NULL) {
		/*
		 * Sorry, we definitly do not want to print lot of checkboxes.
		 */
		if (n > 12)
			break;
#ifdef DEBUG_MORE
		g_message("%s(): '%s'-'%s'",
				__func__,
				li->data, 
				program_get_option_description(producer, li->data));
#endif
	print_checkbox_for_option(outfile, t, producer, li->data);
next_item:
		li = li->next;
		n++;
	}
	fprintf(outfile, _("  </frame>\n"));

	g_list_free(list);
	return list_length;
}

static gint
print_prompt_options(FILE   *outfile,
		Terminal    *t,
		const gchar *producer)
{
	gchar    *variable_name = NULL;
	GList    *list;
	guint     list_length;
	GList    *li;
	gint      rank;
	gboolean  is_on;
	gchar    *restrictions[] = {
			    "    (importance/@prompt   = 'true') "
			"and not (importance/@ignored  = 'true') "
			"and not (importance/@indirect = 'true') "
			"and not (importance/@standard = 'true') "
			,
			    "    (importance/@prompt   = 'true') "
			"and not (importance/@ignored  = 'true') "
			"and not (importance/@indirect = 'true') "
			"and not (importance/@standard = 'true') "
			"and     (description/@length < 50) "
			,
			    "    (importance/@prompt   = 'true') "
			"and not (importance/@ignored  = 'true') "
			"and not (importance/@indirect = 'true') "
			"and not (importance/@standard = 'true') "
			"and     (description/@length < 50) "
			"and     (importance/counter/@in-history > 10)"
			,
			NULL};
#ifdef DEBUG
	g_message("%s(): Starting.", __func__);
#endif
			
	/*
	 * Here we try to create a not too long list of the important 
	 * options. If there are too much we get, we try a more rigorous 
	 * selection.
	 */
	rank = 0;
	while (restrictions[rank] != NULL) {
		list = program_get_option_list(producer, restrictions[rank]);
		list_length = g_list_length(list);
		if (list_length < 8)
			break;
step_up:
		++rank;
	}
	
	if (list_length == 0) {
		if (rank == 0)
			return 0;
		/*
		 * Step back to a longer list.
		 */
		rank--;
		list = program_get_option_list(producer, restrictions[rank]);
		list_length = g_list_length(list);
		if (list_length == 0)
			return 0;
	}

	/* 
	 * The options box and widgets. 
	 */
	g_fprintf(outfile, _("  <frame Prompt/Overwrite Options>\n"));
	g_fprintf(outfile,
"    <radiobutton>\n"
"      <variable>AUTO_OPTION_OVERWRITE_NONE</variable>\n"
"      <label>Use the default setting.</label>\n"
"      <action>refresh:COMMAND</action>\n"
"    </radiobutton>\n");
	
	li = list;
	while (li != NULL) {
#ifdef DEBUG_MORE
		g_message("%s(): '%s'-'%s'",
				__func__,
				li->data, 
				program_get_option_description(producer, li->data));
#endif
		is_on = parse_tree_check_option(t->parse_tree, producer, li->data);
		variable_name = option_to_variable_name(li->data);
		fprintf(outfile,
"    <radiobutton>\n"
"      <variable>%s</variable>\n"
"      <label>%s</label>\n"
"      <action>refresh:COMMAND</action>\n"
"      <input>echo %s</input>\n"
"    </radiobutton>\n", variable_name, 
			program_get_option_description(producer, li->data), 
			is_on ? "true" : "false");
		g_free(variable_name);
next_item:
		li = li->next;
	}
	fprintf(outfile, _("  </frame>\n"));

	g_list_free(list);
	return TRUE;
}

static gint
print_standard_options(
		FILE        *outfile,
		Terminal    *t,
		const gchar *producer)
{
	gchar   *variable_name = NULL;
	GList   *list;
	guint    list_length;
	GList   *li;
	gint     rank;
	gchar   *restrictions[] = {
			    "not (importance/@prompt   = 'true') "
			"and     (importance/@standard = 'true') "
			"and not (importance/@ignored  = 'true') "
			"and not (importance/@indirect = 'true') "
			,
			    "not (importance/@prompt   = 'true') "
			"and     (importance/@standard = 'true') "
			"and not (importance/@ignored  = 'true') "
			"and not (importance/@indirect = 'true') "
			"and     (description/@length < 50) "
			,
			    "not (importance/@prompt   = 'true') "
			"and     (importance/@standard = 'true') "
			"and not (importance/@ignored  = 'true') "
			"and not (importance/@indirect = 'true') "
			"and     (description/@length < 50) "
			"and     (importance/counter/@in-history > 10)"
			,
			NULL};
#ifdef DEBUG
	g_message("%s(): Starting.", __func__);
#endif
	/*
	 * Here we try to create a not too long list of the important 
	 * options. If there are too much we get, we try a more rigorous 
	 * selection.
	 */
	rank = 0;
	while (restrictions[rank] != NULL) {
		list = program_get_option_list(producer, restrictions[rank]);
		list_length = g_list_length(list);
		if (list_length < 8)
			break;
step_up:
		++rank;
	}
	
	if (list_length == 0) {
		if (rank == 0)
			return 0;
		/*
		 * Step back to a longer list.
		 */
		rank--;
		list = program_get_option_list(producer, restrictions[rank]);
		list_length = g_list_length(list);
		g_message("%s(): selection %d, got %d item", 
				__func__, rank, list_length);
		if (list_length == 0)
			return 0;
	}
	/* 
	 * The options box and widgets. 
	 */
	fprintf(outfile, _("  <frame Standard Options>\n"));
	
	li = list;
	while (li != NULL) {
#ifdef DEBUG_MORE
		g_message("%s(): '%s'-'%s'",
				__func__,
				li->data, 
				program_get_option_description(producer, li->data));
#endif
		print_checkbox_for_option(outfile, t, producer, li->data);
next_item:
		li = li->next;
	}
	fprintf(outfile, _("  </frame>\n"));

	g_list_free(list);
	return list_length;
}

static gint
print_file_selection(FILE *outfile,
		const gchar *producer,
		gboolean iterate,
		gboolean optional)
{
#ifdef DEBUG
	g_message("%s(): Starting.", __func__);
#endif
	if (!iterate && optional) 
		g_fprintf(outfile, 
"  <frame File>\n"
"    <checkbox>\n"
"      <label>Use the following filename:</label>\n"
"      <variable>USE_FILE</variable>\n"
"      <action>echo $USE_FILE</action>\n"
"      <action>if true enable:FILE_OPTION</action>\n"
"      <action>if true enable:FILE_BROWSE</action>\n"
"      <action>if false enable:FILE_OPTION</action>\n"
"      <action>if false enable:FILE_BROWSE</action>\n"
"    </checkbox>\n"
"    <hbox>\n"
"      <entry>\n"
"        <variable>FILE_OPTION</variable>\n"
"      </entry>\n"
"      <button>\n"
"        <label>Browse</label>\n"
"        <input file stock=\"gtk-open\"></input>\n"
"        <variable>FILE_BROWSE</variable>\n"
"        <action type=\"fileselect\">FILE_OPTION</action>\n"
"      </button>\n"
"    </hbox>\n"
"  </frame>\n");
	else if (!iterate && !optional)
		g_fprintf(outfile, 
"  <frame File>\n"
"    <hbox>\n"
"      <entry>\n"
"        <variable>FILE_OPTION</variable>\n"
"      </entry>\n"
"      <button>\n"
"        <label>Browse</label>\n"
"        <input file stock=\"gtk-open\"></input>\n"
"        <variable>FILE_BROWSE</variable>\n"
"        <action type=\"fileselect\">FILE_OPTION</action>\n"
"      </button>\n"
"    </hbox>\n"
"  </frame>\n");
	else // incomplete
		g_fprintf(outfile, 
"  <frame Files>\n"
"    <hbox>\n"
"      <list>\n"
"        <variable>FILE_LIST</variable>\n"
"      </list>\n"
"    </hbox>\n"
"    <hbox>\n"
"      <entry>\n"
"        <variable>FILE_OPTION</variable>\n"
"      </entry>\n"
"      <button>\n"
"        <input file stock=\"gtk-open\"></input>\n"
"        <variable>FILE_BROWSE</variable>\n"
"        <action type=\"fileselect\">FILE_OPTION</action>\n"
"      </button>\n"
"      <button>\n"
"        <input file stock=\"gtk-add\"></input>\n"
"        <action type=\"insert\">FILE_OPTION,FILE_LIST</action>\n"
"        <action type=\"refresh\">COMMAND</action>\n"
"      </button>\n"
"      <button>\n"
"        <input file stock=\"gtk-remove\"></input>\n"
"        <action type=\"removeselected\">FILE_LIST</action>\n"
"        <action type=\"refresh\">COMMAND</action>\n"
"      </button>\n"
"    </hbox>\n"
"  </frame>\n");

}

static gboolean
print_program_for_no_producer(
		FILE        *outfile, 
		Terminal    *t)
{
	GList *list, *li;
	
	g_assert(outfile != NULL && t != NULL);
	
#ifdef DEBUG
	g_message("%s(): Starting.", __func__);
#endif
	
	g_fprintf(outfile,
"#\n"
"# This function will assemble a command line using the state\n"
"# of the widgets of the graphical user interface.\n"
"#\n"
"function command_line() {\n"
"  echo -n $COMMAND_LIST\n"
"}\n\n");


	/* The start of the MAIN_DIALOG. */
	g_fprintf(outfile, 
"export MAIN_DIALOG='\n"
"<window title=\"Cached Programs\" window_position=\"2\">"
"  <vbox>\n"
);
	/* The description and the command entry. */	
	fprintf(outfile,
_(
"  <frame Description>\n"
"    <hbox>\n"
"      <pixmap>\n"
"        <input file stock=\"gtk-info\"></input>\n"
"      </pixmap>\n"
"      <text>\n"
"        <label>\n"
"\"This dialog window is created using the cached informations Cuiterm collected "
"analyzing the communication. The cache can be expanded by using commands giving "
"informaton about programs (e.g. the man program).\"\n"
"        </label>\n"
"      </text>\n"
"    </hbox>\n"
"    <entry>\n"
"      <variable>COMMAND</variable>\n"
"      <visible>disabled</visible>\n"
"      <input>command_line</input>\n"
"    </entry>\n"
"  </frame>\n"));
	
	/* The start of command list */
	g_fprintf(outfile,
_(
"  <frame Cached Programs>\n"
"    <tree rules_hint=\"true\">\n"
"      <width>300</width><height>200</height>"
"      <variable>COMMAND_LIST</variable>\n"
"      <action signal=\"cursor_changed\" type=\"refresh\">COMMAND</action>\n"
"      <action type=\"exit\">OK</action>\n"
"      <label>Program|Description</label>\n"));
	/*
	 * Inserting the program names and descriptions into the table.
	 */
	list = program_get_program_list();
	li = list;
	while (li != NULL) {
		g_fprintf(outfile,
"       <item>%s|%s</item>\n", 
		li->data,
		program_get_description(li->data));
next_item:
		li = li->next;
	}
	
	/* The end of command list */
	g_fprintf(outfile,
_(
"    </tree>\n"
"  </frame>\n"));
	
	
	/* The buttons. */
	g_fprintf(outfile,
"  <hbox>\n"
"    <button ok></button>\n"
"    <button cancel></button>\n"
"  </hbox>\n");

	/* The end of the MAIN_DIALOG. */
	g_fprintf(outfile, 
"  </vbox>\n"
"</window>\n"
"'\n");

	return TRUE;
}


static gboolean
print_program_for_single_producer(
		FILE        *outfile, 
		Terminal    *t,
		const gchar *producer)
{
	const gchar *long_description;
	const gchar *short_description;
	
	g_assert(outfile != NULL && producer != NULL);
#ifdef DEBUG
	g_message("%s(): Starting.", __func__);
#endif
	
	print_command_line_function(outfile, producer);
	
	long_description = program_get_long_description(producer);
	short_description = program_get_description(producer);
	if (g_utf8_strlen(long_description, -1) == 0)
		long_description = short_description;
	
	/* The start of the MAIN_DIALOG. */
	g_fprintf(outfile, 
"export MAIN_DIALOG='\n"
"<window title=\"%s\" window_position=\"2\">\n"
"  <vbox>\n", short_description);
	
	/* The description and the command entry. */	
	fprintf(outfile,
_(
"  <frame Description>\n"
//"    <hbox>\n"
//"      <pixmap>\n"
//"        <input file stock=\"gtk-info\"></input>\n"
//"      </pixmap>\n"
"      <text>\n"
"        <label>\n"
"\"%s\"\n"
"        </label>\n"
"      </text>\n"
//"    </hbox>\n"
"    <entry>\n"
"      <variable>COMMAND</variable>\n"
"      <visible>disabled</visible>\n"
"      <input>command_line</input>\n"
"    </entry>\n"
"  </frame>\n"), long_description);
	/*
	 * The important options.
	 */
	const gchar *category;
	gboolean     iterate, optional;
	gint   n;
	category = program_get_argument_category(producer, 1);
	if (g_ascii_strcasecmp(category, "any-option") == 0)
		n = print_important_options(outfile, t, producer);
	
	print_prompt_options(outfile, t, producer);
	print_standard_options(outfile, t, producer);
	
	category = program_get_argument_category(producer, 2);
	optional = program_get_argument_optional(producer, 2);
	iterate = program_get_argument_iterate(producer, 2);
	if (g_ascii_strcasecmp(category, "file") == 0)
		print_file_selection(outfile, producer, iterate, optional);
	
	/* The buttons at the bottom: OK, Cancel and Help */
	g_fprintf(outfile,
"  <hbox>\n"
"    <button ok></button>\n"
"    <button cancel></button>\n"
"  </hbox>\n"
"", producer
);
	/* The end of the MAIN_DIALOG. */
	g_fprintf(outfile, 
"  </vbox>"
"</window>\n"
"'\n");
	return TRUE;
}

/**
 *
 * @returns: A filename or NULL in the case of some problems.
 */
static gchar *
emit_program(Terminal *t)
{
	GtkWidget *dialog;
	gchar     *filename;
	FILE      *outfile;
	gchar     *producer;      /* The producer is the first program name. */
	
#ifdef DEBUG
	g_message("%s(): Starting.", __func__);
#endif
	g_assert(t != NULL);
	/*
	 * Initialization.
	 */
	if (t->parse_tree != NULL)
		producer = parse_tree_get_producer_program(t->parse_tree);
	else
		producer = NULL;
	/*
	 * The first part of the generated file.
	 */
	filename = create_temporary_program_name();
	outfile = open_temporary_program_for_write(filename);
	if (outfile == NULL) {
		g_free(filename);
		return NULL;
	}
	
	if (!print_first_line(outfile, filename)) {
		g_free(filename);
		return NULL;
	}
	/*
	 * Now the question is if we have a command partially typed or we
	 * have to offer a list of programs.
	 */
	if (producer != NULL) {
		/*
		 * If the producer is not in the cache, we apologize and return.
		 */
		if (!program_info_in_cache(producer)) {
			 dialog = gtk_message_dialog_new (NULL,
	                                  GTK_DIALOG_DESTROY_WITH_PARENT,
	                                  GTK_MESSAGE_ERROR,
	                                  GTK_BUTTONS_CLOSE,
					_(
						"Documentation for the '%s' program is not in the cache. "
						"Please use some command (e.g. 'man %s') to cache the information "
						"needed to build the dialog box."), producer, producer);
	 		gtk_dialog_run (GTK_DIALOG (dialog));
	 		gtk_widget_destroy (dialog);
			fclose(outfile);
			g_remove(filename);
			g_free(filename);
			return NULL;
		}
	
		if (!print_program_for_single_producer(outfile, t, producer)) {
			fclose(outfile);
			g_remove(outfile);
			g_free(filename);
			return NULL;
		}
	} else {
		if (!print_program_for_no_producer(outfile, t)) {
			fclose(outfile);
			g_remove(outfile);
			g_free(filename);
			return NULL;
		}	
	}
	
	fclose(outfile);
	outfile = NULL;
	g_chmod(filename, S_IRUSR | S_IWUSR | S_IXUSR);
	return filename;
}

/**
 * gui_completion_activate:
 * @t: a #Terminal.
 *
 * Called when the user asks for a GUI of the given command line by pressing
 * the Ctrl+G key combination or selectin this feature from the menu or the 
 * toolbar.
 */
gpointer 
gui_completion_activate(Terminal *t)
{
	gchar    *filename;
	FILE     *pipe;
	gchar     line[512];
	gchar    *end;
	gboolean  ok_pressed;
	gchar    *new_command = NULL;
	
	g_assert(t != NULL);

	filename = emit_program(t);
	if (filename == NULL)
	//	g_thread_exit(NULL);
		return NULL;

#ifdef DEBUG
	g_message("%s(): Starting on command: '%s'", __func__, 
			t->current_command);
#endif
	
	pipe = popen(filename, "r");
	ok_pressed = FALSE;
	while (fgets(line, 511, pipe) != NULL) {
		//g_message("line: '%s'", line);
		if (strncmp(line, "COMMAND=\"", 9) == 0) {
			new_command = g_strdup(&line[9]);
			end = strchr(new_command, '"');
			if(end != NULL)
				*end = '\0';
		}
		if (g_ascii_strcasecmp(line, "EXIT=\"OK\"\n") == 0)
			ok_pressed = TRUE;
	}

	if (ok_pressed && new_command != NULL) 
		terminal_send_command(t, new_command, FALSE);

finalize:
	fclose(pipe);
	g_free(filename);
	//g_thread_exit(NULL);
	return NULL;
}
