/*
 * Copyright (C) 2004 Laszlo Pere
 * $Id:$
 *
 * This is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Library General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Log:$
 */
#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#define _GNU_SOURCE
#include <gnome.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <stdio.h>

#include <bonobo.h>
#include <gnome.h>
#include <gconf/gconf-client.h>

#include <libxml/xmlmemory.h>
#include <libxml/parser.h>
#include <libxml/xpathInternals.h>
#include <libxml/tree.h>

#include "support.h"
#include "config.h"
#include "history.h"

#if DEBUG_ALL
#   define DEBUG
#endif


static GList *history_list = NULL;


/*****************************************************************************
 * Static function declarations.                                             *
 *                                                                           *
 *****************************************************************************/
static gboolean is_option_ignored(const gchar *name, const gchar *option, const gchar *description);
static gboolean is_option_prompt(const gchar *name, const gchar *option, const gchar *description);
static gboolean is_option_indirect(const gchar *name, const gchar *option, const gchar *description);
static void add_description_to_option(xmlXPathContextPtr xpath_ctx, 
		const gchar *name, 
		const gchar *option, 
		const gchar *description);


/*****************************************************************************
 * String manipulation functions.                                            *
 *****************************************************************************/
gchar *
canonicalize_option(gchar *option)
{
	gchar *copy = g_strdup(option);
	
	g_strstrip(copy);
	if (copy[0] != '\0' && copy[1] != '\0')
		copy[2] = '\0';
	
	return copy;
}

gchar *
canonicalize_long_option(gchar *option)
{
	gchar *copy = g_strdup(option);
	gchar *end;
	
	g_strstrip(copy);
	end = g_utf8_strchr(copy, -1, '=');
	if (end != NULL)
		*end = '\0';
	end = g_utf8_strchr(copy, -1, '[');
	if (end != NULL)
		*end = '\0';
	
	return copy;
}

gchar *
canonicalize_sentence(gchar *option)
{
	gint state = 0;
	gunichar c, last = ' ';
	GString *copy = g_string_new("");
	gint n = 0;
	/*
		state:
			0: normal text
			1: a space came, we eliminate all spaces
			2: a '-' came, we eliminate it when followed by ' \n'
			3: a '-\n' came, now we eat spaces.
	 */
	c = g_utf8_get_char(option);
	option = g_utf8_next_char(option);
	while (c != '\0') {
		switch (state) {
			case 0:	
				if (c == ' ')
					state = 1;
				if (c == '-') {
					state = 2;
					goto get_next_char;
				}
				if (c == '\n')
					goto get_next_char;
				break;
			case 1:
				if (c == ' ')
					goto get_next_char;
				if (c != ' ')
					state = 0;
				break;
			case 2:
				if (c == '\n') {
					state = 3;
					goto get_next_char;
				}else{
					g_string_append_unichar(copy, '-');
					state = 0;
				}
				break;
			case 3:
				if (c == ' ')
					goto get_next_char;
				else
					state = 0;
		}
		if (n == 0)
			c = g_unichar_toupper(c);
		last = c;
		g_string_append_unichar(copy, c);
		n++;
	get_next_char:
		c = g_utf8_get_char(option);
		option = g_utf8_next_char(option);
	}
	
	if (last != '.')
		g_string_append_unichar(copy, '.');

	return g_string_free(copy, FALSE);
}

gchar *
canonicalize_paragraph(gchar *option)
{
	gint state = 0;
	gunichar c, last = ' ';
	GString *copy = g_string_new("");
	gint n = 0;
	/*
		state:
			0: normal text
			1: a space came, we eliminate all spaces
			2: a '-' came, we eliminate it when followed by ' \n'
			3: a '-\n' came, now we eat spaces.
	 */
	c = g_utf8_get_char(option);
	option = g_utf8_next_char(option);
	while (c != '\0') {
		switch (state) {
			case 0:	
				if (c == ' ')
					state = 1;
				if (c == '-') {
					state = 2;
					goto get_next_char;
				}
				if (c == '\n')
					goto get_next_char;
				break;
			case 1:
				if (c == ' ')
					goto get_next_char;
				if (c != ' ')
					state = 0;
				break;
			case 2:
				if (c == '\n') {
					state = 3;
					goto get_next_char;
				}else{
					g_string_append_unichar(copy, '-');
					state = 0;
				}
				break;
			case 3:
				if (c == ' ')
					goto get_next_char;
				else
					state = 0;
		}
		if (n == 0)
			c = g_unichar_toupper(c);
		last = c;
		g_string_append_unichar(copy, c);
		n++;
	get_next_char:
		c = g_utf8_get_char(option);
		option = g_utf8_next_char(option);
	}
	
	if (last != '.')
		g_string_append_unichar(copy, '.');

	return g_string_free(copy, FALSE);
}

/*
 * Returns FALSE if the parameter is not an application name supposed to get
 * filenames to edit or view.
 */
static gboolean
application(gchar *program) {
	gint n;
	gchar *not_application[] = {
		"cp", "mv", "ls", "cd", "rm", "getfacl", "setfacl", "scp",
		"chmod", "chown", "du", "locate", "mkdir", "grep", "tree", 
		"stat", "diff", "make", "man", ".", "pushd",  

		"ssh", 
		NULL
	};
	
	for (n = 0; not_application[n] != NULL; ++n)
		if (strcmp(not_application[n], program) == 0)
			return FALSE;
	
	if (strchr(program, '=') != NULL)
		return FALSE;
	
	return TRUE;
}

/*
 * Returns TRUE if the given command name is an internal BASH command.
 */
static gboolean
bash_internal(gchar *program) {
	gint n;
	gchar *bash_internals[] = {
		"source",    "alias",     "bg",        "bind",
		"break",     "builtin",   "cd",        "caller", 
		"command",   "compgen",   "complete",  "continue",
		"declare",   "typeset",   "dirs",      "disown",
		"echo",      "enable",    "eval",      "exec",
		"exit",      "export",    "fc",        "fg",
		"getopts",   "hash",      "help",      "history",
		"jobs",      "kill",      "let",       "local",
		"logout",    "popd",      "printf",    "pushd",
		"pwd",       "read",      "readonly",  "return",
		"set",       "shift",     "shopt",     "suspend",
		"test",      "times",     "trap",      "type",
		"ulimit",    "umask",     "unalias",   "unset",
		"wait",
		NULL
	};
	
	for (n = 0; bash_internals[n] != NULL; ++n)
		if (strcmp(bash_internals[n], program) == 0)
			return TRUE;
	
	return FALSE;
}

gchar *
handle_text_editors(gchar *appname, ApplicationOpcode opcode) {
	gint n, higher;
	static app_freq text_editors[] = {
		{"vi",         0},
		{"vim",        0},
		{"gvim",       0},
		{"emacs",      0},
		{"xemacs",     0}, 
		{"gedit",      0},
		{NULL,         0}
	};

	switch (opcode) {
		case OpTryAdd:
			for (n = 0; text_editors[n].appname != NULL; ++n)
				if (strcmp(text_editors[n].appname, appname) == 0){
					text_editors[n].freq++;
					return NULL;
				}
			break;
		case OpSearch:
			higher = 0;
			for (n = 1; text_editors[n].appname != NULL; ++n)
				if (text_editors[n].freq > text_editors[higher].freq)
					higher = n;
			return text_editors[higher].appname;
			break;
		case OpReset:
			for (n = 1; text_editors[n].appname != NULL; ++n)
				text_editors[n].freq = 0;
			break;
		default:
			g_warning("%s(): Unknown opcode: %d", __func__, opcode);
	}
	
	return NULL;
}

static void
history_line_red(
		file_read_progress *file_progress,
		gchar *line)
{
	gint       argc;
	gchar    **argv;
	GError    *error = NULL;
	gchar     *basename;
	gint       n;
	gchar     *colored_command, tmp;	
	
#ifdef DEBUG
	g_message("%s(): line: '%s'", __func__, line);
#endif
	/*
	 *
	 */
	if (!g_shell_parse_argv(line, &argc, &argv, &error)) {
		g_assert(error != NULL);
		g_warning("%s(): %s", error->message);
		return;
	}
	/*
	 * Incrementing program option counters.
	 */
	n = 1;
	while (n < argc) {
		if (*argv[n] == '-') 
			program_increment_option_history_counter(argv[0], argv[n]);
		++n;
	}
#if 0
	/*
	 * We store the commands so we can use it to visualize the history on
	 * the screen.
	 */
	colored_command = g_strdup_printf(
			"<span color=\"dark green\">%s</span>", argv[0]); 
	n = 1;
	while (n < argc) {
		tmp = colored_command;
		g_message("----------------%s(): '%s'", __func__, argv[n]);
		if (*argv[n] == '-') {
			colored_command = g_strconcat(
					colored_command,
					" <span color=\"brown\">",
					argv[n],
					"</span>",
					NULL);
		}else if (*argv[n] == '<' || *argv[n] == '>') {
			colored_command = g_strconcat(
					colored_command,
					" <span color=\"red\">\\",
					argv[n],
					"</span>",
					NULL);
		} else {
			colored_command = g_strconcat(
					colored_command,
					" ", 
					argv[n],
					NULL);
		}

		++n;
		g_free(tmp);
	}
	g_message("->'%s'", colored_command);
#endif
	colored_command = g_strdup(line);	
	colored_command[strlen(colored_command) - 1] = '\0';
	history_list = g_list_append(history_list, colored_command);
	
	handle_text_editors(argv[0], OpTryAdd);
	
	//g_message("%s(): argc=%d argv[0]: '%s'", __func__, argc, argv[0]);
	if (!application(argv[0]))
		goto finalize;
	
	for (n = 1; n < argc; ++n) {
		if (argv[n][0] != '-'){
			basename = g_path_get_basename(argv[n]);
			g_free(basename);
		}
	}
finalize:
	g_strfreev(argv);
}

static gboolean 
_read_from_history(GIOChannel *source, 
		GIOCondition condition, 
		gpointer data)
{
	gchar     *line;
	gint       length;
	GError    *error = NULL;
	GIOStatus  status;
	static gint n = 0;
	file_read_progress *file_progress = (file_read_progress *)data;
#ifdef DEBUG
	g_message("%s(): Start.", __func__);
#endif
	g_assert(file_progress != NULL);
	
	/*
	 * If there is no more data we close the channel.
	 */
	if (condition == G_IO_HUP){
		g_message("end of file");
		//g_io_channel_shutdown(source, TRUE, NULL);
		return FALSE;
	}
	/*
	 * Reading one line from the history file.
	 */
	status = g_io_channel_read_line(source, &line, &length, NULL, &error);
	if (status != G_IO_STATUS_NORMAL || line == NULL) {
		if (error != NULL)
			g_warning("%s(): %s", __func__, error->message);
		gnome_appbar_set_progress_percentage(
				GNOME_APPBAR(file_progress->appbar), 0.0);
		gnome_appbar_pop(GNOME_APPBAR(file_progress->appbar));
#ifdef DEBUG
		g_message("Text Editor: '%s'", handle_text_editors(NULL, OpSearch));
#endif
		return FALSE;
	}

	file_progress->red += length;
	n++;
	if (n >= 50) {
		gnome_appbar_set_progress_percentage(
				GNOME_APPBAR(file_progress->appbar),
				(gfloat)file_progress->red / 
				  file_progress->size);
		n = 0;
	}
	if (length > 1)
		history_line_red(file_progress, line);

	g_free(line);
	
	return TRUE;
}

/**
 * history_read_bash_history:
 * widget: a #GtkWidget to find the message bar by the "appbar1" name.
 *
 * Reads and processes the BASH history file. This function must be executed as
 * a separate thread.
 */
gpointer
history_read_bash_history(GtkWidget *widget)
{
	GIOChannel *history_channel;
	file_read_progress	*file_progress;
	GtkWidget   *appbar;
	gchar       *filename;
	gchar       *message;
	struct stat  stat_buf;
	gchar       line[255];
	gint        red = 0;
#ifdef DEBUG
	g_message("%s(): Start.", __func__);
#endif
	/*
	 * This time we reset the counters we use to find out which command
	 * options are used frequently.
	 */
	program_reset_all_history_counter();
	file_progress = g_new(file_read_progress, 1);

	/* Filename and filesize. */
	filename = g_build_path(G_DIR_SEPARATOR_S, 
			g_getenv("HOME"), ".bash_history", NULL);
	if (g_stat(filename, &stat_buf) != 0){
		g_warning("Unable to stat file '%s'", filename);
		g_free(filename);
		return NULL;
	}
	file_progress->size = stat_buf.st_size;
	file_progress->red = 0;
	
	/* Widgets: message and progress */
	file_progress->appbar = lookup_widget(widget, "appbar1");
	message = g_strdup_printf(_("Reading file '%s'"), filename);
  	gnome_appbar_push(GNOME_APPBAR(file_progress->appbar), message);
	g_free(message);
	
	/* Opening file and creating channel. */
	file_progress->ifile = g_open(filename, O_RDONLY);
	if (file_progress->ifile < 0) {
		g_warning("Unable to open file '%s'", filename);
		return NULL;
	}
	history_channel = g_io_channel_unix_new(file_progress->ifile);
	g_io_add_watch_full(history_channel, 
			G_PRIORITY_DEFAULT_IDLE, 
			G_IO_IN | G_IO_HUP, 
			_read_from_history, (gpointer) file_progress, NULL);
	g_thread_exit(NULL);
	return NULL;
}

/*****************************************************************************
 * Type checking and correcting with the man cache.                          *
 * I'm changing the program to handle man cache with an xml document.        *
 *****************************************************************************/
//static GTree *programs = NULL;

static xmlDocPtr document = NULL;
/*
xmlNodePtr	xmlNewCDataBlock	(xmlDocPtr doc, 
					 const xmlChar * content, 
					 int len)

xmlNodePtr	xmlNewChild		(xmlNodePtr parent, 
					 xmlNsPtr ns, 
					 const xmlChar * name, 
					 const xmlChar * content)

xmlNodePtr	xmlNewDocFragment	(xmlDocPtr doc)

xmlNodePtr	xmlNewDocRawNode	(xmlDocPtr doc, 
					 xmlNsPtr ns, 
					 const xmlChar * name, 
					 const xmlChar * content)

xmlNodePtr	xmlNewNode		(xmlNsPtr ns, 
					 const xmlChar * name)

xmlAttrPtr	xmlNewProp		(xmlNodePtr node, 
					 const xmlChar * name, 
					 const xmlChar * value)

xmlNodePtr	xmlNewText		(const xmlChar * content)

xmlNodePtr	xmlReplaceNode		(xmlNodePtr old, 
					 xmlNodePtr cur)

xmlNodePtr	xmlAddChild		(xmlNodePtr parent, 
					 xmlNodePtr cur)

void	xmlUnlinkNode			(xmlNodePtr cur)

void	xmlFreeNode			(xmlNodePtr cur)

 */
gboolean
xml_document_new(GtkWidget *app,
		xmlDocPtr *document)
{
	GtkWidget *dialog;
	gchar tmp[128];
	gchar *tmp_p;
	xmlNodePtr tree, subtree, subsubtree;

	xmlKeepBlanksDefault(0);

	// Document and root node
	*document = xmlNewDoc("1.0");
	(*document)->children = xmlNewDocNode(*document, 
			NULL, 
			"man-cache", 
			NULL);
	xmlSetProp((*document)->children, "version", "0.1");
	
	tree = xmlNewChild((*document)->children, NULL, "creator", NULL);
	subtree = xmlNewChild(tree, NULL, "app", NULL);
	subsubtree = xmlNewChild(subtree, NULL, "name", PACKAGE);
	subsubtree = xmlNewChild(subtree, NULL, "version", VERSION);

	tree = xmlNewChild((*document)->children, NULL, "programs", NULL);
	return TRUE;
}

gboolean
xml_document_save(GtkWidget *app, 
		xmlDocPtr *document,
		gchar *filename)
{
	g_assert(document != NULL && filename != NULL);
#ifdef DEBUG
	g_message("%s(): start on file: '%s'", __func__, filename);
#endif
	xmlKeepBlanksDefault(0);
        if (!xmlSaveFormatFileEnc(filename, *document, "UTF-8", 1)){
		g_error("%s(): error saving document: %m", __func__);
	}
	return TRUE;
}

gboolean
xml_document_load(GtkWidget *app, 
		xmlDocPtr *document, 
		gchar *filename)
{
	GtkWidget *dialog;

	g_assert(filename != NULL && document != NULL);
#ifdef DEBUG
	g_message("%s(): Start.", __func__);
#endif
	xmlKeepBlanksDefault(0);
	*document = xmlParseFile(filename);

	if (*document == NULL) {
		return xml_document_new(app, document);
	}

	return TRUE;
}

static void 
man_cache_finalize(void)
{
	gchar *filename;
	gchar *backup_filename;
#ifdef DEBUG	
	g_message("%s(): Start.", __func__);
#endif
	
	filename = g_build_path(G_DIR_SEPARATOR_S,
			g_getenv("HOME"), ".cuiterm_man_cache", NULL);
	backup_filename = g_build_path(G_DIR_SEPARATOR_S,
			g_getenv("HOME"), ".cuiterm_man_cache.bak", NULL);
	
	g_rename(filename, backup_filename);
	
	xml_document_save(NULL, &document, filename);
	
finalize:
	g_remove(backup_filename);
	g_free(filename);
	g_free(backup_filename);
}

static inline void 
man_cache_initialize(void)
{
	static gboolean initialized = FALSE;
	gchar *filename;
#ifdef BEBUG
	g_message("%s(): Start.", __func__);
#endif
	if (initialized)
		return;
	initialized = TRUE;

	filename = g_build_path(G_DIR_SEPARATOR_S,
			g_getenv("HOME"), ".cuiterm_man_cache", NULL);
	
	xml_document_load(NULL, &document, filename);	
	/*
	 * This function will save the man cache at exit.
	 */
	g_atexit(man_cache_finalize);
}

/****************************************************************************
 * Utility functions for XML manipulation.                                  *
 *                                                                          *
 ****************************************************************************/
static inline gint 
xpath_to_gint(const gchar       *command,
	     xmlXPathContextPtr xpath_ctx)
{
	gint n;
	xmlXPathObjectPtr xpath_obj;
	
	xpath_obj = xmlXPathEvalExpression(command, xpath_ctx);
	n = (gint) xmlXPathCastToNumber(xpath_obj);
	xmlXPathFreeObject(xpath_obj);
	return n;
}

/**
 * xpath_get_node:
 *
 * Tryes to find a node, if it not exists will create it when create is TRUE.
 */
static inline xmlNodePtr
xpath_get_node(const gchar         *nodepath,
	      xmlXPathContextPtr   xpath_ctx,
	      gboolean             create)
{
	xmlXPathObjectPtr xpath_obj;
	xmlNodePtr        nodes_node, new_node;
	gchar             *nodepath_parent, *tmp;
#ifdef DEBUG
	g_message("%s(): searching node: '%s'", __func__, nodepath);
#endif
	// Searching the node.
	xpath_obj = xmlXPathEvalExpression(nodepath, xpath_ctx);
	// If we found, it is all right, we return it.
	if (xmlXPathCastToBoolean(xpath_obj)/*xpath_obj != NULL && xpath_obj->nodesetval != NULL*/) {
		nodes_node = xpath_obj->nodesetval->nodeTab[0];
		xmlXPathFreeObject(xpath_obj);
		return nodes_node;
	}
	// If we don't have to create it, we return NULL.
	if (!create)
		return NULL;
	// If we want to create, we get the parent and create it.
	nodepath_parent = g_strdup(nodepath);
	tmp = g_utf8_strrchr(nodepath_parent, -1, '/');
	if (tmp == NULL) {
		g_warning("%s: The root node '%s' has not been found.", __func__, nodepath);
		return NULL;
	}
	*tmp = '\0';
	nodes_node = xpath_get_node(nodepath_parent, xpath_ctx, TRUE);
	if (nodes_node == NULL) {
		g_free(nodepath_parent);
		return NULL;
	}
	new_node = xmlNewChild(nodes_node, NULL, tmp + 1, NULL);
	g_free(nodepath_parent);
	return new_node;
}

/**
 * xpath_create_new_node:
 *
 * Searches the parent and creates a new node. This function always 
 * returns with a new node, except if the root node could not be found.
 */
static inline xmlNodePtr
xpath_create_new_node(const gchar         *nodepath,
	             xmlXPathContextPtr   xpath_ctx)
{
	xmlNodePtr   nodes_node, new_node;
	gchar       *nodepath_parent;
	gchar       *tmp;
	
	/* Let's find (or maybe create) the parent. */
	nodepath_parent = g_strdup(nodepath);
	tmp = g_utf8_strrchr(nodepath_parent, -1, '/');
	if (tmp == NULL) {
		g_warning("%s: Tryed to create a root node '%s'.", __func__, nodepath);
		g_free(nodepath_parent);
		return NULL;
	}
	*tmp = '\0';
	nodes_node = xpath_get_node(nodepath_parent, xpath_ctx, TRUE);
	if (nodes_node == NULL) {
		g_warning("%s: Could not found parent '%s'.", __func__, nodepath_parent);
		g_free(nodepath_parent);
		return NULL;
	}
	/* We create a new node here. */
	new_node = xmlNewChild(nodes_node, NULL, tmp + 1, NULL);
	g_free(nodepath_parent);
	return new_node;
}

/**
 *
 * Creates a description node for the given node if the node hasn1t got a write
 * protected description node.
 */
static inline gboolean
xml_create_description_node(xmlNodePtr node,
		const gchar           *description,
                xmlXPathContextPtr     xpath_ctx)
{
	xmlNodePtr  description_node;
	gchar      *protection;
	gboolean    protected;
	gchar      *tmp;

	g_assert(node != NULL);
	/*
	 * First we try to find an existing description node by walking through the childrens.
	 */
	description_node = node->xmlChildrenNode;
	while (description_node != NULL) {
		if (strcmp(description_node->name, "description") == 0)
			break;
		description_node = description_node->next;
	}
	/*
	 * If not found we create one.
	 */
	if (description_node == NULL) {
		description_node = xmlNewChild(node, NULL, "description", description);
		tmp = g_strdup_printf("%ld", g_utf8_strlen(description, -1));
		xmlSetProp(description_node, "length", tmp);
		xmlSetProp(description_node, "protection", "read-write");
		return TRUE;
	}
	/*
	 * We check the protection and if not set we modify the description.
	 */
	protection = xmlGetProp(description_node, "protection");
	protected = protection != NULL && strcmp(protection, "read-only") == 0;
	g_free(protection);
	if (!protected) {
		xmlNodeSetContent(description_node, g_strdup(description));
		tmp = g_strdup_printf("%ld", g_utf8_strlen(description, -1));
		xmlSetProp(description_node, "length", tmp);
		return TRUE;
	}
	/*
	 * The description is protected.
	 */
	return FALSE;
}
		
/**
 *
 * Creates a description node for the given node if the node hasn't got a write
 * protected description node.
 */
static inline gboolean
xml_create_long_description_node(xmlNodePtr node,
		const gchar           *description,
                xmlXPathContextPtr     xpath_ctx)
{
	xmlNodePtr  description_node;
	gchar      *protection;
	gboolean    protected;
	gchar      *tmp;

	g_assert(node != NULL);
	/*
	 * First we try to find an existing description node by walking through the childrens.
	 */
	description_node = node->xmlChildrenNode;
	while (description_node != NULL) {
		if (strcmp(description_node->name, "long-description") == 0)
			break;
		description_node = description_node->next;
	}
	/*
	 * If not found we create one.
	 */
	if (description_node == NULL) {
		description_node = xmlNewChild(node, NULL, "long-description", description);
		tmp = g_strdup_printf("%ld", g_utf8_strlen(description, -1));
		xmlSetProp(description_node, "length", tmp);
		xmlSetProp(description_node, "protection", "read-write");
		return TRUE;
	}
	/*
	 * We check the protection and if not set we modify the description.
	 */
	protection = xmlGetProp(description_node, "protection");
	protected = protection != NULL && strcmp(protection, "read-only") == 0;
	g_free(protection);
	if (!protected) {
		xmlNodeSetContent(description_node, g_strdup(description));
		tmp = g_strdup_printf("%ld", g_utf8_strlen(description, -1));
		xmlSetProp(description_node, "length", tmp);
		return TRUE;
	}
	/*
	 * The description is protected.
	 */
	return FALSE;
}

/******************************************************************************
 * Helper functions to handle program nodes.                                  *
 *                                                                            *
 ******************************************************************************/

/*
 * Search the nodes for the program and returns it if it is not write protected.
 * If the node doesn't exists this function will create it.
 */
static inline xmlNodePtr
_program_get_program_node_for_write(
		xmlXPathContextPtr  xpath_ctx,
		const gchar        *name)
{
	xmlNodePtr  program_node;
	gchar      *path = g_strdup_printf("/man-cache/programs/program[@name='%s']", name);
	gchar      *protection;
	gboolean    protected;
	
	/* Seaarching. */
	program_node = xpath_get_node(path, xpath_ctx, FALSE);
	g_free(path);
	/* If not exists we create it with default protection. */
	if (program_node == NULL) {
		program_node = xpath_create_new_node("/man-cache/programs/program", xpath_ctx);
		xmlSetProp(program_node, "name", name);
		xmlSetProp(program_node, "protection", "read-write");
		return program_node;
	}
	/* If there is no protection we return it. */
	protection = xmlGetProp(program_node, "protection");
	if (protection == NULL)
		return program_node;
	
	protected = strcmp(protection, "read-only") == 0;
	xmlFree(protection);
	if (protected)
		return NULL;
	else
		return program_node;
}

/*
 * 
 */
static inline xmlNodePtr
_program_get_option_node_for_write(
		xmlXPathContextPtr  xpath_ctx,
		const gchar        *name,
		const gchar        *option)
{
	xmlNodePtr     program_node;
	xmlNodePtr     option_node;
	gchar         *option_name;
	gchar         *protection;
	gboolean       protected;
	/*
	 * If we can't open the program node for write we abort.
	 */
	program_node = _program_get_program_node_for_write(xpath_ctx, name);
	if (program_node == NULL)
		return NULL;
	/*
	 *
	 */
	option_node = program_node->xmlChildrenNode;
	while (option_node != NULL) {
		if (strcmp(option_node->name, "option") != 0)
			goto next_child;
		option_name = xmlGetProp(option_node, "option-name");
		/* If this is not the option we want. */
		if (option_name == NULL)
			goto next_child;
		if (strcmp(option_name, option) != 0) {
			xmlFree(option_name);
			goto next_child;
		}
		xmlFree(option_name);
		
		/* Checking the write protection. */
		protection = xmlGetProp(option_node, "protection");
		if (protection == NULL)
			return program_node;
		protected = strcmp(protection, "read-only");
		xmlFree(protection);
		if (protected)
			return NULL;
		else
			return option_node;

next_child:
		option_node = option_node->next;
	}
	/*
	 * Now we could open the program for writing, but we
	 * could not find the option node, so we can create it.
	 */
	option_node = xmlNewChild(program_node, NULL, "option", NULL);
	xmlSetProp(option_node, "option-name", xmlEncodeEntitiesReentrant(document, option));
	return option_node;
}

/*
 * 
 */
static inline xmlNodePtr
_program_get_long_option_node_for_write(
		xmlXPathContextPtr  xpath_ctx,
		const gchar        *name,
		const gchar        *option)
{
	xmlNodePtr     program_node;
	xmlNodePtr     option_node;
	gchar         *option_name;
	gchar         *protection;
	gboolean       protected;
	/*
	 * If we can't open the program node for write we abort.
	 */
	program_node = _program_get_program_node_for_write(xpath_ctx, name);
	if (program_node == NULL)
		return NULL;
	/*
	 *
	 */
	option_node = program_node->xmlChildrenNode;
	while (option_node != NULL) {
		if (strcmp(option_node->name, "option") != 0)
			goto next_child;
		option_name = xmlGetProp(option_node, "long-option-name");
		/* If this is not the option we want. */
		if (option_name == NULL)
			goto next_child;
		if (strcmp(option_name, option) != 0) {
			xmlFree(option_name);
			goto next_child;
		}
		xmlFree(option_name);
		
		/* Checking the write protection. */
		protection = xmlGetProp(option_node, "protection");
		if (protection == NULL)
			return program_node;
		protected = strcmp(protection, "read-only");
		xmlFree(protection);
		if (protected)
			return NULL;
		else
			return option_node;

next_child:
		option_node = option_node->next;
	}
	/*
	 * Now we could open the program for writing, but we
	 * could not find the option node, so we can create it.
	 */
	option_node = xmlNewChild(program_node, NULL, "option", NULL);
	xmlSetProp(option_node, "long-option-name", xmlEncodeEntitiesReentrant(document, option));
	return option_node;
}
/**
 * synopsis_create_node:
 * @name: the name of the node
 * @returns: a new node
 *
 * Creates a single node which later must be added to the document.
 * See #synopsis_add_node_to_program how to add a new synopsis node
 * to the program.
 */
xmlNodePtr 
synopsis_create_node(const gchar *name) 
{
	xmlNodePtr   node;
	gchar       *arg_description;
	gchar       *category;
	
	g_assert(name != NULL);
	
	node = xmlNewNode(NULL, "argument");
#ifdef DEBUG
	xmlSetProp(node, "creator", __func__);
#endif
	
	arg_description = g_utf8_strdown(name, -1);
	
	category = "unknown";
	if (g_str_equal(arg_description, "option") ||
			g_str_equal(arg_description, "options"))
		category = "any-option";
	if (g_strrstr(arg_description, "file") != NULL)
		category = "file";
	
	xmlSetProp(node, "arg-description", arg_description);
	xmlSetProp(node, "category", category);
	
	return node;
}

/**
 * synopsis_create_node_with_child:
 * @name: the name of the new node
 * @child: the child node of the new node
 * @returns: a new node
 *
 * Creates a node which later must be added to the document and adds
 * a child node to it.
 * See #synopsis_add_node_to_program how to add a new synopsis node
 * to the program.
 */
xmlNodePtr 
synopsis_create_node_with_child(const gchar *name, 
		xmlNodePtr child) 
{
	xmlNodePtr	node;
	node = xmlNewNode(NULL, name);
#ifdef DEBUG
	xmlSetProp(node, "creator", __func__);
#endif
	xmlAddChild(node, child);
	return node;
}

xmlNodePtr
synopsis_create_operator_node(gchar operator,
		xmlNodePtr arg1,
		xmlNodePtr arg2)
{
	gchar op[2] = "?";
	xmlNodePtr	node;
	
	op[0] = operator;
	node = xmlNewNode(NULL, xmlStrdup("binop"));
	xmlSetProp(node, "op", op);
#ifdef DEBUG
	xmlSetProp(node, "creator", __func__);
#endif
	xmlAddChild(node, arg1);
	xmlAddChild(node, arg2);
	return node;
}

xmlNodePtr
synopsis_create_alternative(xmlNodePtr arg1,
		xmlNodePtr arg2)
{
	xmlNodePtr	node;
	
	node = xmlNewNode(NULL, "alternative");
#ifdef DEBUG
	xmlSetProp(node, "creator", __func__);
#endif
	xmlAddChild(node, arg1);
	xmlAddChild(node, arg2);
	return node;
}

xmlNodePtr
synopsis_catenate_nodes(xmlNodePtr node1,
		xmlNodePtr node2)
{
	xmlAddSibling(node1, node2); 
	return node1;
}

/**
 * synopsis_create_node_with_option:
 * @option: the option name starting with a dash
 * @returns: the new node
 *
 * Creates a new node representing an option.
 */
xmlNodePtr 
synopsis_create_node_with_option(const gchar *option) 
{
	xmlNodePtr	node;
	node = xmlNewNode(NULL, "argument");
#ifdef DEBUG
	xmlSetProp(node, "creator", __func__);
#endif
	xmlSetProp(node, "option-name", option);
	return node;
}


/**
 * synopsis_create_node_with_long_option:
 * @option: the long option name starting with two dashes
 * @returns: the new node
 *
 * Creates a new node representing a long option.
 */
xmlNodePtr 
synopsis_create_node_with_long_option(const gchar *option) 
{
	xmlNodePtr	node;
	node = xmlNewNode(NULL, "argument");
#ifdef DEBUG
	xmlSetProp(node, "creator", __func__);
#endif
	xmlSetProp(node, "long-option-name", option);
	return node;
}

xmlNodePtr 
synopsis_create_node_with_option_argument(
		const gchar *option,
		const gchar *argument, 
		const gchar *op) 
{
	xmlNodePtr	node;
	node = xmlNewNode(NULL, "argument");
#ifdef DEBUG
	xmlSetProp(node, "creator", __func__);
#endif
	xmlSetProp(node, "option-name", option);
	xmlSetProp(node, "argument", argument);
	xmlSetProp(node, "op", op);
	return node;
}

xmlNodePtr 
synopsis_create_node_with_long_option_argument(
		const gchar *option,
		const gchar *argument,
		const gchar *op) 
{
	xmlNodePtr	node;
	node = xmlNewNode(NULL, "argument");
#ifdef DEBUG
	xmlSetProp(node, "creator", __func__);
#endif
	xmlSetProp(node, "long-option-name", option);
	xmlSetProp(node, "argument", argument);
	xmlSetProp(node, "op", op);
	return node;
}

xmlNodePtr
synopsis_set_node_optional(xmlNodePtr node)
{
	xmlSetProp(node, "optional", "true");
	return node;
}
	
xmlNodePtr
synopsis_set_node_iterate(xmlNodePtr node)
{
	xmlSetProp(node, "iterate", "true");
	return node;
}

xmlNodePtr
program_set_node_argument(xmlNodePtr node,
		const gchar *argument_name)
{
	//g_message("--->'%s'", argument_name);
	xmlSetProp(node, "argument-name",
			g_utf8_strdown(g_strdup(argument_name), -1));
	return node;
}

xmlNodePtr
program_set_node_description(xmlNodePtr node,
		const gchar *description)
{
	xmlXPathContextPtr xpath_ctx; 
	xpath_ctx = xmlXPathNewContext(document);
	
	g_assert(node != NULL);
	g_assert(description != NULL);
	description = canonicalize_sentence(description);
	xml_create_description_node(node, description, xpath_ctx);
	return node;
}

xmlNodePtr
program_set_options_equal(xmlNodePtr node1,
		xmlNodePtr node2)
{
	xmlXPathContextPtr xpath_ctx; 
	xpath_ctx = xmlXPathNewContext(document);
	gint   n;
	gchar *tmp;	
	gchar *names[] = { 
		"long-option-name", "option-name", "argument-name", NULL
	};

	g_message("%s(): %p %p", __func__, node1, node2);
	/*
	 * Yeah, we know about that.
	 */
	if (node1 == node2) {
		g_warning("equal");
		return node1;
	}
	if (node1 == NULL) {
		g_warning("node1 == NULL");
		return node2;
	}
	if (node2 == NULL) {
		g_warning("node2 == NULL");
		return node1;
	}
			
	/*
	 * We copy a finite set of attributes if we see them.
	 */
	n = 0;
	while (names[n] != NULL) {
		tmp = xmlGetProp(node2, names[n]);
		if (tmp != NULL) {
			g_message("'%s'='%s'", names[n], tmp);
			xmlSetProp(node1, names[n], g_strdup(tmp));
		}
		n++;
	}
	/*
	 * Destroying the secondary node.
	 */
	//xmlUnlinkNode(node2);
	//xmlFreeNode(node2);
	//xmlSetProp(node1, "node1", g_strdup("yes"));
	xmlSetProp(node2, "deleted", g_strdup("yes"));
	
	return node1;	
}


void
synopsis_remove_nodes_from_program(const gchar *name)
{
	gchar             *path; 
	xmlXPathContextPtr xpath_ctx; 
	xmlNodePtr         synopsis_node; 

	/* Initialization */
	man_cache_initialize();
	xpath_ctx = xmlXPathNewContext(document);
	/* Get or create the synopsis node of the program. */
	path = g_strdup_printf(
			"/man-cache/programs/program[@name='%s']/synopsis", name);
	synopsis_node = xpath_get_node(path, xpath_ctx, FALSE);
	/* Removing */
	if (synopsis_node != NULL) {
		xmlUnlinkNode(synopsis_node);
		xmlFreeNode(synopsis_node);
	}
finalize:
	xmlXPathFreeContext(xpath_ctx); 
	g_free(path);
}

void
synopsis_add_node_to_program(
		const gchar *name, 
		xmlNodePtr node)
{
	xmlXPathContextPtr xpath_ctx; 
	xmlNodePtr         synopsis_node; 
	xmlNodePtr         child;
	gchar             *path; 

	g_assert(name != NULL && node != NULL);
	
	/* Initialization */
	man_cache_initialize();
	xpath_ctx = xmlXPathNewContext(document);
	/* Get or create the synopsis node of the program. */
	path = g_strdup_printf("/man-cache/programs/program[@name='%s']/synopsis", name);
	synopsis_node = xpath_get_node(path, xpath_ctx, TRUE);
	/* Set the text content for the node */
	child = xmlAddChild(synopsis_node, node);
	if (child == NULL)
		g_warning("%s(): Internal program error.", __func__);
finalize:
	xmlXPathFreeContext(xpath_ctx); 
	g_free(path);
}


void
program_reset_all_history_counter(void)
{
#ifdef DEBUG
	g_warning("%s(): Not implemented yet.", __func__);
#endif
}

static void
do_increment_short_option_history_counter(
		xmlXPathContextPtr xpath_ctx,
		const gchar *name, 
		const gchar *option) 
{
	xmlNodePtr     option_node;
	xmlNodePtr     counter_node;
	gint           counter;
	gchar         *path;
	gchar         *tmp;
	
	// If we are at the end of the option we simply return.
	if (*option == '\0')
		return;
#ifdef DEBUG
	g_message("%s(): option: -'%c'", __func__, *option);	
#endif

	/* 
	 * We try to find the option itself. If it does not exists we will 
	 * not create it.	
	 */
	path = g_strdup_printf("/man-cache/programs/program[@name='%s']/option[@option-name='-%c']", name, *option);
	option_node = xpath_get_node(path, xpath_ctx, FALSE);
	g_free(path);
	if (option_node == NULL) 
		goto next_char;
	
	// We create the counters node if we have to
	path = g_strdup_printf("/man-cache/programs/program[@name='%s']/option[@option-name='-%c']/importance/counter",
			name, *option);
	counter_node = xpath_get_node(path, xpath_ctx, TRUE);
	g_free(path);
	if (option_node == NULL) 
		goto next_char;

	// Getting and incrementing the counter value.
	tmp = xmlGetProp(counter_node, "in-history");
	if (tmp != NULL) {
		counter = atoi(tmp);
		xmlFree(tmp);
	}else{
		counter = 0;
	}
	++counter;
	tmp = g_strdup_printf("%d", counter);
	xmlSetProp(counter_node, "in-history", tmp);
	g_free(tmp);
	
next_char:
	// We call the function recursively to handle the end of the option.
	do_increment_short_option_history_counter(xpath_ctx, name, option + 1);
}

void
program_increment_option_history_counter(
		const gchar *name, 
		const gchar *option) 
{
	xmlXPathContextPtr xpath_ctx; 
	
	g_assert(name != NULL && option != NULL);
#ifdef DEBUG
	g_message("%s(): name: '%s' option: '%s'", __func__, name, option);
#endif
	/* Initializing. */
	man_cache_initialize();
	xpath_ctx = xmlXPathNewContext(document);

	if (option[0] != '-')
		return;
	if (option[1] != '-')
		do_increment_short_option_history_counter(xpath_ctx, name, option + 1);
	
finelize:
	xmlXPathFreeContext(xpath_ctx); 
}
	
/**
 * program_info_in_cache:
 * @name: the name of the program
 * @returns: TRUE if we have this program in cache
 *
 * Checks if we have information in the cache about the given
 * program.
 */
gboolean
program_info_in_cache(gchar *name)
{
	xmlXPathContextPtr xpath_ctx; 
	xmlXPathObjectPtr xpath_obj;
	gchar *path;
	gint   n;	
	/*
	 * Initialization...
	 */
	man_cache_initialize();
	path = g_strdup_printf("count(/man-cache/programs/program[@name='%s'])", name);
	xpath_ctx = xmlXPathNewContext(document);
	/*
	 * Search...
	 */
	xpath_obj = xmlXPathEvalExpression(path, xpath_ctx);
	n = (gint) xmlXPathCastToNumber(xpath_obj);
	/*
	 * Finalize...
	 */
	g_free(path);
	xmlXPathFreeObject(xpath_obj);

	return n != 0;
}

/**
 * program_add_description:
 * @name: the name of the program
 * @descr: the short description of the program
 * @returns: TRUE if the operation could be performed
 *
 * Creates a new node for the program, then stores the description for it.
 * This function will not change the description if the program or the 
 * description itself has a protection="read-write" attribute setting.
 */
gboolean
program_add_description(
	gchar *name,
	gchar *descr)
{
	xmlXPathContextPtr  xpath_ctx; 
	xmlNodePtr          program_node, description_node;
	gboolean            retval;
	
	g_assert(name != NULL && descr != NULL);
	
#ifdef DEBUG
	g_message("%s(): Start", __func__);
#endif
	/* Initialization. */
	man_cache_initialize();
	xpath_ctx = xmlXPathNewContext(document);
	/*
	 * If the program has no node, we create one;
	 */
	program_node = _program_get_program_node_for_write(xpath_ctx, name);
	if (program_node == NULL) {
		xmlXPathFreeContext(xpath_ctx); 
		return FALSE;
	}
	/*
	 * We create or set the description node if it is not exist as a 
	 * write-protected node.
	 */
	descr = canonicalize_sentence(descr);
	retval = xml_create_description_node(program_node, descr, xpath_ctx);

finalize:
	xmlXPathFreeContext(xpath_ctx); 
	g_free(descr); // This is not the original, it's a copy!
	return retval;
}

/**
 * program_get_description:
 * @name: the name of the program
 * @returns: the description for the program
 *
 * Returns the description belongs to the program.
 */
const gchar *
program_get_description(gchar *name)
{
	xmlXPathContextPtr xpath_ctx; 
	xmlXPathObjectPtr xpath_obj;
	gchar *desc;
	gchar *path;
	
	/*
	 * Initialization...
	 */
	man_cache_initialize();
	path = g_strdup_printf("/man-cache/programs/program[@name='%s']/description", name);
	xpath_ctx = xmlXPathNewContext(document);
	/*
	 * Search...
	 */
	xpath_obj = xmlXPathEvalExpression(path, xpath_ctx);
	desc = (gchar *) xmlXPathCastToString(xpath_obj);
	/*
	 * Finalize...
	 */
	g_free(path);
	xmlXPathFreeObject(xpath_obj);
	xmlXPathFreeContext(xpath_ctx); 
	return desc;
}

/**
 * program_add_long_description:
 * @name: the name of the program
 * @descr: the long description of the program
 * @returns: TRUE if the operation could be performed
 *
 * Creates a new node for the program, then stores the long description 
 * for it. This function will not change the long description if the 
 * program or the long description itself has a protection="read-write" 
 * attribute setting.
 */
gboolean
program_add_long_description(
	gchar *name,
	gchar *descr)
{
	xmlXPathContextPtr  xpath_ctx; 
	xmlNodePtr          program_node, description_node;
	gboolean            retval;
	
	g_assert(name != NULL && descr != NULL);
	
#ifdef DEBUG
	g_message("%s(): Start", __func__);
#endif
	/* Initialization. */
	man_cache_initialize();
	xpath_ctx = xmlXPathNewContext(document);
	/*
	 * If the program has no node, we create one;
	 */
	program_node = _program_get_program_node_for_write(xpath_ctx, name);
	if (program_node == NULL) {
		xmlXPathFreeContext(xpath_ctx); 
		return FALSE;
	}
	/*
	 * We create or set the long description node if it is not exist 
	 * as a write-protected node.
	 */
	descr = canonicalize_paragraph(descr);
	retval = xml_create_long_description_node(program_node, descr, xpath_ctx);

finalize:
	xmlXPathFreeContext(xpath_ctx); 
	g_free(descr); // This is not the original, it's a copy!
	return retval;
}

/**
 * program_get_long_description:
 * @name: the name of the program
 * @returns: name long description of the program
 * 
 * Returns the long description belongs to the given program.
 */
const gchar *
program_get_long_description(gchar *name)
{
	xmlXPathContextPtr xpath_ctx; 
	xmlXPathObjectPtr xpath_obj;
	gchar *longdesc;
	gchar *path;
	
	/*
	 * Initialization...
	 */
	man_cache_initialize();
	path = g_strdup_printf("/man-cache/programs/program[@name='%s']/long-description", name);
	xpath_ctx = xmlXPathNewContext(document);
	/*
	 * Search...
	 */
	xpath_obj = xmlXPathEvalExpression(path, xpath_ctx);
	longdesc = (gchar *) xmlXPathCastToString(xpath_obj);
	/*
	 * Finalize...
	 */
	g_free(path);
	xmlXPathFreeObject(xpath_obj);
	xmlXPathFreeContext(xpath_ctx); 
	return longdesc;
}

xmlNodePtr
program_add_long_option(gchar *name,
		gchar *long_option)
{
	xmlXPathContextPtr xpath_ctx; 
	xmlNodePtr         long_option_node;

	/* Initiailzation */
	man_cache_initialize();
	long_option = canonicalize_long_option(long_option);
	xpath_ctx = xmlXPathNewContext(document);
	/* creating the option */
	long_option_node = _program_get_long_option_node_for_write(xpath_ctx, name, long_option);
	return long_option_node;
}

/**
 * program_add_option:
 * @name: the name of the program
 * @option: the option name starting with a dash
 * @returns: TRUE if the opertation suceeded
 *
 * Adds a new option to the given program.
 */
xmlNodePtr
program_add_option(gchar *name, 
		gchar *option)
{
	xmlXPathContextPtr xpath_ctx; 
	xmlNodePtr         option_node;

	/* Initiailzation */
	man_cache_initialize();
	option = canonicalize_option(option);
	xpath_ctx = xmlXPathNewContext(document);
	/* creating the option */
	option_node = _program_get_option_node_for_write(xpath_ctx, name, option);
	return option_node;
}

gboolean
program_add_option_with_descr( gchar *name, 
			gchar *option,
			gchar *descr)
{
	xmlXPathContextPtr xpath_ctx; 
	xmlNodePtr         option_node;
	xmlNodePtr         description_node;
	xmlNodePtr	   importance_node;
	gchar *path;
	gint n;
	
#ifdef DEBUG
	g_message("%s(): Start", __func__);
#endif
	/* Initiailzation */
	man_cache_initialize();
	option = canonicalize_option(option);
	descr = canonicalize_sentence(descr);
	xpath_ctx = xmlXPathNewContext(document);
	/*
	 * We try to find an option with the given long option name. 
	 */
	path = g_strdup_printf(
			"/man-cache/programs/program[@name='%s']/"
			"option[@option-name='%s']", name, option);
	option_node = xpath_get_node(path, xpath_ctx, FALSE);
	g_free(path);
	if (option_node != NULL) 
		goto option_exists;
	/*
	 * Now we create a new <option> node, because it is not existing with this long option yet.
	 */
	path = g_strdup_printf(
			"/man-cache/programs/program[@name='%s']/option", 
			name);
	option_node = xpath_create_new_node(path, xpath_ctx);
	g_free(path);
	xmlSetProp(option_node, "option-name", option);
#ifdef DEBUG
	xmlSetProp(option_node, "creator", __func__);
#endif
	
option_exists:
	/*
	 * If the description exists, we will not change it.
	 */
	path = g_strdup_printf(
			"count(/man-cache/programs/program[@name='%s']/"
			"option[@option-name='%s']/description)", name, option);
	n = xpath_to_gint(path, xpath_ctx);
	g_free(path);
	if (n != 0) 
		goto finalize;
	/*
	 * We add a new description to the long option.
	 */
	add_description_to_option(xpath_ctx, name, option, descr);
	/*
	 * If we added a new description, we try to find out some information
	 * about the importance of this option and we set it to the <option>
	 * node.
	 */
	importance_node = xmlNewChild(option_node, NULL, "importance", NULL);
	if (is_option_indirect(name, option, descr))
		xmlSetProp(importance_node, "indirect", "true");
	if (is_option_prompt(name, option, descr))
		xmlSetProp(importance_node, "prompt", "true");
	if (is_option_ignored(name, option, descr))
		xmlSetProp(importance_node, "ignored", "true");
#ifdef DEBUG
	xmlSetProp(importance_node, "creator", __func__);
#endif
	
finalize:
	g_free(option); // This is not the original, it's a copy!
	g_free(descr);   // This is not the original, it's a copy!
	xmlXPathFreeContext(xpath_ctx); 
	return TRUE;
}



/**
 * program_get_option_description:
 * @name: the name of the program
 * @option: the short or long option starting with a dash
 * @returns: one sentence description in a newly allocated string
 *
 * Returns the one sentence description of the option in a newly allocated xml
 * encoded string.
 */
gchar *
program_get_option_description(
		const gchar *name,
		const gchar *option)
{
	xmlXPathContextPtr  xpath_ctx; 
	xmlXPathObjectPtr   xpath_obj;
	gchar              *path;
	gchar              *description;
	/* 
	 * Initialization. 
	 */
	man_cache_initialize();
	xpath_ctx = xmlXPathNewContext(document);
	/* 
	 * Searching the description. 
	 */
	path = g_strdup_printf(
			"/man-cache/programs/program[@name='%s']"
			"/option[@option-name='%s' or @long-option-name='%s']/description", 
			name, option, option);
	xpath_obj = xmlXPathEvalExpression(path, xpath_ctx);
	description = (gchar *)xmlEncodeEntitiesReentrant(document, xmlXPathCastToString(xpath_obj));
	
	g_free(path);
	xmlXPathFreeObject(xpath_obj);
	return description;
}

/**
 * program_get_argument_category:
 * @name: the name of the argument
 * @position: which argument to read
 * @returns: the category of the given argument
 *
 * Returns the category of the nth argument.
 */
const gchar *
program_get_argument_category(
		const gchar *name,
		gint position)
{
	xmlXPathContextPtr xpath_ctx; 
	xmlXPathObjectPtr xpath_obj;
	gchar *path;
	gchar *category;
	
	g_assert(name != NULL);
#ifdef DEBUG
	g_message("%s(): name: '%s' position: %d", __func__, name, position);
#endif
	/* Initialization. */
	man_cache_initialize();
	xpath_ctx = xmlXPathNewContext(document);
	/* Getting the category. */
	path = g_strdup_printf(
			"/man-cache/programs/program[@name='%s']/"
			"synopsis/argument-list/argument[position()=%d]/@category",
			name, position); 
	
	xpath_obj = xmlXPathEvalExpression(path, xpath_ctx);
	category = xmlXPathCastToString(xpath_obj);
	
	g_free(path);
	xmlXPathFreeObject(xpath_obj);
	return category;
}

/**
 * program_get_argument_optional:
 * @name: the name of the program
 * @position: which argument to check
 * @returns: TRUE if the given argument is optional
 *
 * Returns TRUE if the argument can be omitted.
 */
gboolean 
program_get_argument_optional(
		const gchar *name,
		gint position)
{
	xmlXPathContextPtr xpath_ctx; 
	xmlXPathObjectPtr  xpath_obj;
	gchar             *path;
	gchar             *optional;
	gboolean           retval;
	
	g_assert(name != NULL);
#ifdef DEBUG
	g_message("%s(): name: '%s' position: %d", __func__, name, position);
#endif
	/* Initialization. */
	man_cache_initialize();
	xpath_ctx = xmlXPathNewContext(document);
	/* Getting the attribute. */
	path = g_strdup_printf(
			"/man-cache/programs/program[@name='%s']/"
			"synopsis/argument-list/argument[position()=%d]/@optional",
			name, position); 
	
	xpath_obj = xmlXPathEvalExpression(path, xpath_ctx);
	optional = xmlXPathCastToString(xpath_obj);
	if (optional == NULL)
		optional = FALSE;
	else
		optional = g_ascii_strcasecmp(optional, "true") == 0;
	
	g_free(path);
	xmlXPathFreeObject(xpath_obj);
	return optional;
}

/**
 * program_get_argument_iterate:
 * @name: the name of the program
 * @position: which argument to check
 * @returns: TRUE if the argument is iterated
 *
 * Returns TRUE if the given argument can be repeated.
 */
gboolean 
program_get_argument_iterate(
		const gchar *name,
		gint position)
{
	xmlXPathContextPtr xpath_ctx; 
	xmlXPathObjectPtr  xpath_obj;
	gchar             *path;
	gchar             *iterate;
	gboolean           retval;
	
	g_assert(name != NULL);
#ifdef DEBUG
	g_message("%s(): name: '%s' position: %d", __func__, name, position);
#endif
	/* Initialization. */
	man_cache_initialize();
	xpath_ctx = xmlXPathNewContext(document);
	/* Getting the attribute. */
	path = g_strdup_printf(
			"/man-cache/programs/program[@name='%s']/"
			"synopsis/argument-list/argument[position()=%d]/@iterate",
			name, position); 
	
	xpath_obj = xmlXPathEvalExpression(path, xpath_ctx);
	iterate = xmlXPathCastToString(xpath_obj);
	if (iterate == NULL)
		retval = FALSE;
	else
		retval = g_ascii_strcasecmp(iterate, "true") == 0;
	
	g_free(path);
	xmlXPathFreeObject(xpath_obj);
	return retval;
}

GList *
program_get_program_list(void)
{
	xmlXPathContextPtr  xpath_ctx; 
	xmlXPathObjectPtr   xpath_obj;
	GList              *list = NULL;
	gchar              *path;
	gchar              *command_name;
	gint                i, n;
	/*
	 * Initialization.
	 */
	man_cache_initialize();
	xpath_ctx = xmlXPathNewContext(document);
	/*
	 * Counting the options.
	 */
	path = g_strdup_printf("count(/man-cache/programs/program)");
	n = xpath_to_gint(path, xpath_ctx);
	g_free(path);
#ifdef DEBUG
	g_message("%s(): n=%d", __func__, n);
#endif
	/*
	 * Inserting program names into the list.
	 */
	for (i = 1; i <= n; ++i){
		path = g_strdup_printf("/man-cache/programs/program[position()=%d]/@name", i);
		xpath_obj = xmlXPathEvalExpression(path, xpath_ctx);
		command_name = xmlXPathCastToString(xpath_obj);	
		g_free(path);
		xmlXPathFreeObject(xpath_obj);
		list = g_list_append(list, command_name);
	}
	
	return list;
}

/**
 * program_get_option_list:
 *
 * Returns a list with the options (short or long option names) for the given
 * program with the given restraint. The list contains strings statically 
 * allocated, should not be tamered with.
 */
GList *
program_get_option_list(const gchar *name,
		const gchar *restraint)
{
	xmlXPathContextPtr xpath_ctx; 
	xmlXPathObjectPtr xpath_obj;
	GList *list = NULL;
	gchar *path;
	gchar *count;
	gchar *option_name;
	gint  n, i;
	/*
	 * Initialization.
	 */
	man_cache_initialize();
	xpath_ctx = xmlXPathNewContext(document);
	count = g_strdup_printf("count(/man-cache/programs/program[@name='%s']/option)", name);

	/*
	 * Counting the options.
	 */
	n = xpath_to_gint(count, xpath_ctx);
	/*
	 * Adding the options one by one.
	 */
	for (i = 1; i <= n; ++i){
		if (restraint == NULL)
			path = g_strdup_printf(
					"/man-cache/programs/program[@name='%s']/"
					"option[position()=%d]/@option-name", name, i);
		else
			path = g_strdup_printf(
					"/man-cache/programs/program[@name='%s']/"
					"option[(position()=%d) and %s]/@option-name", 
					name, i, restraint);
#ifdef DEBUG
		g_message("%s():---> \"%s\"", __func__, path);
#endif
		xpath_obj = xmlXPathEvalExpression(path, xpath_ctx);
		option_name = xmlXPathCastToString(xpath_obj);	
#ifdef DEBUG
		g_message("%s():---> \"%s\"", __func__, option_name);
#endif
		
		g_free(path);
		xmlXPathFreeObject(xpath_obj);
		
		if (option_name[0] == '\0') {
			if (restraint == NULL)
				path = g_strdup_printf(
						"/man-cache/programs/program[@name='%s']/"
						"option[position()=%d]/@long-option-name", 
						name, i);
			else
				path = g_strdup_printf(
						"/man-cache/programs/program[@name='%s']/"
						"option[(position()=%d) and %s]/@long-option-name",
						name, i, restraint);
			xpath_obj = xmlXPathEvalExpression(path, xpath_ctx);
			option_name = xmlXPathCastToString(xpath_obj);
			g_free(path);
			xmlXPathFreeObject(xpath_obj);
		}
		
		if (option_name[0] != '\0')
			list = g_list_append(list, option_name);
	}

finalize:
	g_free(count);
	xmlXPathFreeContext(xpath_ctx); 
	
	return list;
}


static gboolean
is_option_ignored(const gchar *name,
		const gchar *option,
		const gchar *description)
{
	if (g_ascii_strncasecmp(description, "(ignored)", 9) == 0 ||
			g_ascii_strncasecmp(description, "ignored", 7) == 0)
		return TRUE;
	return FALSE;
}

static gboolean
is_option_prompt(const gchar *name,
		const gchar *option,
		const gchar *description)
{
	if (g_str_equal(option, "--force") || 
			g_str_equal(option, "--interactive"))
		return TRUE;
	
	if (g_ascii_strncasecmp(description, "Prompt", 6) == 0 ||
			g_ascii_strncasecmp(description, "Do not prompt", 13) == 0)
		return TRUE;
	return FALSE;
}

static gboolean
is_long_option_standard(const gchar *name,
		const gchar *long_option,
		const gchar *description)
{
	if (g_str_equal(long_option, "--version")            || 
			g_str_equal(long_option, "--help")   ||
			g_str_equal(long_option, "--quiet")  ||
			g_str_equal(long_option, "--silent") ||
			g_str_equal(long_option, "--verbose")
			)
		return TRUE;
	return FALSE;
}

static gboolean
is_option_indirect(const gchar *name,
		const gchar *option,
		const gchar *description)
{
	if (g_ascii_strncasecmp(description, "Same as -", 9) == 0 ||
			g_ascii_strncasecmp(description, "Likewise", 8) == 0 ||
			g_ascii_strncasecmp(description, "Like -", 6) == 0 ||
			g_ascii_strncasecmp(description, "Equivalent", 10) == 0)
		return TRUE;
	return FALSE;
}

#if 0
gboolean
program_add_longopt(
	gchar *name, 
	gchar *longopt)
{
	xmlXPathContextPtr xpath_ctx; 
	xmlXPathObjectPtr xpath_obj;
	xmlNodePtr new_node, new_node_child, grandchild;
	xmlNodePtr nodes_node;
	gchar *path = g_strdup_printf("/man-cache/programs/%s", name);
	gchar *count = g_strdup_printf("count(/man-cache/programs/%s)", name);
	gchar *tmp;
	gint n;
#ifdef DEBUG
	g_message("%s(): Start", __func__);
#endif
	man_cache_initialize();
	longopt = canonicalize_long_option(longopt);
	xpath_ctx = xmlXPathNewContext(document);
	/*
	 * If the program has no node, we create one;
	 */
	n = xpath_to_gint(count, xpath_ctx);
	if (n == 0) {
		xpath_obj = xmlXPathEvalExpression("/man-cache/programs", xpath_ctx);
		nodes_node = xpath_obj->nodesetval->nodeTab[0];
		new_node = xmlNewChild(nodes_node, NULL, name, NULL);
		xmlXPathFreeObject(xpath_obj);
		goto longopt_not_exists;
	}
	/*
	 * If the long option exists, we don't have to create it.
	 */
	g_free(count);
	count = g_strdup_printf("count(/man-cache/programs/%s/option[@long-option-name='%s'])",
			name, longopt);
	xpath_obj = xmlXPathEvalExpression(count, xpath_ctx);
	n = (gint) xmlXPathCastToNumber(xpath_obj);
	if (n != 0)
		goto finalize;
	
longopt_not_exists:
	xpath_obj = xmlXPathEvalExpression(path, xpath_ctx);
	nodes_node = xpath_obj->nodesetval->nodeTab[0];
	new_node = xmlNewChild(nodes_node, NULL, "option", NULL);
	xmlSetProp(new_node, "long-option-name", longopt);
#ifdef DEBUG
	xmlSetProp(new_node, "creator", __func__);
#endif


finalize:
	xmlXPathFreeObject(xpath_obj);
	xmlXPathFreeContext(xpath_ctx); 
	g_free(longopt); // This is not the original, it's a copy!
	g_free(path);
	g_free(count);
	return TRUE;
}
#endif


static void
add_description_to_option(xmlXPathContextPtr xpath_ctx,
 			  const gchar *name,
			  const gchar *option,
			  const gchar *description)
{
	xmlXPathObjectPtr xpath_obj;
	xmlNodePtr nodes_node, new_node;
	gchar *path, *tmp;
	
	g_assert(name != NULL && option != NULL && description != NULL);
	/*
	 * Searching the option.
	 */
	path = g_strdup_printf(
			"/man-cache/programs/program[@name='%s']/"
			"option[@option-name='%s' or @long-option-name='%s']", 
			name, option, option);
	xpath_obj = xmlXPathEvalExpression(path, xpath_ctx);
	if (xpath_obj == NULL) {
		g_free(path);
		g_warning("%s(): Option '%s' of program '%s' not found.", __func__, option, name);
		return;
	}
	/*
	 * Adding a new node.
	 */
	nodes_node = xpath_obj->nodesetval->nodeTab[0];
	new_node = xmlNewChild(nodes_node, NULL, "description", description);
	/* 
	 * The length of the description. 
	 */
	tmp = g_strdup_printf("%ld", g_utf8_strlen(description, -1));
	xmlSetProp(new_node, "length", tmp);
#ifdef DEBUG
	xmlSetProp(new_node, "creator", __func__);
#endif

finalize:
	xmlXPathFreeObject(xpath_obj);
	g_free(path);
	g_free(tmp);
}



gboolean
program_add_longopt_with_descr(gchar *name, 
			       gchar *longopt,
			       gchar *descr)
{
	xmlXPathContextPtr xpath_ctx; 
	xmlNodePtr         option_node;
	xmlNodePtr         description_node;
	xmlNodePtr	   importance_node;
	gchar *path;
	gint n;
	
#ifdef DEBUG
	g_message("%s(): Start", __func__);
#endif
	/* Initiailzation */
	man_cache_initialize();
	longopt = canonicalize_long_option(longopt);
	descr = canonicalize_sentence(descr);
	xpath_ctx = xmlXPathNewContext(document);
	/*
	 * We try to find an option with the given long option name. 
	 */
	path = g_strdup_printf(
			"/man-cache/programs/program[@name='%s']/"
			"option[@long-option-name='%s']", name, longopt);
	option_node = xpath_get_node(path, xpath_ctx, FALSE);
	g_free(path);
	if (option_node != NULL) 
		goto longopt_exists;
	/*
	 * Now we create a new <option> node, because it is not 
	 * existing with this long option yet.
	 */
	path = g_strdup_printf("/man-cache/programs/program[@name='%s']/option", name);
	option_node = xpath_create_new_node(path, xpath_ctx);
	g_free(path);
	xmlSetProp(option_node, "long-option-name", longopt);
#ifdef DEBUG
	xmlSetProp(option_node, "creator", __func__);
#endif
	
longopt_exists:
	/*
	 * If the description exists, we will not change it.
	 */
	path = g_strdup_printf(
			"count(/man-cache/programs/program[@name='%s']/"
			"option[@long-option-name='%s']/description)", name, longopt);
	n = xpath_to_gint(path, xpath_ctx);
	g_free(path);
	if (n != 0) 
		goto finalize;
	/*
	 * We add a new description to the long option.
	 */
	add_description_to_option(xpath_ctx, name, longopt, descr);
	/*
	 * If we added a new description, we try to find out some information
	 * about the importance of this option and we set it to the <option>
	 * node.
	 */
	importance_node = xmlNewChild(option_node, NULL, "importance", NULL);
	if (is_option_indirect(name, longopt, descr))
		xmlSetProp(importance_node, "indirect", "true");
	if (is_option_prompt(name, longopt, descr))
		xmlSetProp(importance_node, "prompt", "true");
	if (is_option_ignored(name, longopt, descr))
		xmlSetProp(importance_node, "ignored", "true");
	if (is_long_option_standard(name, longopt, descr))
		xmlSetProp(importance_node, "standard", "true");
#ifdef DEBUG
	xmlSetProp(importance_node, "creator", __func__);
#endif
	
finalize:
	g_free(longopt); // This is not the original, it's a copy!
	g_free(descr);   // This is not the original, it's a copy!
	xmlXPathFreeContext(xpath_ctx); 
	return TRUE;
}


gboolean
program_add_options_with_descr( gchar *name,
				gchar *option,
				gchar *longopt,
				gchar *descr)
{
	xmlXPathContextPtr xpath_ctx; 
	xmlNodePtr         option_node;
	xmlNodePtr         description_node;
	xmlNodePtr	   importance_node;
	gchar *path;
	gint n;
	
#ifdef DEBUG
	g_message("%s(): Start", __func__);
#endif
	/* Initiailzation */
	man_cache_initialize();
	longopt = canonicalize_long_option(longopt);
	option = canonicalize_option(option);
	descr = canonicalize_sentence(descr);
	xpath_ctx = xmlXPathNewContext(document);
	/*
	 * We try to find an option with the given long option name. 
	 */
	// FIXME: This is not the way to do it, what if we did know the short version of the option, and learning the
	// long version just now?
	path = g_strdup_printf(
			"/man-cache/programs/program[@name='%s']/"
			"option[(@long-option-name='%s') or (@option='%s')]", 
			name, longopt, option);
	option_node = xpath_get_node(path, xpath_ctx, FALSE);
	g_free(path);
	if (option_node != NULL) 
		goto longopt_exists;
	/*
	 * Now we create a new <option> node, because it is not existing with this long option yet.
	 */
	path = g_strdup_printf("/man-cache/programs/program[@name='%s']/option", name);
	option_node = xpath_create_new_node(path, xpath_ctx);
	g_free(path);
	xmlSetProp(option_node, "long-option-name", longopt);
	xmlSetProp(option_node, "option-name", option);
#ifdef DEBUG
	xmlSetProp(option_node, "creator", __func__);
#endif
	
longopt_exists:
	/*
	 * If the description exists, we will not change it.
	 */
	path = g_strdup_printf(
			"count(/man-cache/programs/program[@name='%s']/"
			"option[@long-option-name='%s' or @option='%s']/description)", 
			name, longopt, option);
	n = xpath_to_gint(path, xpath_ctx);
	g_free(path);
	if (n != 0) 
		goto finalize;
	/*
	 * We add a new description to the long option.
	 */
	add_description_to_option(xpath_ctx, name, longopt, descr);
	/*
	 * If we added a new description, we try to find out some information
	 * about the importance of this option and we set it to the <option>
	 * node.
	 */
	importance_node = xmlNewChild(option_node, NULL, "importance", NULL);
	if (is_option_indirect(name, longopt, descr))
		xmlSetProp(importance_node, "indirect", "true");
	if (is_option_prompt(name, longopt, descr))
		xmlSetProp(importance_node, "prompt", "true");
	if (is_option_ignored(name, longopt, descr))
		xmlSetProp(importance_node, "ignored", "true");
	if (is_long_option_standard(name, longopt, descr))
		xmlSetProp(importance_node, "standard", "true");
#ifdef DEBUG
	xmlSetProp(importance_node, "creator", __func__);
#endif
	
finalize:
	g_free(longopt); // This is not the original, it's a copy!
	g_free(option); // This is not the original, it's a copy!
	g_free(descr);   // This is not the original, it's a copy!
	xmlXPathFreeContext(xpath_ctx); 
	return TRUE;

}

/**
 *
 */
gboolean
type_check_command_name(CuiLexElement *lexelement)
{
	gchar *path;
	GtkTextBuffer *text_buffer;
	GtkTextIter begin, end;
#ifdef DEBUG	
	g_message("%s(): Start: '%s'", __func__, lexelement->text);
#endif
	
	text_buffer = gtk_text_mark_get_buffer(lexelement->begin);
	gtk_text_buffer_get_iter_at_mark(text_buffer, 
				&begin, lexelement->begin);
	gtk_text_buffer_get_iter_at_mark(text_buffer, 
				&end, lexelement->end);

	path = g_find_program_in_path(lexelement->text);
	if (path == NULL && !bash_internal(lexelement->text)) {
		gtk_text_buffer_apply_tag_by_name(text_buffer,
				"underline-error",
				&begin, &end);
		return FALSE;
	} else {
		gtk_text_buffer_remove_tag_by_name(text_buffer,
				"underline-error",
				&begin, &end);
		return TRUE;
	}
}


TypingCheckResult
program_check_longopt_typing(
		const gchar *name, 
		const gchar *longopt)
{
	xmlXPathContextPtr xpath_ctx; 
	xmlXPathObjectPtr xpath_obj;
	gchar *count;
	gint n;
	
	man_cache_initialize();
	longopt = canonicalize_long_option(longopt);
	xpath_ctx = xmlXPathNewContext(document);
	/*
	 * If the program has no node we can't check the long option.
	 */
	count = g_strdup_printf("count(/man-cache/programs/program[@name='%s'])", name);
	n = xpath_to_gint(count, xpath_ctx);
	if (n = 0) {
		g_free(longopt);
		g_free(count);
		return TypingNotChecked;	
	}
	/*
	 * Otherwise if there is an option with this long name, the spelling
	 * is correct.
	 */
	g_free(count);
	count = g_strdup_printf(
			"count(/man-cache/programs/program[@name='%s']/"
			"option[@long-option-name='%s'])",
			name, longopt);
	xpath_obj = xmlXPathEvalExpression(count, xpath_ctx);
	n = (gint) xmlXPathCastToNumber(xpath_obj);
	xmlXPathFreeObject(xpath_obj);
	xmlXPathFreeContext(xpath_ctx); 
	g_free(longopt); // This is a copy.
	g_free(count);
	if (n != 0) {
		return TypingCorrect;
	}else{
		return TypingError;
	}
}

gboolean
type_check_long_option(CuiLexElement *longopt, gchar *name)
{
	GtkTextBuffer *text_buffer;
	GtkTextIter begin, end;
	
	text_buffer = gtk_text_mark_get_buffer(longopt->begin);
	gtk_text_buffer_get_iter_at_mark(text_buffer, 
				&begin, longopt->begin);
	gtk_text_buffer_get_iter_at_mark(text_buffer, 
				&end, longopt->end);

	if (program_check_longopt_typing(name, longopt->text) == TypingError) {
		gtk_text_buffer_apply_tag_by_name(text_buffer,
				"underline-error",
				&begin, &end);
		return FALSE;
	} else {
		gtk_text_buffer_remove_tag_by_name(text_buffer,
				"underline-error",
				&begin, &end);
		return TRUE;
	}
}


/*******************************************************************
 * Functions for the history window.                               *
 *******************************************************************/
void
on_history_close_button_clicked        (GtkButton       *button,
                                        gpointer         user_data)
{
	GtkWidget *widget;
	widget = lookup_widget(GTK_WIDGET(button), "history_window");
	gtk_widget_destroy(widget);
}


void
on_history_treeview_realize            (GtkWidget       *widget,
                                        gpointer         user_data)
{
	GtkTreeView *tree_view = GTK_TREE_VIEW(widget);
	GtkTreeViewColumn *column;
	GtkCellRenderer   *renderer;
	GtkTreeStore *tree_store;
	GtkTreeIter iter_level0;
	
	gint n;
	GList *actual;
#ifdef DEBUG
	g_message("%s(): Start.", __func__);
#endif
	/*
	 * Tree model
	 */
	tree_store = gtk_tree_store_new(1,
			G_TYPE_STRING    /* The command itself.         */
			);
	n = 1;
	actual = history_list;
	while(actual != NULL) {
#ifdef DEBUG
		g_message("%s(): reading into the store: %d", __func__, n);
#endif
		gtk_tree_store_append(tree_store, &iter_level0, NULL);
		gtk_tree_store_set(tree_store, &iter_level0, 
	  		0, actual->data, -1);
next_item:
		actual = actual->next;
		n++;
	}
	gtk_tree_view_set_model(tree_view, GTK_TREE_MODEL(tree_store));
	
	/*
	 * The columns.
	 */
  	/* Column0: the name */
  	renderer = gtk_cell_renderer_text_new ();
  	column = gtk_tree_view_column_new_with_attributes(
		  _("The command"), 
		  renderer, "text" /*"markup"*/, 0, NULL);
  	gtk_tree_view_append_column(tree_view, column);
	gtk_tree_view_set_rules_hint(tree_view, TRUE);
}


