%{
/*
 * Copyright (C) 2004 Laszlo Pere
 * $Id: ls_parser.y,v 1.1 2006/02/24 08:19:15 pipas Exp pipas $
 *
 * This is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Library General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Log: ls_parser.y,v $
 * Revision 1.1  2006/02/24 08:19:15  pipas
 * Initial revision
 *
 */
#include <gnome.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

#include "ls_parser_c.h"
#include "cuiterm.h"
#include "terminal.h"
#include "cuiwindow.h"
#include "ls_parser.h"
#include "filename.h"
#include "dnd.h"
extern CuiWindow *window_for_lex;

typedef enum {
	UsualLine,
	DeviceLine
} LineType;

/*
 * We use this structure to store all the lines we process, so we can go back
 * when all of the lines are well known;
 */
typedef struct one_line {
	CuiLexElement *mode;       /* the mode string */
	CuiLexElement *lnumber;    /* link number */
	CuiLexElement *uname;      /* the owner */
	CuiLexElement *gname;
	CuiLexElement *size; 
	CuiLexElement *timestamp;
	CuiLexElement *filename;
	CuiLexElement *origin;  /* target if the entry is a link */
	LineType       line_type;
} one_line;


static GList *list = NULL; /* The list we store the long lines */

/**
 * Frees the allocated resources and resets the parser.
 */
static inline void
finalize_parser(void)
{
	GList *l;
	one_line *line;
	/*
	 * Destroying the list. This will not destroy only the few elements we 
	 * left in the list after the last filename has been processed.
	 */
	l = list;
	while (l != NULL){
		line = l->data;
		cui_lex_element_destroy(line->mode, TRUE); 
		g_free(line->mode);
		
		cui_lex_element_destroy(line->lnumber, TRUE); 
		g_free(line->lnumber);
		
		cui_lex_element_destroy(line->uname, TRUE); 
		g_free(line->uname);
		
		cui_lex_element_destroy(line->gname, TRUE); 
		g_free(line->gname);
		
		cui_lex_element_destroy(line->size, TRUE); 
		g_free(line->size);
		
		cui_lex_element_destroy(line->timestamp, TRUE); 
		g_free(line->timestamp);
		
		cui_lex_element_destroy(line->filename, TRUE); 
		g_free(line->filename);

		if (line->origin != NULL)
			cui_lex_element_destroy(line->origin, TRUE);
		g_free(line->origin);

		l = l->next;
	}
	g_list_free(list);
	list = NULL;
}

static inline void 
append_long_line(CuiLexElement *mode,
	CuiLexElement *lnumber,
	CuiLexElement *uname,
	CuiLexElement *gname,
	CuiLexElement *size,
	CuiLexElement *timestamp,
	CuiLexElement *filename,
	CuiLexElement *origin,
	LineType       type)
{
	one_line *new_line = g_new(one_line, 1);
	new_line->mode      = g_memdup(mode, sizeof(CuiLexElement));	
	new_line->lnumber   = g_memdup(lnumber, sizeof(CuiLexElement));	
	new_line->uname     = g_memdup(uname, sizeof(CuiLexElement));	
	new_line->gname     = g_memdup(gname, sizeof(CuiLexElement));	
	new_line->size      = g_memdup(size, sizeof(CuiLexElement));	
	new_line->timestamp = g_memdup(timestamp, sizeof(CuiLexElement));	
	new_line->filename  = g_memdup(filename, sizeof(CuiLexElement));
	if (origin != NULL)
		new_line->origin  = g_memdup(origin, sizeof(CuiLexElement));
	else
		new_line->origin  = NULL;
	new_line->line_type = type;

	list = g_list_append(list, new_line);
}

static inline gdouble 
sum_long_lines(void)
{
	GList *le;
	one_line *line;
	gdouble retval = 0;

	le = list;
	while (le != NULL){
		line = le->data;
		if (line->line_type != DeviceLine)
			retval += line->size->double_val;
		le = le->next;
	}
	return retval;
}

#undef PROFILE

#ifdef PROFILE
#   include <time.h>
#endif

static inline void
process_long_lines(CuiWindow *window)
{
	GList *le;
	one_line *line;
	gdouble full_size = sum_long_lines();
	CuiFilename *fn;
	CuiWidget *widget;
	Background *background;
	CuiFileSizeChange change;
#ifdef PROFILE
	clock_t start_time, end_time;
#endif

	g_return_if_fail(CUI_IS_WINDOW(window));
	background = cui_widget_get_background(CUI_WIDGET(window));
	/*
	 * We go through the lines and create the widgets.
	 */
	le = list;
	g_message("%s(): length of list: %d", __func__, g_list_length(list));
	while (le != NULL){
		line = le->data;
		/* 
		 * Filename, stores information about the file.
		 */
#ifdef PROFILE
			start_time = clock();
#endif
		fn = cui_filename_new(line->filename->text, window->wd->str); 
#ifdef PROFILE
			end_time = clock();
			g_message("%s(): Parsing time: %.12f\n", 
					__func__,
					((double)(end_time-start_time))/CLOCKS_PER_SEC);
#endif
		cui_filename_set_mode_string(fn, line->mode->text);
		cui_filename_set_owner_and_group(fn, 
			line->uname->text, 
			line->gname->text);
		/*
		 * Link action area.
		 */
		cui_action_area_new_for_link(window, background,
			fn,
			line->lnumber->begin,
			line->lnumber->end);
		/*
		 * Filename action area.
		 */
		cui_action_area_new_for_filename(
			window,
			background, 
			fn,
			line->filename->begin, 
			line->filename->end);
		/*
		 * Filename menu.
		 */
	   	cui_file_menu_new_for_filename(window, background, fn,
	   		line->filename->begin, 
			line->filename->end);
		/*
		 * Mode action area, and dnd source.
		 */
		cui_action_area_new_for_mode(window, background, 
			fn, 
	   		line->mode->begin, 
			line->mode->end);
		cui_dnd_source_new_with_file(window, background, TargetFileForMode, 
			fn, line->mode->begin, line->mode->end);
		/*
		 * Size, or device number action areas.
		 */
		if (line->line_type != DeviceLine){
			cui_action_area_new_for_size(window, background, 
				fn, 
	  	 		line->size->begin, 
				line->size->end);
		} else {
			cui_action_area_new_for_node_numbers(window, background,
				fn,
					line->size->begin,
				line->size->end);
		}
		/*
		 * Date timestamp action area.
		 */
		cui_action_area_new_for_moddate(window, background,
			fn,
			line->timestamp->begin,
			line->timestamp->end);
	   	/* 
		 * User menu and owner dnd source
		 */
		cui_action_area_new_for_owner(window, background,
			fn,
			line->uname->begin,
			line->uname->end);
	   	widget = cui_user_menu_new(line->uname->text);
	   	cui_widget_set_background(widget, background);
	   	cui_widget_set_area_by_marks(widget, 
			line->uname->begin, 
			line->uname->end);
	   	/*
		 * Filename entry. 
		 */
	   	cui_entry_new_for_filename(window, background, fn, 
	   		line->filename->begin, 
			line->filename->end, 
			line->filename->text);
	   	/* 
		 * Timestamp entry. 
		 */
	   	widget = cui_entry_new();
	   	cui_widget_set_background(widget, background);
	   	cui_widget_set_area_by_marks(widget, 
			line->timestamp->begin, 
			line->timestamp->end);
	   	cui_entry_set_default_text(CUI_ENTRY(widget), 
			line->timestamp->text);
	   	/* 
		 * Owner entry.
		 */
	   	widget = cui_entry_new();
	   	cui_widget_set_background(widget, background);
	   	cui_widget_set_area_by_marks(widget, line->uname->begin, line->uname->end);
	   	cui_entry_set_default_text(CUI_ENTRY(widget), line->uname->text);
	   	cui_widget_append_file_command(widget, FileChangeOwner, fn);
		cui_widget_append_window_command(widget, WindowRefresh, window);
		cui_action_area_new_for_group(window, background,
			fn,
			line->gname->begin,
			line->gname->end);
		cui_dnd_source_new_with_file(window, background, TargetFileForOwner, fn,
			line->uname->begin,
			line->gname->end);
	
	

	   	/*
		 * Group entry and dnd source
		 */
	   	widget = cui_entry_new();
	   	cui_widget_set_background(widget, background);
	   	cui_widget_set_area_by_marks(widget, line->gname->begin, line->gname->end);
	   	cui_entry_set_default_text(CUI_ENTRY(widget), line->gname->text);
	   	cui_widget_append_file_command(widget, FileChangeGroup, fn);
		cui_widget_append_window_command(widget, WindowRefresh, window);
		cui_dnd_source_new_with_file(window, background, TargetFileForGroupOwner, fn,
			line->gname->begin,
			line->gname->end);
	   	/* 
		 * Mode menu.
		 */
	   	widget = cui_mode_menu_new();
	   	cui_widget_set_background(widget, background);
	   	cui_widget_set_area_by_marks(widget, 
			line->mode->begin, 
			line->mode->end);
	   	cui_mode_menu_set_filename(CUI_MODE_MENU(widget), fn);
		cui_widget_append_window_command(widget, WindowRefresh, window);
		/* 
		 * Colored sizebar. 
		 */
		if (line->line_type != DeviceLine){
			widget = cui_size_bar_new();
	 		cui_widget_set_background(widget, background);
	   		cui_widget_set_area_by_marks(widget, line->mode->begin, line->mode->end);
			
	   		g_object_set(G_OBJECT(widget), 
				"percent", line->size->double_val / full_size,
				NULL);
				
			// FIXME: we should differentiate between the exact size and the human readable format.
			cui_filename_set_file_size(fn, (gint64)line->size->double_val);
	 		change = cui_filename_get_size_change(fn);
			if (change == FileSizeIncreasing)
				g_object_set(G_OBJECT(widget),
			   		"decoration", DecorIncrease,
					NULL);
			if (change == FileSizeDecreasing)
				g_object_set(G_OBJECT(widget),
		   			"decoration", DecorDecrease,
					NULL);
		}

		/********************************************************
		 * Drag&Drop target area.                               *
		 ********************************************************/
		GtkTargetList *tlist;
		widget = cui_rectangle_new();
	   	cui_widget_set_background(widget, background);
	   	cui_widget_set_area_by_marks(widget, 
			line->mode->begin, 
			line->filename->end);
		cui_widget_set_width_by_percent(widget, 1.00);
		if (cui_filename_is_dir(fn))
			tlist = dnd_get_dir_target_list();
		else
			tlist = dnd_get_file_target_list();
		cui_rectangle_set_target_list(CUI_RECTANGLE(widget), tlist);
		cui_rectangle_set_target_object(CUI_RECTANGLE(widget), 
			G_OBJECT(fn));
		cui_rectangle_set_is_dnd_target(CUI_RECTANGLE(widget), TRUE);
	        cui_widget_append_window_command(widget, WindowRefresh, window);
		
		/*
		 * DnD source for filename.
		 */
		cui_dnd_source_new_with_file(window, background, TargetFileForName, fn,
			line->filename->begin,
			line->filename->end);

		
		if (line->origin != NULL){
			/* 
			 * Filename becomes origin of the link.
			 */
			fn = cui_filename_new(line->origin->text, window->wd->str); 
			/*
			 * Filename action area for the origin.
			 */
			cui_action_area_new_for_filename(
				window,
				background, 
				fn,
				line->origin->begin, 
				line->origin->end);
			/*
			 * Filename menu for the origin.
			 */
			cui_file_menu_new_for_filename(window, background, fn,
				line->origin->begin, 
				line->origin->end);
			/* 
			 * Filename syntax highlight for the origin. 
			 */
			cui_syntax_highlight_apply_filename(background,
				fn,
				line->origin->begin,
				line->origin->end);
			/********************************************************
			 * Drag&Drop target and source area.                    *
			 ********************************************************/
			GtkTargetList *tlist;
			widget = cui_rectangle_new();
			cui_widget_set_background(widget, background);
			cui_widget_set_area_by_marks(widget, 
				line->origin->begin, 
				line->origin->end);
			cui_widget_set_width_by_percent(widget, 1.00);
			if (cui_filename_is_dir(fn))
				tlist = dnd_get_dir_target_list();
			else
				tlist = dnd_get_file_target_list();
			cui_rectangle_set_target_list(CUI_RECTANGLE(widget), tlist);
			cui_rectangle_set_target_object(CUI_RECTANGLE(widget), 
				G_OBJECT(fn));
			cui_rectangle_set_is_dnd_target(CUI_RECTANGLE(widget), TRUE);
			cui_widget_append_window_command(widget, WindowRefresh, window);

			/* DnD source for filename */	
			widget = cui_rectangle_new();
			cui_widget_set_background(widget, background);
			cui_widget_set_area_by_marks(widget, 
			line->filename->begin, 
			line->filename->end);
			cui_rectangle_set_dnd_source_data(CUI_RECTANGLE(widget),
			TargetFileForName, fn, sizeof(CuiFilename));
			cui_rectangle_set_is_dnd_source(CUI_RECTANGLE(widget), TRUE);
		}
next_line:
		le = le->next;
	}
}

%}

%parse-param {CuiWindow *window}
%union { 
	CuiLexElement sval;
	Direntry      eval;	
}
 
%type  <eval> filename
%token <sval> STRING DIRNAME FILENAME LINKNAME
%token <sval> L_TOTAL L_MODE L_NUMBER L_DNUMBERS L_STRING L_TIMESTAMP L_ARROW 
%type  <sval> uname gname 
%token <sval> ERROR_MESSAGE
%token        END_OF_WINDOW

/* 
 * This grammar describes the language the 'ls' command uses to print out file
 * information. It is very important, I'm working on it...
 */
%%
lines: lines line { /*g_message("%s(): lines line", __func__); */}
     | line       { /*g_message("%s(): lines line", __func__); */}
     ;

line: shortline
    | longline
    | ERROR_MESSAGE 	{ 
    		ls_place_error_message(window, $1.text, $1.begin, $1.end); 
		window->error_code++;
	}
    | END_OF_WINDOW {
		Background *bg;
		if (list != NULL){
			process_long_lines(window_for_lex);
			finalize_parser();
			bg = cui_widget_get_background(CUI_WIDGET(window));
			if (background_check_stored_command(bg, CommandNewFile) ||
					background_check_stored_command(bg, CommandNewDirectory)){
						if (background_get_stored_widget(bg) != NULL)
							cui_entry_realize(CUI_ENTRY(background_get_stored_widget(bg)));
						background_store_command(bg, CommandNone);
					}
		}
		return 0;
	}
    | error { window->error_code++; } 
    ;

longline: L_TOTAL
	| L_MODE L_NUMBER uname gname L_NUMBER L_TIMESTAMP L_STRING {
				append_long_line(&$1, &$2, &$3, &$4, &$5, &$6, &$7, NULL, UsualLine);
		}
	| L_MODE L_NUMBER uname gname L_NUMBER L_TIMESTAMP L_STRING L_ARROW L_STRING {
				append_long_line(&$1, &$2, &$3, &$4, &$5, &$6, &$7, &$9, UsualLine);
		}
	| L_MODE L_NUMBER uname gname L_DNUMBERS L_TIMESTAMP L_STRING {
				append_long_line(&$1, &$2, &$3, &$4, &$5, &$6, &$7, NULL, DeviceLine);
		}
	;

uname:	  L_STRING { $$ = $1; }
     |    L_NUMBER { $$ = $1; g_warning("UID for username."); } 
     ;

gname:	  L_STRING { $$ = $1; }
     |    L_NUMBER { $$ = $1; g_warning("GID for username."); } 
     ;

shortline: filename 		{ }
         | shortline filename	{ }
         ;

filename: STRING	{  
			CuiFilename *fn;
			Background *background;
			background =
				cui_widget_get_background(CUI_WIDGET(window));
		   	fn = cui_filename_new($1.text, window->wd->str); 
	   		/* Action area */
			cui_action_area_new_for_filename(window, background, 
				fn, $1.begin, $1.end);
			/* File menu */
		   	cui_file_menu_new_for_filename(window, background, 
				fn, $1.begin, $1.end);
	   		/* Filename entry */
			cui_entry_new_for_filename(window, background, 
				fn, $1.begin, $1.end, $1.text);
	   		/* Filename syntax highlight */
			cui_syntax_highlight_apply_filename(background,
	   			fn,
				$1.begin,
				$1.end);
			}
        | DIRNAME       {
			CuiFilename *fn;
			Background *background;
			background =
				cui_widget_get_background(CUI_WIDGET(window));
		   	fn = cui_filename_new($1.text, window->wd->str); 
			cui_filename_set_filetype_dir(fn);
	   		/* Action area */
	   		cui_action_area_new_for_filename(
				window, background, 
				fn, $1.begin, $1.end);
			/* File menu */
		   	cui_file_menu_new_for_filename(window, background, 
				fn, $1.begin, $1.end);
	   		/* Filename entry */
			cui_entry_new_for_filename(window, background, 
				fn, $1.begin, $1.end, $1.text);
	   		/* Filename syntax highlight */
			cui_syntax_highlight_apply_filename(background,
	   			fn,
				$1.begin,
				$1.end);
			}
        | FILENAME	{
			CuiFilename *fn;
			Background *background;
			background =
				cui_widget_get_background(CUI_WIDGET(window));
		   	fn = cui_filename_new($1.text, window->wd->str); 
			cui_filename_set_filetype_regular(fn);
	   		/* Action area */
	   		cui_action_area_new_for_filename( window, background, 
				fn, $1.begin, $1.end);
			/* File menu */
		   	cui_file_menu_new_for_filename(window, background, 
				fn, $1.begin, $1.end);
	   		/* Filename entry */
			cui_entry_new_for_filename(window, background, 
				fn, $1.begin, $1.end, $1.text);
	   		/* Filename syntax highlight */
			cui_syntax_highlight_apply_filename(background,
	   			fn,
				$1.begin,
				$1.end);
			}
        | LINKNAME	{
			CuiFilename *fn;
			Background *background;
			background =
				cui_widget_get_background(CUI_WIDGET(window));
		   	fn = cui_filename_new($1.text, window->wd->str); 
			cui_filename_set_filetype_link(fn);
	   		/* Action area */
	   		cui_action_area_new_for_filename(window, background, 
				fn, $1.begin, $1.end);
			/* File menu */
		   	cui_file_menu_new_for_filename(window, background, 
				fn, $1.begin, $1.end);
	   		/* Filename entry */
			cui_entry_new_for_filename(window, background, 
				fn, $1.begin, $1.end, $1.text);
	   		/* Filename syntax highlight */
			cui_syntax_highlight_apply_filename(background,
	   			fn,
				$1.begin,
				$1.end);
			}
        ;
%%

extern int ls_lineno;

int ls_error(CuiWindow *window, char *c )
{
	gchar *error_message;
	
	error_message = g_strdup_printf(
		"Directory Contents Error: %s in line %d", 
		c,
		ls_lineno);
	g_warning(error_message);
	cui_window_set_title(window, error_message);
	g_free(error_message);
	return 1;
}

int ls_wrap(void)
{
	return 1;
}
