/*
 * Copyright (C) 2004 Laszlo Pere
 * $Id: widgets.c,v 1.1 2006/02/24 17:14:15 pipas Exp pipas $
 *
 * This is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Library General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Log: widgets.c,v $
 * Revision 1.1  2006/02/24 17:14:15  pipas
 * Initial revision
 *
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gnome.h>

#include "cuiterm.h"
#include "icons.h"
#include "widgets.h"
#include "cuiwindow.h"
#include "interface.h"
#include "callbacks.h"
#include "interface.h"
#include "support.h"
#include "configure.h"
#include "popup_menus.h"
#if DEBUG_ALL
#  define DEBUG 1
#endif

/*****************************************************************************
 * CuiWidget                                                                 *
 *                                                                           *
 *****************************************************************************/

#define PROPERTY_NAME 1
#define PROPERTY_ACTIVE 2
#define PROPERTY_TOOLTIP 3
#define PROPERTY_DIRECTORY 4
#define PROPERTY_AREA 5
#define PROPERTY_BACKGROUND 6
#define PROPERTY_APPBAR 7
#define PROPERTY_TERMINAL 8
#define PROPERTY_ASTYLE 9

static void
cui_widget_set_property(GObject *object, 
		guint           property_id, 
		const GValue    *value, 
		GParamSpec      *pspec);
static void
cui_widget_get_property(GObject *object, 
		guint            property_id, 
		GValue           *value, 
		GParamSpec       *pspec);

/*
 * Dispose and finalize functions for the CuiWidgets
 */
static void 
cui_widget_dispose(GObject * obj)
{
	CuiWidget *self = (CuiWidget *) obj;
#if 0
	if (self->private->dispose_has_run) {
		return;
	}
	
	object->private->dispose_has_run = TRUE;
#endif
	//G_OBJECT_CLASS(GObjectClass)->dispose(obj);
}

static void 
cui_widget_finalize(GObject * obj)
{
	CuiWidget *self = (CuiWidget *) obj;
	GList *l;
	
	//g_message("%s(): start", __func__);
	
	/*
	 * Freeing some strings.
	 */
	g_free(self->name);
	g_free(self->area);
	g_free(self->groupname);
	
	if (self->begin_mark != NULL && self->background != NULL &&
			!gtk_text_mark_get_deleted(self->begin_mark))
		gtk_text_buffer_delete_mark(self->background->text_buffer, 
				self->begin_mark);
	if (self->end_mark != NULL && self->background != NULL &&
			!gtk_text_mark_get_deleted(self->end_mark))
		gtk_text_buffer_delete_mark(self->background->text_buffer, 
				self->end_mark);
	/*
	 * Freeing the command list.
	 */
	l = self->commands;
	while (l != NULL){
		g_free(l->data);
		l = l->next;
	}
	g_list_free(self->commands);
	
	//G_OBJECT_CLASS(GObjectClass)->finalize(obj);
}


static void 
cui_widget_class_init(
		gpointer g_class, 
		gpointer g_class_data) 
{
	GObjectClass   *gobject_class = G_OBJECT_CLASS(g_class);
	CuiWidgetClass *klass = CUI_WIDGET_CLASS(g_class);
	GParamSpec     *pspec;
	/*
	 * Dispose and finalize functions.
	 */
	gobject_class->dispose = cui_widget_dispose;
	gobject_class->finalize = cui_widget_finalize;
	
	/*
	 * Default values for this class.
	 */
	klass->default_inactive_cursor = gdk_cursor_new(GDK_TOP_LEFT_ARROW);
	klass->default_active_cursor = gdk_cursor_new(GDK_HAND2);
	/*
	 * Setting and getting properties.
	 */
	gobject_class->set_property = cui_widget_set_property;
  	gobject_class->get_property = cui_widget_get_property;

	/*
	 * Properties.
	 */
	pspec = g_param_spec_string (
			"name", "name",
                        "The name of this widget instance.",
                        "unknown name",
                        G_PARAM_READWRITE);
	g_object_class_install_property(gobject_class, PROPERTY_NAME, pspec);

	pspec = g_param_spec_boolean (
			"active", "widget-active",
                        "If this widget is shown and working.",
                        TRUE,
                        G_PARAM_READWRITE);
	g_object_class_install_property(gobject_class, PROPERTY_ACTIVE, pspec);
	
	pspec = g_param_spec_string (
			"tooltip", "tooltip-text",
                        "Some descriptive text.",
                        "",
                        G_PARAM_READWRITE);
	g_object_class_install_property(gobject_class, PROPERTY_TOOLTIP, pspec);
	
	pspec = g_param_spec_string (
			"active-style", "active_style",
                        "Text style for the active state if any.",
                        "",
                        G_PARAM_READWRITE);
	g_object_class_install_property(gobject_class, PROPERTY_ASTYLE, pspec);
	
	pspec = g_param_spec_string (
			"directory", "working-directory",
                        "The working directory of the creator command.",
                        "",
                        G_PARAM_READWRITE);
	g_object_class_install_property(gobject_class, 
			PROPERTY_DIRECTORY, 
			pspec);
	
	pspec = g_param_spec_pointer(
			"background", "background-pointer",
                        "The background containing this widget.",
                        G_PARAM_READWRITE);
	g_object_class_install_property(gobject_class, 
			PROPERTY_BACKGROUND, 
			pspec);
	
	pspec = g_param_spec_pointer(
			"terminal", "terminal-pointer",
                        "The terminal we send the commands to.",
                        G_PARAM_READWRITE);
	g_object_class_install_property(gobject_class, 
			PROPERTY_TERMINAL, 
			pspec);

	pspec = g_param_spec_pointer(
			"appbar", "appbar-widget",
                        "The GtkWidget to send the tooltip information.",
                        G_PARAM_READWRITE);
	g_object_class_install_property(gobject_class, 
			PROPERTY_APPBAR, 
			pspec);
	
	pspec = g_param_spec_pointer(
			"area", "activation-area",
                        "The place and the size of the widget.",
                        G_PARAM_READWRITE);
	g_object_class_install_property(gobject_class, 
			PROPERTY_AREA, 
			pspec);
}

static void
cui_widget_base_class_init(CuiWidgetClass *klass){
}


static void
cui_widget_instance_init(
		GTypeInstance   *instance, 
		gpointer         g_class){
	CuiWidget *self = (CuiWidget *)instance;
	CuiWidgetClass *klass = (CuiWidgetClass *)g_class;
	
	/*
	 * Setting the defaults from the class structure.
	 */
	self->mouse_cursor_active = klass->default_active_cursor;
	self->mouse_cursor_inactive = klass->default_inactive_cursor;
	
	self->name = NULL;
	self->active = TRUE;
	self->tooltip = NULL;
	self->directory = NULL;
	self->area = NULL;
	self->background = NULL;
	self->terminal = NULL;
	self->appbar = NULL;
	self->commands = NULL;
	self->groupname = NULL;
	self->visible = TRUE;
	self->mouse_is_over = FALSE;
	self->begin_mark = NULL;
	self->end_mark = NULL;
	self->active_style = NULL;
}

static void
cui_widget_set_property(GObject *object, 
		guint           property_id, 
		const GValue    *value, 
		GParamSpec      *pspec)
{
	CuiWidget *self = (CuiWidget *) object;
	#ifdef DEBUG_WIDGETS	
	g_message("%s(): start.", __func__);
	#endif
	switch(property_id){
		case PROPERTY_NAME:
			g_free(self->name);
			self->name = g_value_dup_string(value);
			break;
		case PROPERTY_TOOLTIP:
			g_free(self->tooltip);
			self->tooltip = g_value_dup_string(value);
			break;
		case PROPERTY_ASTYLE:
			g_free(self->active_style);
			self->active_style = g_value_dup_string(value);
			break;
		case PROPERTY_DIRECTORY:
			g_free(self->directory);
			self->directory = g_value_dup_string(value);
			break;
		case PROPERTY_BACKGROUND:
			self->background = g_value_get_pointer(value);
			break;
		case PROPERTY_TERMINAL:
			self->terminal = g_value_get_pointer(value);
			break;
		case PROPERTY_AREA:
			g_free(self->area);
			self->area = g_memdup(g_value_get_pointer(value),
					sizeof(GdkRectangle));
			break;
		case PROPERTY_ACTIVE:
			self->active = g_value_get_boolean(value);
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID(
					object, property_id, pspec);
	}
}


static void
cui_widget_get_property(GObject *object, 
		guint            property_id, 
		GValue           *value, 
		GParamSpec       *pspec)
{
	CuiWidget *self = (CuiWidget *) object;
	
	#ifdef DEBUG_WIDGETS	
	g_message("%s(): start: %d", __func__, property_id);
	#endif
	switch(property_id){
		case PROPERTY_NAME:
			g_value_set_string(value, self->name);
			break;
		case PROPERTY_TOOLTIP:
			g_value_set_string(value, self->tooltip);
			break;
		case PROPERTY_ASTYLE:
			g_value_set_string(value, self->active_style);
			break;
		case PROPERTY_DIRECTORY:
			g_value_set_string(value, self->directory);
			break;
		case PROPERTY_BACKGROUND:
			g_value_set_pointer(value, self->background);
			break;
		case PROPERTY_TERMINAL:
			g_value_set_pointer(value, self->terminal);
			break;
		case PROPERTY_AREA:
			g_value_set_pointer(value, self->area);
			break;
		case PROPERTY_ACTIVE:
			g_value_set_boolean(value, self->active);
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID(
					object, property_id, pspec);
	}
	#ifdef DEBUG_WIDGETS	
	g_message("%s(): end", __func__);
	#endif
}


GType 
cui_widget_get_type(void) 
{
	static GType type = 0;
	
	#ifdef DEBUG_WIDGETS	
	g_message("%s(): start.", __func__);
	#endif
	if (type == 0) {
		static const GTypeInfo info = {
			sizeof(CuiWidgetClass),
			(GBaseInitFunc) cui_widget_base_class_init, 
			(GBaseFinalizeFunc) NULL,                  
			(GClassInitFunc) cui_widget_class_init,   
			NULL,
			NULL,
			sizeof(CuiWidget),
			0,
			(GInstanceInitFunc) cui_widget_instance_init 
		};
		type = g_type_register_static(G_TYPE_OBJECT, 
				"CuiWidget", &info, 0);
	}
	return type;
}

/**
 * cui_widget_set_name:
 * @widget: a #CuiWidget
 * @name: the name of the widget
 *
 * Sets the name for this widget instance. The name appears in messages, help
 * texts and several other places.
 */
void
cui_widget_set_name(CuiWidget *widget, const gchar *name){
	GValue val = {0,};
	g_return_if_fail(CUI_IS_WIDGET(widget));
	g_value_init(&val, G_TYPE_STRING);
	g_value_set_string (&val, name);
	g_object_set_property (G_OBJECT (widget), "name", &val);
}

/**
 * cui_widget_get_name:
 * @widget: a #CuiWidget
 * @returns: a newly allocated string
 *
 * Gets the name of this widget instance. The returned string 
 * should be freed.
 */
gchar *
cui_widget_get_name(CuiWidget *widget)
{
	gchar *name;
	
	#ifdef DEBUG_WIDGETS	
	g_message("%s(): start.", __func__);
	#endif
	
	g_return_if_fail(CUI_IS_WIDGET(widget));
	g_object_get(G_OBJECT(widget), "name", &name, NULL);
	return name;
}

/**
 * cui_widget_get_active:
 * @widget: a #CuiWidget
 * @returns: TRUE if the widget can be activated
 *
 * Gets if the widget can be activated by the mouse.
 */
gboolean
cui_widget_get_active(CuiWidget *widget)
{	
	#ifdef DEBUG_WIDGETS	
	g_message("%s(): start.", __func__);
	#endif

	g_return_if_fail(CUI_IS_WIDGET(widget));	
	return widget->active;
}

/**
 * cui_widget_set_active:
 * @widget: a #CuiWidget
 * @active: sets if the widget is an active one
 *
 * Sets is the widget can be activated by the mouse.
 */
void
cui_widget_set_active(CuiWidget *widget, gboolean active)
{
	#ifdef DEBUG_WIDGETS	
	g_message("%s(): start.", __func__);
	#endif
	
	g_return_if_fail(CUI_IS_WIDGET(widget));
	widget->active = active;
	g_object_notify(G_OBJECT(widget), "active");
}

/**
 * cui_widget_set_tooltip:
 * @widget: a #CuiWidget
 * @tooltip: the tooltip text of the widget
 *
 * Sets the tooltip text for this widget instance. The tooltip text is a short
 * remainder appears in a different #GtkWidget.
 */
void
cui_widget_set_tooltip(CuiWidget *widget, const gchar *tooltip)
{
	GValue val = {0,};
	
	#ifdef DEBUG_WIDGETS	
	g_message("%s(): start.", __func__);
	#endif
	g_return_if_fail(CUI_IS_WIDGET(widget));
	g_value_init(&val, G_TYPE_STRING);
	g_value_set_string (&val, tooltip);
	g_object_set_property (G_OBJECT (widget), "tooltip", &val);
}

/**
 * cui_widget_set_active_style:
 * @widget: a #CuiWidget
 * @style: the active style text of the widget
 *
 * Sets the active style name text for this widget instance. The active style is
 * a text style used when the mouse is over the widget.
 */
void
cui_widget_set_active_style(CuiWidget *widget, const gchar *style)
{
	GValue val = {0,};
	
	#ifdef DEBUG_WIDGETS	
	g_message("%s(): start.", __func__);
	#endif
	g_return_if_fail(CUI_IS_WIDGET(widget));
	g_value_init(&val, G_TYPE_STRING);
	g_value_set_string (&val, style);
	g_object_set_property (G_OBJECT (widget), "active_style", &val);
}

/**
 * cui_widget_add_tag_by_name:
 * @widget: a #CuiWidget widget
 * @tag: the name of the tag to apply
 *
 * Convinient utility function to highlight the text belongs to the widget.
 */
void
cui_widget_add_tag_by_name(CuiWidget *widget, const gchar *tag)
{
	GtkTextIter begin, end;
	g_return_if_fail(CUI_IS_WIDGET(widget));
	g_return_if_fail(tag != NULL);
	g_return_if_fail(widget->background != NULL);
	
	if (widget->begin_mark == NULL || widget->end_mark == NULL)
		return;
	
	gtk_text_buffer_get_iter_at_mark(widget->background->text_buffer,
				&begin, widget->begin_mark);
	gtk_text_buffer_get_iter_at_mark(widget->background->text_buffer,
				&end, widget->end_mark);
	gtk_text_buffer_apply_tag_by_name(widget->background->text_buffer, 
				tag, &begin, &end);
}
	

/**
 * cui_widget_set_directory:
 * @widget: a #CuiWidget
 * @directory: the working directory of the widget's creator
 *
 * Sets the directory for this widget instance. The directory shows what was the
 * working directory when this widget has been created.
 */
void
cui_widget_set_directory(CuiWidget *widget, const gchar *directory)
{
	GValue val = {0,};
	#ifdef DEBUG_WIDGETS	
	g_message("%s(): start.", __func__);
	#endif
	
	g_return_if_fail(CUI_IS_WIDGET(widget));
	g_value_init(&val, G_TYPE_STRING);
	g_value_set_string (&val, directory);
	g_object_set_property (G_OBJECT (widget), "directory", &val);
}

/**
 * cui_widget_get_directory:
 * @widget: a #CuiWidget
 * @returns: a newly allocated string
 *
 * Gets the working directory of this widget instance.
 */
gchar *
cui_widget_get_directory(CuiWidget *widget)
{
	gchar *directory;
	
	#ifdef DEBUG_WIDGETS	
	g_message("%s(): start.", __func__);
	#endif
	g_return_if_fail(CUI_IS_WIDGET(widget));
	g_object_get(G_OBJECT(widget), "directory", &directory, NULL);
	return directory;
}

/**
 * cui_widget_set_background:
 * @widget: a #CuiWidget
 * @background: the background this widget belongs to
 *
 * Sets the background for this widget instance. The background is the place
 * where we show the widget. This function also registers the widget calling
 * #cui_background_append_widget.
 */
void
cui_widget_set_background(CuiWidget *widget, 
		const Background *background)
{
	GValue val = {0,};
	Terminal *terminal;
	
	#ifdef DEBUG_WIDGETS	
	g_message("%s(): start.", __func__);
	#endif
	g_return_if_fail(CUI_IS_WIDGET(widget));
	/*
	 * Setting the "background" property.
	 */
	g_value_init(&val, G_TYPE_POINTER);
	g_value_set_pointer (&val, (gpointer)background);
	g_object_set_property (G_OBJECT (widget), "background", &val);
	/*
	 * Registering the new widget as the new part of the background.
	 */
	cui_background_append_widget(g_value_get_pointer(&val), widget);
	/*
	 * Setting the "terminal" property.
	 */
	terminal = g_object_get_data(G_OBJECT(background->text_view),
			"terminal");
	g_assert(terminal != NULL);
	g_value_set_pointer (&val, (gpointer)terminal);
	g_object_set_property (G_OBJECT (widget), "terminal", &val);
}

/**
 * cui_widget_set_terminal:
 * @widget: a #CuiWidget
 * @terminal: the terminal we use to send the commands to the shell
 *
 * Sets the terminal for this widget instance. The terminal is a #Terminal
 * object we send the commands to execute.
 */
void
cui_widget_set_terminal(CuiWidget *widget, const Terminal *terminal)
{
	GValue val = {0,};
	
	#ifdef DEBUG_WIDGETS	
	g_message("%s(): start.", __func__);
	#endif
	g_return_if_fail(CUI_IS_WIDGET(widget));
	g_value_init(&val, G_TYPE_POINTER);
	g_value_set_pointer (&val, (gpointer)terminal);
	g_object_set_property (G_OBJECT (widget), "terminal", &val);
}

/**
 * cui_widget_get_terminal:
 * @widget: a #CuiWidget
 * @returns: a pointer to the background 
 *
 * Gets the terminal of this widget. The terminal is a #Terminal object we use
 * to send the command to the shell.
 */
Terminal *
cui_widget_get_terminal(CuiWidget *widget){
	Terminal *terminal;
	
	#ifdef DEBUG_WIDGETS	
	g_message("%s(): start.", __func__);
	#endif
	g_return_if_fail(CUI_IS_WIDGET(widget));
	g_object_get(G_OBJECT(widget), "terminal", &terminal, NULL);
	return terminal;
}

/**
 * cui_widget_set_area:
 * @widget: a #CuiWidget
 * @rectangle: the rectangle of the activation area for this widget
 *
 * Sets the activation area for this widget instance. This function will copy
 * the #GdkRectangle to an allocated memory, so it can be freed.
 */
void
cui_widget_set_area(CuiWidget *widget, 
		const GdkRectangle *rectangle)
{
	GValue val = {0,};
	
	#ifdef DEBUG_WIDGETS	
	g_message("%s(): start.", __func__);
	#endif
	g_return_if_fail(CUI_IS_WIDGET(widget));
	g_return_if_fail(rectangle != NULL);
	
	g_value_init(&val, G_TYPE_POINTER);
	g_value_set_pointer (&val, (gpointer)rectangle);
	g_object_set_property (G_OBJECT (widget), "area", &val);
}

/**
 * cui_widget_move:
 * @widget: the widget to move
 * @dx: movement on the x axis
 * @dy: movement on the y axis
 *
 * Moves the widget by modifying its coordinates.
 */
void 
cui_widget_move(CuiWidget *widget, gint dx, gint dy)
{
	#ifdef DEBUG_WIDGETS	
	g_message("%s(): start.", __func__);
	#endif
	g_return_if_fail(CUI_IS_WIDGET(widget));
	g_return_if_fail(widget->area != NULL);
	if (dx >= 0)
		widget->area->x = dx;
	if (dy >= 0)
		widget->area->y = dy;
}

/**
 * cui_widget_move_rel:
 * @widget: the widget to move
 * @dx: movement on the x axis
 * @dy: movement on the y axis
 *
 * Moves the widget by modifying its coordinates.
 */
void 
cui_widget_move_rel(CuiWidget *widget, 
		gint dx, 
		gint dy)
{
	#ifdef DEBUG_WIDGETS	
	g_message("%s(): start.", __func__);
	#endif
	g_return_if_fail(CUI_IS_WIDGET(widget));
	g_return_if_fail(widget->area != NULL);
	if (dx != 0)
		widget->area->x += dx;
	if (dy != 0)
		widget->area->y += dy;
}

/**
 * cui_widget_set_area_by_coord:
 * @widget: a #CuiWidget
 * @x: the x coordinate 
 * @y: the y coordinate 
 * @dx: the width 
 * @dy: the height
 *
 * Sets the activation area for this widget instance.
 */
void
cui_widget_set_area_by_coord(
		CuiWidget *widget, 
		gint x, gint y,
		gint dx, gint dy){
	GdkRectangle rec;
	GValue val = {0,};
	
	#ifdef DEBUG_WIDGETS	
	g_message("%s(): start.", __func__);
	#endif
	g_return_if_fail(CUI_IS_WIDGET(widget));
	
	rec.x = x;
	rec.y = y;
	rec.width = dx;
	rec.height = dy;
	g_value_init(&val, G_TYPE_POINTER);
	g_value_set_pointer (&val, &rec);
	g_object_set_property (G_OBJECT (widget), "area", &val);
}

/**
 * cui_widget_set_width_by_percent:
 * @widget: a #CuiWidget
 * @percent: the new width of the widget as a fraction of 100
 *
 * Sets the width of the widget as the fraction of the background text window.
 */
void
cui_widget_set_width_by_percent(
		CuiWidget *widget, 
		gdouble percent)
{
	GdkWindow *window;
	gint width;
	gint max;
	
	#ifdef DEBUG_WIDGETS	
	g_message("%s(): start.", __func__);
	g_message("percent: %f", percent);
	#endif
	
	g_return_if_fail(CUI_IS_WIDGET(widget));
	g_return_if_fail(widget->area != NULL && widget->background != NULL);
 
	window = gtk_text_view_get_window(widget->background->text_view,
			GTK_TEXT_WINDOW_WIDGET);
	gdk_drawable_get_size(GDK_DRAWABLE(window), &width, NULL);
	
	max = (width - LEFT_MARGIN - 9) - (widget->area->x - LEFT_MARGIN);
	width = percent * max;
	if (width <= 0)
		width = 2;
	widget->area->width = MIN(max, width);
}

/**
 * cui_widget_set_start_and_stop_by_percent:
 * @widget: a #CuiWidget
 * @percent: the new width of the widget as a fraction of 100
 *
 * Sets the left and right coordinates of the widget as the fraction of the 
 * background text window.
 */
void
cui_widget_set_start_and_stop_by_percent(
		CuiWidget *widget, 
		gdouble start,
		gdouble stop)
{
	gdouble    percent = stop - start;
	GdkWindow *window;
	gint       left, width;
	
	#ifdef DEBUG_WIDGETS	
	g_message("%s(): start.", __func__);
	#endif
	g_return_if_fail(CUI_IS_WIDGET(widget));
	g_return_if_fail(widget->area != NULL && widget->background != NULL);
 
	window = gtk_text_view_get_window(widget->background->text_view,
			GTK_TEXT_WINDOW_WIDGET);
	gdk_drawable_get_size(GDK_DRAWABLE(window), &width, NULL);

	left = LEFT_MARGIN + start * (width - LEFT_MARGIN - 9) - (widget->area->x - LEFT_MARGIN);
	width = 1 + percent * (width - LEFT_MARGIN - 9) - (widget->area->x - LEFT_MARGIN);
	if (width == 0)
		width = 2;
	
	#ifdef DEBUG_WIDGETS	
	g_message("%s(): left=%d, width=%d", __func__, left, width);	
	#endif
	
	widget->area->x = MAX(0, left);
	widget->area->width = MAX(0, width);
}

/**
 * cui_widget_set_area_by_marks:
 * @widget: a #CuiWidget
 * @begin: the #GtkTextMark of the upper left corner
 * @end: the #GtkTextMark of the lower right corner
 *
 * Sets the activation area for this widget instance. This function copy the
 * marks so they can be eliminated.
 */
void
cui_widget_set_area_by_marks(
		CuiWidget *widget, 
		GtkTextMark *begin,
		GtkTextMark *end)
{
	GtkTextIter biter, eiter;
	GdkRectangle b_rec, e_rec;
	GdkRectangle rec;
	GValue val = {0,};
	
	#ifdef DEBUG_WIDGETS	
	g_message("%s(): start.", __func__);
	#endif
	g_return_if_fail(CUI_IS_WIDGET(widget));
	g_return_if_fail(begin != NULL && end != NULL);
	/* FIXME: We should not tamper with the background. It would be nice to
	 * call some function instead. */
	g_return_if_fail(widget->background != NULL && 
			widget->background->text_view != NULL);
	/*
	 * Calculating the location and the size of the activation area.
	 */
	gtk_text_buffer_get_iter_at_mark(widget->background->text_buffer, 
			&biter, begin);
	gtk_text_buffer_get_iter_at_mark(widget->background->text_buffer, 
			&eiter, end);
	gtk_text_view_get_iter_location(widget->background->text_view, 
			&biter, &b_rec);
	gtk_text_view_get_iter_location(widget->background->text_view, 
			&eiter, &e_rec);
	gdk_rectangle_union(&b_rec, &e_rec, &rec);
	rec.width -= e_rec.width;
	/*
	 * Storing the activation area.
	 */
	g_value_init(&val, G_TYPE_POINTER);
	g_value_set_pointer (&val, &rec);
	g_object_set_property (G_OBJECT (widget), "area", &val);
	/*
	 * Storing the text marks for further uses
	 */
	if (widget->begin_mark != NULL)
		gtk_text_buffer_delete_mark(widget->background->text_buffer,
				widget->begin_mark);
	if (widget->end_mark != NULL)
		gtk_text_buffer_delete_mark(widget->background->text_buffer,
				widget->end_mark);
	widget->begin_mark = gtk_text_buffer_create_mark(
			widget->background->text_buffer,
			NULL, &biter, TRUE);
	widget->end_mark = gtk_text_buffer_create_mark(
			widget->background->text_buffer,
			NULL, &eiter, TRUE);
#ifdef VISIBLE_MARKS
	gtk_text_mark_set_visible(widget->begin_mark, TRUE);
	gtk_text_mark_set_visible(widget->end_mark, TRUE);
#endif
}

/**
 * cui_widget_set_area_by_iter:
 * @widget: a #CuiWidget to set
 * @iter: the #GtkTextIter storing the place of the char
 *
 * Sets the activation area of the widgets using a #GtkTextIter to find a
 * character. This function is usefull to create action areas for one given 
 * character on the user interface.
 *
 */
void 
cui_widget_set_area_by_iter(CuiWidget *widget, 
		GtkTextIter *iter)
{
	GtkTextMark *begin, *end;
	GtkTextIter end_iter;

	g_return_if_fail(CUI_IS_WIDGET(widget) && iter != NULL);
	g_return_if_fail(widget->background != NULL && 
			 widget->background->text_buffer != NULL);

	end_iter = *iter;
	gtk_text_iter_forward_char(&end_iter);
	
	begin = gtk_text_buffer_create_mark(widget->background->text_buffer, 
			NULL, iter, TRUE);
#ifdef VISIBLE_MARKS
	gtk_text_mark_set_visible(begin, TRUE);
#endif

	end   = gtk_text_buffer_create_mark(widget->background->text_buffer, 
			NULL, &end_iter, TRUE);
#ifdef VISIBLE_MARKS
		gtk_text_mark_set_visible(end, TRUE);
#endif

	cui_widget_set_area_by_marks(widget, begin, end);
	
	gtk_text_buffer_delete_mark(widget->background->text_buffer, begin);
	gtk_text_buffer_delete_mark(widget->background->text_buffer, end);
#if 0
	GdkRectangle rec;
	GValue val = {0,};
	
	#ifdef DEBUG_WIDGETS	
	g_message("%s(): start.", __func__);
	#endif
	g_return_if_fail(CUI_IS_WIDGET(widget));
	/* FIXME: We should not tamper with the background. It would be nice to
	 * call some function instead. */
	g_return_if_fail(widget->background != NULL && 
			widget->background->text_view != NULL);
	gtk_text_view_get_iter_location(widget->background->text_view, 
			iter, &rec);
	/*
	 * Storing the activation area.
	 */
	g_value_init(&val, G_TYPE_POINTER);
	g_value_set_pointer (&val, &rec);
	g_object_set_property (G_OBJECT (widget), "area", &val);
#endif
}

/**
 * cui_widget_get_space:
 * @widget: a #CuiWidget
 * @returns: the size of the widget area
 *
 * Returns the size of the widget area in pixels. If the widget has no
 * associated area this function will return F_MAXINT.
 */
gint
cui_widget_get_size(CuiWidget *widget)
{
	#ifdef DEBUG_WIDGETS	
	g_message("%s(): start.", __func__);
	#endif
	g_return_val_if_fail(CUI_IS_WIDGET(widget), G_MAXINT);
	g_return_val_if_fail(widget != NULL, G_MAXINT);
	return widget->area->width * widget->area->height;
}

/**
 * cui_widget_get_area:
 * @widget: a #CuiWidget
 * @returns: the rectangular area of the widget
 *
 * Gets the activation area of this widget. The return value should not be
 * modified or freed.
 */
GdkRectangle *
cui_widget_get_area(CuiWidget *widget)
{
	GdkRectangle *area;
	
	#ifdef DEBUG_WIDGETS	
	g_message("%s(): start.", __func__);
	#endif
	g_return_if_fail(CUI_IS_WIDGET(widget));
	g_object_get(G_OBJECT(widget), "area", &area, NULL);
	return area;
}

/**
 * cui_widget_get_area_window_coords:
 * @widget: a #CuiWidget
 * @returns: the rectangular area of the widget relative to the text_view
 *
 * Gets the activation area of this widget relative to the text_view this
 * widget is in. Note that the #Background of the widget must be valid
 * and the #GtkTextView of the #Background also must be set.
 *
 * The return value of the function is a dinamically allocated area which
 * should be freed.
 */
GdkRectangle *
cui_widget_get_area_window_coords(CuiWidget *widget){
	GtkTextView *text_view;
	GdkRectangle *area;
	GdkRectangle *ret = g_malloc(sizeof(GdkRectangle));
	
	#ifdef DEBUG_WIDGETS	
	g_message("%s(): start.", __func__);
	#endif
	/*
	 * Checking...
	 */
	g_return_val_if_fail(widget != NULL, ret);
	g_return_val_if_fail(widget->background != NULL, ret);
	/* FIXME: This should be found by a function of the background 
	 * class, but that is not a class yet.*/
	g_return_val_if_fail(widget->background->text_view != NULL, ret);
	/*
	 * Calculating...
	 */
	g_object_get(G_OBJECT(widget), "area", &area, NULL);
	g_return_val_if_fail(area != NULL, ret);
	
	gtk_text_view_buffer_to_window_coords(
			widget->background->text_view, 
			GTK_TEXT_WINDOW_TEXT, 
			area->x, 
			area->y, 
			&ret->x, 
			&ret->y);
	ret->height = area->height;
	ret->width = area->width;
	
	return ret;
}

/**
 * cui_widget_get_background:
 * @widget: a #CuiWidget
 * @returns: a pointer to the background 
 *
 * Gets the background this widget instance appears on.
 */
Background *
cui_widget_get_background(CuiWidget *widget)
{
	Background *background;
	
	#ifdef DEBUG_WIDGETS	
	g_message("%s(): start.", __func__);
	#endif
	g_return_val_if_fail(CUI_IS_WIDGET(widget), NULL);
	g_object_get(G_OBJECT(widget), "background", &background, NULL);
	return background;
}

/**
 * cui_widget_set_appbar:
 * @widget: a #CuiWidget
 * @appbar: the application bar we send the tooltip text
 *
 * Sets the appbar for this widget instance. The appbar is a widget 
 * where we send tooltip text when needed.
 */
void
cui_widget_set_appbar(CuiWidget *widget, GtkWidget *appbar)
{	
	#ifdef DEBUG_WIDGETS	
	g_message("%s(): start.", __func__);
	#endif
	g_return_if_fail(CUI_IS_WIDGET(widget));
	widget->appbar = appbar; 
	g_object_notify(G_OBJECT(widget), "appbar");
}


/**
 * cui_widget_is_mouse_over:
 * @widget: a #CuiWidget
 * @x: the x coordinate of the mouse
 * @y: the y coordinate of the mouse
 * @returns: TRUE if the given coordinates are inside the activation area
 *
 * Checks if the mouse is over the given widget, that is the given coordinates
 * are inside the activation area.
 *
 * This function gets x and y coordinate in the widget coordinate system. Use #
 * gtk_text_view_window_to_buffer_coords to convert mouse coordinates to this
 * coordinate system.
 */
gboolean
cui_widget_is_mouse_over(CuiWidget *widget, 
		gint x, 
		gint y){
	
	#ifdef DEBUG	
	g_message("%s(): start: x: %d, y: %d", __func__, x, y);
	#endif
	g_return_val_if_fail(CUI_IS_WIDGET(widget), FALSE);
	
	if (widget->area == NULL) {
#ifdef DEBUG
		g_warning("%s(): area == NULL", __func__);
#endif
		return FALSE;
	}
#if 0
	g_message("%s(): x=%4d y=%4d (%4d, %4d)(%4d, %4d)",
			__func__, x, y,
			widget->area->x, widget->area->y,
			widget->area->width, widget->area->height
			);
	g_message("%s(): name: '%s'", __func__, widget->name);
#endif
	if (x >= widget->area->x && 
			y >= widget->area->y && 
			x <= widget->area->x + widget->area->width && 
			y <= widget->area->y + widget->area->height)
		return TRUE;
	return FALSE;
}

/**
 * cui_widget_mouse_enter:
 * @widget: a #CuiWidget widget
 *
 * This function must be called when the mouse enters the area of the widget.
 * I'm sure there is a better solution, but I have not found yet... yep, found
 * we should use signals of course. FIXME: we really should use signals.
 */
void
cui_widget_mouse_enter(CuiWidget *widget)
{
	GtkTextIter begin, end;
	
	#ifdef DEBUG_WIDGETS	
	g_message("%s(): start.", __func__);
	#endif
	
	g_return_if_fail(CUI_IS_WIDGET(widget));
	
	if(widget->mouse_is_over){
		g_warning("%s(): Mouse enter already registered for %p",
				__func__, widget);
		return;
	}
	
	widget->mouse_is_over = TRUE;

	// Invalidating.
	if (widget->background != NULL && widget->area != NULL) {
		GdkRectangle *rec;
		rec = cui_widget_get_area_window_coords(widget);
		gdk_window_invalidate_rect(widget->background->window, rec, TRUE);
		g_free(rec);
	}

	
	/* deploying tooltip */
	if (widget->tooltip == NULL || widget->appbar == NULL)
		return;
	gnome_appbar_push(GNOME_APPBAR(widget->appbar), widget->tooltip);

	
	/*
	 * If this widget can not be activated by the mouse, we do 
	 * nothing else.
	 */
	if (!widget->active)
		return;
	/*
	 * Now this is a widget the mouse just entered over.
	 */
	g_return_if_fail(widget->background != NULL);
	if (widget->begin_mark != NULL &&
			widget->end_mark != NULL &&
			widget->active_style != NULL)
	{
		gtk_text_buffer_get_iter_at_mark(
				widget->background->text_buffer,
				&begin,
				widget->begin_mark);
		gtk_text_buffer_get_iter_at_mark(
				widget->background->text_buffer,
				&end,
				widget->end_mark);
		gtk_text_buffer_apply_tag_by_name(
				widget->background->text_buffer, 
				widget->active_style, 
				&begin, &end);
	}
#if 0
	/*
	 * I decided not to tamper with the mouse cursor here.
	 */
	/* Setting the mouse cursor */	
	if (widget->background != NULL && widget->mouse_cursor_active != NULL)
		gdk_window_set_cursor(widget->background->window, 
				widget->mouse_cursor_active);
#endif
}

/**
 * cui_widget_mouse_leave:
 * @widget: a #CuiWidget widget
 *
 * This function must be called when the mouse leaves the area of the widget.
 * I'm sure there is a better solution, but I have not found yet... 
 */
void
cui_widget_mouse_leave(CuiWidget *widget)
{
	GtkTextIter begin, end;
	
	#ifdef DEBUG_WIDGETS	
	g_message("%s(): start.", __func__);
	#endif
	g_return_if_fail(CUI_IS_WIDGET(widget));
	if(!widget->mouse_is_over){
		g_warning("%s(): Mouse leave already registered for %p",
				__func__, widget);
		return;
	}
	widget->mouse_is_over = FALSE;

	// Invalidating.
	if (widget->background != NULL && widget->area != NULL) {
		GdkRectangle *rec;
		rec = cui_widget_get_area_window_coords(widget);
		gdk_window_invalidate_rect(widget->background->window, rec, TRUE);
		g_free(rec);
	}
	
	/*
	 * FIXME: we assume this widget has pushed a tooltip text to the appbar
	 * if it would push now.
	 */
	if (widget->tooltip == NULL || widget->appbar == NULL)
		return;
	gnome_appbar_pop(GNOME_APPBAR(widget->appbar));
	/*
	 * If this widget can not be activated by the mouse, we do 
	 * nothing else.
	 */
	if (!widget->active)
		return;
	/*
	 * Now this is a widget the mouse just left from.
	 */
	g_return_if_fail(widget->background != NULL);
	if (widget->begin_mark != NULL &&
			widget->end_mark != NULL &&
			widget->active_style != NULL)
	{
		gtk_text_buffer_get_iter_at_mark(
				widget->background->text_buffer,
				&begin,
				widget->begin_mark);
		gtk_text_buffer_get_iter_at_mark(
				widget->background->text_buffer,
				&end,
				widget->end_mark);
		gtk_text_buffer_remove_tag_by_name(
				widget->background->text_buffer, 
				widget->active_style, 
				&begin, &end);
	}

#if 0
	/* Setting the mouse cursor */	
	if (widget->background != NULL && 
			widget->mouse_cursor_inactive != NULL)
		gdk_window_set_cursor(widget->background->window, 
				widget->mouse_cursor_inactive);
#endif
}

/**
 * cui_widget_append_file_command:
 * @widget: a #CuiWidget
 * @opcode: the code of this operation
 * @file_arg: the #CuiFilename argument of the command
 *
 * The functions appends a new command to the widget. The commands are
 * instructions we have to execute when the user activates the widget. As it is
 * obvious I'm not exactly sure how we have to develop this command feature, 
 * but it is essential to send lines to the shell.
 *
 * This function handles only commands used to handle files.
 */
void
cui_widget_append_file_command(CuiWidget *widget, 
		CuiCommandOpcode opcode,
		CuiFilename *file_arg)
{
	CuiCommand *command;
	g_return_if_fail(CUI_IS_WIDGET(widget));
	g_return_if_fail(CUI_IS_FILENAME(file_arg));

	command = g_malloc(sizeof(CuiCommand));
	command->type = FileCommand;
	command->opcode = opcode;
	command->parg1 = file_arg;
	command->parg2 = NULL;
	widget->commands = g_list_append(widget->commands, command);
}

/**
 * cui_widget_append_command_with_arg:
 * @widget: a #CuiWidget
 * @opcode: the code of this operation
 * @file_arg: the #CuiFilename argument of the command
 * @arg: the other argument of the command
 *
 * The functions appends a new command to the widget. The commands are
 * instructions we have to execute when the user activates the widget. The
 * command has an argument which is a pointer of arbitrary type. 
 *
 * This function handles only commands used to handle files.
 */
void
cui_widget_append_file_command_with_arg(CuiWidget *widget, 
		CuiCommandOpcode opcode,
		CuiFilename *file_arg,
		gpointer     arg)
{
	CuiCommand *command;
	g_return_if_fail(CUI_IS_WIDGET(widget));
	g_return_if_fail(CUI_IS_FILENAME(file_arg));

	command = g_malloc(sizeof(CuiCommand));
	command->type = FileCommandWithArg;
	command->opcode = opcode;
	command->parg1 = file_arg;
	command->parg2 = arg;
	widget->commands = g_list_append(widget->commands, command);
}

/**
 * cui_widget_append_window_command:
 * @widget: a #CuiWidget
 * @opcode: the opcode of the command
 * @window: a #CuiWindow
 *
 * Appends a new window command to the widget.
 */
void
cui_widget_append_window_command(CuiWidget *widget,
		CuiCommandOpcode opcode,
		gpointer window)
{
	CuiCommand *command;
	g_return_if_fail(CUI_IS_WIDGET(widget));
	g_return_if_fail(CUI_IS_WINDOW(window));

	command = g_malloc(sizeof(CuiCommand));
	command->type = WindowCommand;
	command->opcode = opcode;
	command->parg1 = window;
	command->parg2 = NULL;
	widget->commands = g_list_append(widget->commands, command);
}

/**
 * cui_widget_append_window_command_with_arg:
 * @widget: a #CuiWidget
 * @opcode: the opcode of the command
 * @window: a #CuiWindow
 * @arg: the argument
 *
 * Appends a new window command to the widget.
 */
void
cui_widget_append_window_command_with_arg(CuiWidget *widget,
		CuiCommandOpcode opcode,
		gpointer window,
		gpointer arg)
{
	CuiCommand *command;
	g_return_if_fail(CUI_IS_WIDGET(widget));
	g_return_if_fail(CUI_IS_WINDOW(window));

	command = g_malloc(sizeof(CuiCommand));
	command->type = WindowCommandWithArg;
	command->opcode = opcode;
	command->parg1 = window;
	command->parg2 = arg;
	widget->commands = g_list_append(widget->commands, command);
}

/**
 * cui_widget_activate:
 * @widget: a #CuiWidget
 *
 * Emits the commands associated with the widget. This function should be 
 * called when the user activates the widget.
 *
 */
void
cui_widget_activate(CuiWidget *widget)
{
	Terminal *terminal;
	CuiGenerator *command_generator;
	
	g_return_if_fail(CUI_IS_WIDGET(widget));
	terminal = widget->terminal;
	g_return_if_fail(terminal != NULL);
	command_generator = terminal->command_generator;
	g_return_if_fail(CUI_IS_GENERATOR(command_generator));
	/*
	 * Generating the commands.
	 */
	cui_generator_emit_commands(command_generator, widget->commands);
}



/*****************************************************************************
 * Instantiable classed types: CuiActionArea                                 *
 *                                                                           *
 *****************************************************************************/
static void
cui_action_area_set_property (
		GObject      *object, 
		guint         property_id, 
		const GValue *value, 
		GParamSpec   *pspec) {
}

static void
cui_action_area_get_property (
		GObject      *object, 
		guint         property_id, 
		GValue       *value, 
		GParamSpec   *pspec)
{
	g_value_set_string (value, "value");
}

static gboolean
cui_action_area_expose_handler(CuiActionArea *area,
		gint arg1,
		gpointer user_data)
{
	Background *bg;
	GdkColor red;
	GdkRectangle *position = NULL;
	g_return_val_if_fail(area != NULL, FALSE);

	bg = cui_widget_get_background(CUI_WIDGET(area));
	g_return_val_if_fail(bg != NULL, FALSE);

	if (background_get_view_flag(bg, ActiveAreas)){
		position = cui_widget_get_area_window_coords(CUI_WIDGET(area));
		g_return_val_if_fail(position != NULL, FALSE);
		/*
		 * If the rectangle is very far from here to be visible, we return to
		 * speed up athe process a little bit.
		 */
		if (position->y < 0 && position->y + position->height < 0)
			return -1;

		/*
		 * We begin to draw here.
		 */
		gdk_color_parse("Purple", &red);
		gdk_colormap_alloc_color(bg->colormap, &red, TRUE, TRUE);
		gdk_gc_set_foreground(bg->gc, &red);
		gdk_draw_rectangle(bg->window, 
				bg->gc, 
				FALSE, 
				position->x, position->y, 
				position->width, position->height);
		/*
		 * The cui_widget_get_area_window_coords returning an allocated memory
		 * pointer, however, the cui_widget_get_area is not. Stupid...
		 */
		g_free(position);
	}
	return TRUE;
}

static void
cui_action_area_base_class_init(CuiActionAreaClass *klass){
	static gboolean initialized = FALSE;
	GClosure *default_closure;
	GType params[1];
	guint id;
	
	if (!initialized){
		/* interface signals creation*/
		default_closure = g_cclosure_new(
				G_CALLBACK(cui_action_area_expose_handler),
				NULL, /* user data */
				NULL  /* destroy data */
				);
		params[0] = G_TYPE_INT;
		id = g_signal_newv(
				"expose",
				CUI_ACTION_AREA_TYPE,
				G_SIGNAL_RUN_LAST,
				default_closure,
				NULL, /* Accumulator function */
				NULL, /* Accumulator data */
				g_cclosure_marshal_BOOLEAN__FLAGS, 
				G_TYPE_BOOLEAN, /* return value */
				1, /* param number */
				params /*param type array */
				);
		initialized = TRUE;
	}
}

static void
cui_action_area_class_init(
		gpointer g_class, 
		gpointer g_class_data){

	GObjectClass *gobject_class = G_OBJECT_CLASS (g_class);
	CuiActionAreaClass *klass = CUI_ACTION_AREA_CLASS(g_class);	

	gobject_class->set_property = cui_action_area_set_property;
  	gobject_class->get_property = cui_action_area_get_property;
	
}

static void
cui_action_area_instance_init(
		GTypeInstance   *instance, 
		gpointer         g_class){
}

GType 
cui_action_area_get_type(void)
{
	static GType type = 0;
	
	if (type == 0) {
		static const GTypeInfo info = {
			sizeof(CuiActionAreaClass),
			(GBaseInitFunc) cui_action_area_base_class_init, 
			NULL,	                                        
			(GClassInitFunc) cui_action_area_class_init,
			NULL,	  
			NULL,	  
			sizeof(CuiActionArea),
			0,         
			(GInstanceInitFunc) cui_action_area_instance_init 
		};
		
		type = g_type_register_static(CUI_WIDGET_TYPE, 
				"CuiActionArea", 
				&info, 
				0);
	}
	return type;
}

/**
 * cui_action_area_new:
 * @returns: A newly allocated and initialized action area
 *
 * Creates a newly allocated action area, a rectangular box on the CUI which can
 * be activated with a double click by the user.
 */
CuiWidget *
cui_action_area_new(void)
{
	CuiWidget *area;
	area = g_object_new(CUI_ACTION_AREA_TYPE, NULL);
	return area;
}

/**
 * cui_action_area_redraw:
 * @area: the #CuiActionArea to redraw
 * @debug: if this argument is not 0 the area will be highlighted
 * @returns: the return value of the signal handler
 *
 * Emits the expose event for the widget to redraw the screen.
 */
gboolean 
cui_action_area_redraw(CuiActionArea *area, gint debug){
	gboolean retval;
	
	g_signal_emit_by_name(area, "expose", debug, &retval);
	return retval;	
}


/*****************************************************************************
 * Class: CuiPixmap                                                          *
 *                                                                           *
 *****************************************************************************/
static void
cui_pixmap_set_property (
		GObject      *object, 
		guint         property_id, 
		const GValue *value, 
		GParamSpec   *pspec) 
{
}

static void
cui_pixmap_get_property (
		GObject      *object, 
		guint         property_id, 
		GValue       *value, 
		GParamSpec   *pspec)
{
	//CuiWidget *self = (CuiWidget *) object;
	g_value_set_string (value, "value");
}

static gboolean
cui_pixmap_expose_handler(CuiPixmap *cui_pixmap,
		gint arg1,
		gpointer user_data)
{
	Background *bg;
	GdkRectangle *position = NULL;
	
	/*
	 * Collecting data.
	 */
	g_return_val_if_fail(cui_pixmap != NULL, FALSE);
	bg = cui_widget_get_background(CUI_WIDGET(cui_pixmap));
	g_return_val_if_fail(bg != NULL, FALSE);
	position = cui_widget_get_area_window_coords(CUI_WIDGET(cui_pixmap));
	g_return_val_if_fail(position != NULL, FALSE);
	/*
	 * Actual drawing.
	 */
	gdk_pixbuf_render_to_drawable(cui_pixmap->pixbuf,
                                GDK_DRAWABLE(bg->window),
                                bg->gc,
                                0, 0,
                                position->x, position->y,
                                -1, -1,
                                GDK_RGB_DITHER_MAX,
                                0, 0);
	/*
	 * The cui_widget_get_area_window_coords returning an allocated memory
	 * pointer, however, the cui_widget_get_area is not. Stupid...
	 */
	g_free(position);
	return TRUE;
}

static void
cui_pixmap_base_class_init(CuiPixmapClass *klass)
{
	static gboolean initialized = FALSE;
	GClosure *default_closure;
	GType params[1];
	guint id;
	
	if (!initialized){
		/* interface signals creation*/
		default_closure = g_cclosure_new(
				G_CALLBACK(cui_pixmap_expose_handler),
				NULL, /* user data */
				NULL  /* destroy data */
				);
		params[0] = G_TYPE_INT;
		id = g_signal_newv(
				"expose",
				CUI_PIXMAP_TYPE,
				G_SIGNAL_RUN_LAST,
				default_closure,
				NULL, /* Accumulator function */
				NULL, /* Accumulator data */
				g_cclosure_marshal_BOOLEAN__FLAGS, 
				G_TYPE_BOOLEAN, /* return value */
				1, /* param number */
				params /*param type array */
				);
		initialized = TRUE;
	}
}

static void
cui_pixmap_class_init(
		gpointer g_class, 
		gpointer g_class_data){

	GObjectClass *gobject_class = G_OBJECT_CLASS (g_class);
	CuiPixmapClass *klass = CUI_PIXMAP_CLASS(g_class);	

	gobject_class->set_property = cui_pixmap_set_property;
  	gobject_class->get_property = cui_pixmap_get_property;
	
}

static void
cui_pixmap_instance_init(
		GTypeInstance   *instance, 
		gpointer         g_class)
{
	CuiPixmap *self = (CuiPixmap *)instance;
	
	
	self->filename = NULL;
	self->pixbuf = NULL;
	self->pixmap = NULL;
}

GType 
cui_pixmap_get_type(void)
{
	static GType type = 0;
	
	if (type == 0) {
		static const GTypeInfo info = {
			sizeof(CuiPixmapClass),
			(GBaseInitFunc) cui_pixmap_base_class_init,
			NULL,
			(GClassInitFunc) cui_pixmap_class_init,
			NULL,
			NULL,
			sizeof(CuiPixmap),
			0,
			(GInstanceInitFunc) cui_pixmap_instance_init
		};
		
		type = g_type_register_static(CUI_WIDGET_TYPE, 
				"CuiPixmap", 
				&info, 
				0);
	}
	return type;
}

/**
 * cui_pixmap_set_xy:
 * @cui_pixmap: a #CuiPixmap to set the coordinates for
 * @x: the new x coordinate
 * @y: the new y coordinate
 *
 * Sets the place where the pixmap will appear without tampering with the size
 * of the widget.
 */
void
cui_pixmap_set_xy(CuiPixmap *cui_pixmap, 
		gint x, 
		gint y)
{
	GdkRectangle *area;
	GdkRectangle new_area;
	
	g_return_if_fail(CUI_IS_PIXMAP(cui_pixmap));
	
	area = cui_widget_get_area(CUI_WIDGET(cui_pixmap));
	g_return_if_fail(area != NULL);
	new_area.x = x;
	new_area.y = y;
	new_area.width = area->width;
	new_area.height = area->height;
	cui_widget_set_area(CUI_WIDGET(cui_pixmap), &new_area);
}

/**
 * cui_pixmap_new_thumbnail:
 * @file: we create the thumbnail for this file
 * @returns: a new pixmap as a widget
 * 
 * Creates a thumbnail for the given file. The thumbnail will represent the
 * content of the given file if the file is local and can be opened for read.
 * FIXME: This function must be completed.
 */
CuiWidget *
cui_pixmap_new_thumbnail(CuiFilename *file)
{
	CuiPixmap *cui_pixmap;
	gchar *tmp;
	GdkPixbuf *pixbuf;	
	GError *error = NULL;
	GdkRectangle area;

#ifdef DEBUG
g_message("%s(): Start", __func__);
#endif
	/*
	 * Creating widget and loading pixbuf from file.
	 */
	cui_pixmap = g_object_new(CUI_PIXMAP_TYPE, NULL);
	cui_pixmap->icon = cui_icon_new(file);
	
	/*
	 * getting the pixbuf, rendering, and getting the size
	 */
	cui_pixmap->pixbuf = cui_icon_get_pixbuf(cui_pixmap->icon);
	gdk_pixbuf_render_pixmap_and_mask(
			cui_pixmap->pixbuf, 
			&cui_pixmap->pixmap,
			&cui_pixmap->bitmap, 
			100);
	area.x = 0;
	area.y = 0;
	gdk_drawable_get_size(cui_pixmap->pixmap, &area.width, &area.height);
	cui_widget_set_area(CUI_WIDGET(cui_pixmap), &area);
end_of_job:
	return CUI_WIDGET(cui_pixmap);
}

void
cui_pixmap_set_max_size(CuiPixmap *cui_pixmap, 
		gint max_size)
{
	GdkRectangle area;
	GdkRectangle *orig_area;
	
	g_return_if_fail(CUI_IS_PIXMAP(cui_pixmap));
	if (cui_pixmap->icon == NULL)
		return;
	
	g_object_set(G_OBJECT(cui_pixmap->icon), "max_size", max_size, NULL);
	/*
	 * getting the pixbuf, rendering, and getting the size
	 */
	cui_pixmap->pixbuf = cui_icon_get_pixbuf(cui_pixmap->icon);
	gdk_pixbuf_render_pixmap_and_mask(
			cui_pixmap->pixbuf, 
			&cui_pixmap->pixmap,
			&cui_pixmap->bitmap, 
			100);
	
	orig_area = cui_widget_get_area(CUI_WIDGET(cui_pixmap));
	if (orig_area != NULL) {
		area.x = orig_area->x;
		area.y = orig_area->y;
	} else { 
		area.x = 0;
		area.y = 0;
	}
	gdk_drawable_get_size(cui_pixmap->pixmap, &area.width, &area.height);
	cui_widget_set_area(CUI_WIDGET(cui_pixmap), &area);
}

void
cui_pixmap_set_max_height(CuiPixmap *cui_pixmap, 
		gint max_height)
{
	GdkRectangle area;
	GdkRectangle *orig_area;
	
	g_return_if_fail(CUI_IS_PIXMAP(cui_pixmap));
	if (cui_pixmap->icon == NULL)
		return;
	
	g_object_set(G_OBJECT(cui_pixmap->icon), "max_height", max_height, NULL);
	/*
	 * getting the pixbuf, rendering, and getting the size
	 */
	cui_pixmap->pixbuf = cui_icon_get_pixbuf(cui_pixmap->icon);
	gdk_pixbuf_render_pixmap_and_mask(
			cui_pixmap->pixbuf, 
			&cui_pixmap->pixmap,
			&cui_pixmap->bitmap, 
			100);
	
	orig_area = cui_widget_get_area(CUI_WIDGET(cui_pixmap));
	if (orig_area != NULL) {
		area.x = orig_area->x;
		area.y = orig_area->y;
	} else { 
		area.x = 0;
		area.y = 0;
	}
	gdk_drawable_get_size(cui_pixmap->pixmap, &area.width, &area.height);
	cui_widget_set_area(CUI_WIDGET(cui_pixmap), &area);
}

	
/**
 * cui_pixmap_new_from_package_file:
 * @filename: the name of the file containing pixel information
 * @returns: a new #CuiPixmap
 *
 * Creates a new #CuiPixmap loading pixel and size information from a file. 
 * This function searches the graphic file in the application pixmap directory.
 *
 */
CuiWidget *
cui_pixmap_new_from_package_file(const gchar *filename)
{
	CuiPixmap *cui_pixmap;
	GError *error = NULL;
	GdkRectangle area;
	
#ifdef DEBUG
g_message("%s(): Start", __func__);
#endif
	/*
	 * Creating widget and loading pixbuf from file.
	 */
	cui_pixmap = g_object_new(CUI_PIXMAP_TYPE, NULL);
	cui_pixmap->filename = g_strconcat(PIXMAP_DIR, filename, NULL);
	cui_pixmap->pixbuf = gdk_pixbuf_new_from_file(cui_pixmap->filename, 
			&error);
	if (cui_pixmap->pixbuf == NULL){
		g_warning("%s(): Unable to load pixbuf from '%s': %s", 
				__func__, 
				cui_pixmap->filename,
				error->message);
		goto end_of_job;
	}
	
	/*
	 * Rendering and reading the size.
	 */
	gdk_pixbuf_render_pixmap_and_mask(
			cui_pixmap->pixbuf, 
			&cui_pixmap->pixmap,
			&cui_pixmap->bitmap, 
			100);
	area.x = 0;
	area.y = 0;
	gdk_drawable_get_size(cui_pixmap->pixmap, &area.width, &area.height);
	cui_widget_set_area(CUI_WIDGET(cui_pixmap), &area);
end_of_job:
	return CUI_WIDGET(cui_pixmap);
}

/**
 * cui_pixmap_new_empty:
 * @returns: a new emty #CuiPixmap
 *
 */
CuiWidget*
cui_pixmap_new_empty(void)
{
	CuiWidget *pixmap;

#ifdef DEBUG
g_message("%s(): Start", __func__);
#endif
	pixmap = g_object_new(CUI_PIXMAP_TYPE, NULL);
	return pixmap;
}

/**
 * cui_pixmap_redraw:
 * @pixmap: the #CuiPixmap to redraw
 * @debug: if this argument is not 0 the area will be highlighted
 * @returns: the return value of the signal handler
 *
 * Emits the expose event for the widget to redraw the screen.
 */
gboolean 
cui_pixmap_redraw(CuiPixmap *pixmap, gint debug){
	gboolean retval;

#ifdef DEBUG
g_message("%s(): Start", __func__);
#endif
	g_signal_emit_by_name(pixmap, "expose", debug, &retval);
	return retval;	
}

/*****************************************************************************
 * Instantiable classed types: CuiEntry                                      *
 *                                                                           *
 *****************************************************************************/
static void
cui_entry_set_property (
		GObject      *object, 
		guint         property_id, 
		const GValue *value, 
		GParamSpec   *pspec) {
}

static void
cui_entry_get_property (
		GObject      *object, 
		guint         property_id, 
		GValue       *value, 
		GParamSpec   *pspec)
{
	//CuiEntry *self = (CuiEntry *) object;
	g_value_set_string (value, "value");
}

static void
cui_entry_class_init(
		gpointer g_class, 
		gpointer g_class_data){

	GObjectClass *gobject_class = G_OBJECT_CLASS (g_class);
	CuiEntryClass *klass = CUI_ENTRY_CLASS(g_class);	

	gobject_class->set_property = cui_entry_set_property;
  	gobject_class->get_property = cui_entry_get_property;
}

static void
cui_entry_instance_init(
		GTypeInstance   *instance, 
		gpointer         g_class)
{
	CuiEntry *self = (CuiEntry *) instance;
	self->default_text = NULL;
	self->box = NULL;
	self->window = NULL;
	self->entry = NULL;
	self->ok = NULL;
	self->cancel = NULL;
	self->ok_pixmap = NULL;
	self->cancel_pixmap = NULL;
}

GType 
cui_entry_get_type(void)
{
	static GType type = 0;
	
	if (type == 0) {
		static const GTypeInfo info = {
			sizeof(CuiEntryClass),
			NULL,
			NULL,
			(GClassInitFunc) cui_entry_class_init,
			NULL,
			NULL,
			sizeof(CuiEntry),
			0,
			(GInstanceInitFunc) cui_entry_instance_init
		};
		
		type = g_type_register_static(CUI_WIDGET_TYPE, 
				"CuiEntry", 
				&info, 
				0);
	}
	return type;
}

/**
 * cui_entry_new:
 * @returns: a new #CuiEntry widget
 * 
 * Creates a new #CuiEntry widget.
 */
CuiWidget *
cui_entry_new(void)
{
	CuiWidget *entry;
	entry = g_object_new(CUI_ENTRY_TYPE, NULL);
	return entry;
}

/**
 * cui_entry_set_default_text:
 * @entry: a #CuiEntry
 * @text: the default text for the entry
 * 
 * Sets the default text for the entry. The default text will appear when the
 * entry shown on the screen.
 */
void
cui_entry_set_default_text(CuiEntry *entry, 
		const gchar *text)
{
	g_return_if_fail(CUI_IS_ENTRY(entry));
	entry->default_text = g_strdup(text);
}

static void
entry_unrealize(CuiEntry *entry)
{
	gtk_widget_destroy(entry->window);
	
	entry->ok_pixmap = NULL;
	entry->cancel_pixmap = NULL;
	entry->ok = NULL;
	entry->cancel = NULL;
	entry->entry = NULL;
	entry->box = NULL;
	entry->window = NULL;
}

/**
 * entry_ok_pressed:
 * Callback function for CuiEntry OK button. This function is called
 * automaticly, when the user pressed the OK button. This is where we run the
 * commands of the #CuiEntry type #CuiWidget.
 */
static void 
entry_ok_pressed(GtkWidget *button, gpointer p)
{
	CuiEntry *entry = p;
	const gchar *text;
	Terminal *terminal;
	CuiGenerator *command_generator;
	GList *list;
	CuiCommand *command;
	
	g_return_if_fail(CUI_IS_ENTRY(entry));
	text = gtk_entry_get_text(GTK_ENTRY(entry->entry));
	/*
	 * If we didn't change the text, we can do nothing.
	 */
	if (strcmp(entry->default_text, text) == 0)
		goto nothing_to_do;
	/*
	 * We start the job here.
	 */
	terminal = cui_widget_get_terminal(CUI_WIDGET(entry));
	g_return_if_fail(terminal != NULL);
	command_generator = terminal->command_generator;
	g_return_if_fail(CUI_IS_GENERATOR(command_generator));
	
	list = CUI_WIDGET(entry)->commands;
	while (list != NULL){
		command = list->data;
		switch (command->type){
			case FileCommand:
				cui_generator_emit_file_command_with_arg(
						command_generator, 
						command->opcode, 
						CUI_FILENAME(command->parg1),
						text);
#if 0
				cui_generator_emit_file_command(
						command_generator, 
						command->opcode, 
						CUI_FILENAME(command->parg1));
#endif
				break;
			case WindowCommand:
				cui_generator_emit_window_command(
						command_generator,
						command->opcode,
						CUI_WINDOW(command->parg1));
				break;
			default:
				g_warning("%s(): Unknown command type: %1",
						__func__, command->type);
		}
next_command:
		list = list->next;
	}

nothing_to_do:
	entry_unrealize(entry);
}

/**
 *
 * Callback function for CuiEntry Cancel button.
 */
static void 
entry_cancel_pressed(GtkWidget *button, gpointer p)
{
	CuiEntry *entry = p;
	g_return_if_fail(CUI_IS_ENTRY(entry));
	entry_unrealize(entry);
}

static gboolean
entry_focus_out_event(GtkWidget *widget, 
		GdkEventFocus   *event, 
		gpointer         user_data)
{
	entry_unrealize(user_data);
	return FALSE;
}

/**
 * cui_entry_realize:
 * @entry: a #CuiEntry type widget
 *
 */
void
cui_entry_realize(CuiEntry *entry)
{
	Terminal *terminal;
	Background *background;
	GdkRectangle *activation_area;
	gint w_left, w_top, here_x, here_y;
	GtkStyle *style;
	
	g_return_if_fail(CUI_IS_ENTRY(entry));
	terminal = cui_widget_get_terminal(CUI_WIDGET(entry));
	background = cui_widget_get_background(CUI_WIDGET(entry));
	g_return_if_fail(terminal != NULL && background != NULL);
	activation_area = cui_widget_get_area_window_coords(CUI_WIDGET(entry));
	style = gtk_widget_get_style(GTK_WIDGET(background->text_view));
	/*
	 * Creating a new window to hold the entry and the two buttons. This
	 * window must be a top level window, but without any decoration.
	 */
	gdk_window_get_origin(background->window, &w_left, &w_top);
	entry->window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	gtk_window_set_decorated(GTK_WINDOW(entry->window), FALSE);
	gtk_widget_set_name(entry->window, "CuiEntry");
	gtk_widget_set_uposition(entry->window, 
			activation_area->x + w_left, 
			activation_area->y + w_top);
	/*
	 * Creating a horizontal box to hold the three widgets.
	 */
	entry->box = gtk_hbox_new(FALSE, 0);
	gtk_container_add(GTK_CONTAINER(entry->window), entry->box);
	/*
	 * Creating a new GtkEntry widget for the window.
	 */
	entry->entry = gtk_entry_new();
	gtk_widget_set_name(entry->entry, "CuiEntry");
	if (entry->default_text != NULL)
		gtk_entry_set_text(GTK_ENTRY(entry->entry), 
			entry->default_text);
	gtk_widget_set_usize(entry->entry, 
			activation_area->width + 2,
			activation_area->height + 2);
	gtk_widget_set_style(entry->entry, style);
	/*
	 * Creating an ok button.
	 */
	//g_message("Loading pixmap from: '%s'", PIXMAP_DIR "apply.png");
	entry->ok_pixmap = gtk_image_new_from_file(PIXMAP_DIR "apply.png");
	entry->ok = gtk_button_new();
	gtk_container_set_border_width(GTK_CONTAINER(entry->ok), 0);
	gtk_container_add(GTK_CONTAINER(entry->ok), entry->ok_pixmap);
	GTK_WIDGET_SET_FLAGS(entry->ok, GTK_CAN_DEFAULT);
	gtk_widget_set_usize(entry->ok, 22, activation_area->height);
	g_signal_connect(G_OBJECT(entry->ok), "clicked",
			G_CALLBACK(entry_ok_pressed), entry);
	/*
	 * Creating a cancel button.
	 */
	//g_message("Loading pixmap from: '%s'", PIXMAP_DIR "cancel.png");
	entry->cancel_pixmap = gtk_image_new_from_file(PIXMAP_DIR "cancel.png");
	entry->cancel = gtk_button_new();
	gtk_container_set_border_width(GTK_CONTAINER(entry->cancel), 0);
	gtk_container_add(GTK_CONTAINER(entry->cancel), entry->cancel_pixmap);
	gtk_widget_set_usize(entry->cancel, 22, activation_area->height);
	g_signal_connect(G_OBJECT(entry->cancel), "clicked",
			G_CALLBACK(entry_cancel_pressed), entry);
	/*
	 * We have to put all together.
	 */
	gtk_box_pack_start(GTK_BOX(entry->box), GTK_WIDGET(entry->entry), 
			TRUE, TRUE, 0);
	gtk_box_pack_start(GTK_BOX(entry->box), GTK_WIDGET(entry->ok), 
			TRUE, TRUE, 0);
	gtk_box_pack_start(GTK_BOX(entry->box), GTK_WIDGET(entry->cancel), 
			TRUE, TRUE, 0);
	/*
	 * Handling focus...
	 */
	gtk_widget_grab_focus(entry->window);
	gtk_widget_grab_focus(entry->entry);
	gtk_widget_grab_default(entry->ok);
	gtk_entry_set_activates_default(GTK_ENTRY(entry->entry), TRUE);
	/*
	 * Let's put the new window into the text_view window
	 */
	gtk_widget_show_all(entry->window);
	gtk_grab_add(entry->window);
	/*
	 * If this window loses the focus we destroy it in this callback
	 * function.
	 */
	g_signal_connect(G_OBJECT(entry->window), "focus_out_event",
                    G_CALLBACK(entry_focus_out_event), entry);
	g_free(activation_area);
}

/*****************************************************************************
 * Instantiable classed types: CuiCalendar                                   *
 *                                                                           *
 *****************************************************************************/
static void
cui_calendar_set_property (
		GObject      *object, 
		guint         property_id, 
		const GValue *value, 
		GParamSpec   *pspec) 
{
}

static void
cui_calendar_get_property (
		GObject      *object, 
		guint         property_id, 
		GValue       *value, 
		GParamSpec   *pspec)
{
	g_value_set_string (value, "value");
}

static void
cui_calendar_class_init(
		gpointer g_class, 
		gpointer g_class_data){

	GObjectClass *gobject_class = G_OBJECT_CLASS (g_class);
	CuiCalendarClass *klass = CUI_CALENDAR_CLASS(g_class);	

	gobject_class->set_property = cui_calendar_set_property;
  	gobject_class->get_property = cui_calendar_get_property;
}

static void
cui_calendar_instance_init(
		GTypeInstance   *instance, 
		gpointer         g_class)
{
	CuiCalendar *self = (CuiCalendar *) instance;
	self->hbox = NULL;
	self->vbox = NULL;
	self->window = NULL;
	self->calendar = NULL;
	self->ok = NULL;
	self->cancel = NULL;
	self->ok_pixmap = NULL;
	self->cancel_pixmap = NULL;
	self->default_date_set = FALSE;
	self->date_chosen = FALSE;
}

GType 
cui_calendar_get_type(void)
{
	static GType type = 0;
	
	if (type == 0) {
		static const GTypeInfo info = {
			sizeof(CuiCalendarClass),
			NULL,
			NULL,
			(GClassInitFunc) cui_calendar_class_init,
			NULL,
			NULL,
			sizeof(CuiCalendar),
			0,
			(GInstanceInitFunc) cui_calendar_instance_init
		};
		
		type = g_type_register_static(CUI_WIDGET_TYPE, 
				"CuiCalendar", 
				&info, 
				0);
	}
	return type;
}

/**
 * cui_calendar_new:
 * @returns: a new #CuiCalendar widget
 *
 * Creates a new #CuiCalendar widget.
 */
CuiWidget *
cui_calendar_new(void)
{
	CuiWidget *calendar;
	calendar = g_object_new(CUI_CALENDAR_TYPE, NULL);
	return calendar;
}

static void
calendar_unrealize(CuiCalendar *calendar)
{
	gtk_widget_destroy(calendar->window);
	
	calendar->ok_pixmap = NULL;
	calendar->cancel_pixmap = NULL;
	calendar->ok = NULL;
	calendar->cancel = NULL;
	calendar->calendar = NULL;
	calendar->hbox = NULL;
	calendar->vbox = NULL;
	calendar->window = NULL;
}

/**
 * cui_calendar_activated:
 *
 *
 */
static void
cui_calendar_activated(CuiCalendar *calendar)
{
	gchar *text;
	Terminal *terminal;
	CuiGenerator *command_generator;
	GList *list;
	CuiCommand *command;
	/*
	 * Getting the date and destroying window.
	 */
	g_return_if_fail(CUI_IS_CALENDAR(calendar));
	gtk_calendar_get_date(GTK_CALENDAR(calendar->calendar), 
			&calendar->year, 
			&calendar->month, 
			&calendar->day);
	++calendar->month;
	calendar_unrealize(calendar);
	/*
	 * If we didn't change the date, we can do nothing.
	 */
	if (calendar->default_date_set &&
			calendar->year == calendar->default_year &&
			calendar->month == calendar->default_month &&
			calendar->day == calendar->default_day)
		return;
	/*
	 * Creating the date string.
	 */
	text = g_strdup_printf("%04u-%02u-%02u", 
			calendar->year,
			calendar->month,
			calendar->day);
	/*
	 * We start the job here.
	 */
	terminal = cui_widget_get_terminal(CUI_WIDGET(calendar));
	g_return_if_fail(terminal != NULL);
	command_generator = terminal->command_generator;
	g_return_if_fail(CUI_IS_GENERATOR(command_generator));
	
	list = CUI_WIDGET(calendar)->commands;
	while (list != NULL){
		command = list->data;
		cui_generator_emit_file_command_with_arg(
				command_generator, 
				command->opcode, 
				CUI_FILENAME(command->parg1),
				text);
next_command:
		list = list->next;
	}
}

static void 
calendar_day_selected(
		GtkCalendar *widget,
		gpointer user_data)
{
	CuiCalendar *calendar = user_data;
	g_return_if_fail(CUI_IS_CALENDAR(calendar));
	cui_calendar_activated(calendar);
	calendar_unrealize(calendar);
}

/**
 *
 * Callback function for CuiCalendar OK button.
 */
static void 
calendar_ok_pressed(GtkWidget *button, gpointer p)
{
	CuiCalendar *calendar = p;
	g_return_if_fail(CUI_IS_CALENDAR(calendar));
	cui_calendar_activated(calendar);
	calendar_unrealize(calendar);
}

/**
 *
 * Callback function for CuiCalendar Cancel button.
 */
static void 
calendar_cancel_pressed(
		GtkWidget *button, 
		gpointer p)
{
	CuiCalendar *calendar = p;
	g_return_if_fail(CUI_IS_CALENDAR(calendar));
	calendar_unrealize(calendar);
}

static gboolean
calendar_focus_out_event(GtkWidget *widget, 
		GdkEventFocus   *event, 
		gpointer         user_data)
{
	CuiCalendar *calendar = user_data;
	g_return_if_fail(CUI_IS_CALENDAR(calendar));
	calendar_unrealize(calendar);
	return FALSE;
}

/**
 * calendar_set_default_date:
 * @calendar: a #CuiCalendar widget
 * @date: the default date
 *
 * Sets the default date for the calendar. The date must be in the form 
 * yyyy-mm-dd.
 */
void
cui_calendar_set_default_date(CuiCalendar *calendar, 
		gchar *date)
{
	g_return_if_fail(CUI_IS_CALENDAR(calendar) && date != NULL);
	
	calendar->default_date_set = FALSE;
	if (sscanf(date, "%4u-%2u-%2u", 
				&calendar->default_year,
				&calendar->default_month,
				&calendar->default_day) != 3) 
		g_warning("%s(): Default date '%s' is malformed.", date);
	else
		calendar->default_date_set = TRUE;
}

/**
 * cui_calendar_realize:
 * @calendar: a #CuiCalendar type widget
 *
 * Fires up the calendar on the screen
 */
void
cui_calendar_realize(CuiCalendar *calendar)
{
	Terminal *terminal;
	Background *background;
	GdkRectangle *activation_area;
	gint w_left, w_top, here_x, here_y;
	PangoFontDescription *font;

	g_return_if_fail(CUI_IS_CALENDAR(calendar));
	terminal = cui_widget_get_terminal(CUI_WIDGET(calendar));
	background = cui_widget_get_background(CUI_WIDGET(calendar));
	g_return_if_fail(terminal != NULL && background != NULL);
	activation_area = 
		cui_widget_get_area_window_coords(CUI_WIDGET(calendar));
	/*
	 * Creating a new window to hold the calendar and the two buttons. 
	 * This window must be a top level window, but without any decoration.
	 */
	gdk_window_get_origin(background->window, &w_left, &w_top);
	calendar->window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	gtk_window_set_decorated(GTK_WINDOW(calendar->window), FALSE);
	gtk_widget_set_name(calendar->window, "CuiCalendar");
	gtk_widget_set_uposition(calendar->window, 
			activation_area->x + w_left, 
			activation_area->y + w_top);
	/*
	 * Creating a horizontal and a vertical box to hold the three widgets.
	 */
	calendar->hbox = gtk_hbox_new(FALSE, 0);
	calendar->vbox = gtk_vbox_new(FALSE, 0);
	/*
	 * Creating a new GtkCalendar widget for the window.
	 */
	calendar->calendar = gtk_calendar_new();
	gtk_widget_set_name(calendar->calendar, "CuiCalendar");
	if (calendar->default_date_set){
		gtk_calendar_select_month(GTK_CALENDAR(calendar->calendar), 
				calendar->default_month - 1, 
				calendar->default_year);
		gtk_calendar_select_day(GTK_CALENDAR(calendar->calendar), 
				calendar->default_day);
	}
	font = pango_font_description_new();
	pango_font_description_set_size(font, 10 * PANGO_SCALE);
	gtk_widget_modify_font(calendar->calendar, font);
	g_signal_connect(G_OBJECT(calendar->calendar), 
			"day-selected-double-click",
                    	G_CALLBACK(calendar_day_selected), calendar);
	/*
	 * Creating an ok button.
	 */
	calendar->ok_pixmap = gtk_image_new_from_file(PIXMAP_DIR "apply.png");
	calendar->ok = gtk_button_new();
	gtk_container_set_border_width(GTK_CONTAINER(calendar->ok), 0);
	gtk_container_add(GTK_CONTAINER(calendar->ok), calendar->ok_pixmap);
	GTK_WIDGET_SET_FLAGS(calendar->ok, GTK_CAN_DEFAULT);
	gtk_widget_set_usize(calendar->ok, 22, activation_area->height);
	g_signal_connect(G_OBJECT(calendar->ok), "clicked",
			G_CALLBACK(calendar_ok_pressed), calendar);
	/*
	 * Creating a cancel button.
	 */
	calendar->cancel_pixmap = 
		gtk_image_new_from_file(PIXMAP_DIR "cancel.png");
	calendar->cancel = gtk_button_new();
	gtk_container_set_border_width(GTK_CONTAINER(calendar->cancel), 0);
	gtk_container_add(GTK_CONTAINER(calendar->cancel), 
			calendar->cancel_pixmap);
	gtk_widget_set_usize(calendar->cancel, 22, activation_area->height);
	g_signal_connect(G_OBJECT(calendar->cancel), "clicked",
			G_CALLBACK(calendar_cancel_pressed), calendar);
	/*
	 * We have to put all together.
	 */
	gtk_box_pack_start(GTK_BOX(calendar->hbox), 
			GTK_WIDGET(calendar->calendar), 
			TRUE, TRUE, 0);
	gtk_container_add(GTK_CONTAINER(calendar->window), calendar->hbox);
	gtk_container_add(GTK_CONTAINER(calendar->hbox), calendar->vbox);
	gtk_box_pack_start(GTK_BOX(calendar->vbox), 
			GTK_WIDGET(calendar->ok), 
			TRUE, TRUE, 0);
	gtk_box_pack_start(GTK_BOX(calendar->vbox), 
			GTK_WIDGET(calendar->cancel),
			TRUE, TRUE, 0);
	/*
	 * Handling focus...
	 */
	gtk_widget_grab_focus(calendar->window);
	gtk_widget_grab_focus(calendar->calendar);
	gtk_widget_grab_default(calendar->ok);
	/*
	 * Let's put the new window into the text_view window
	 */
	gtk_widget_show_all(calendar->window);
	gtk_grab_add(calendar->window);
	/*
	 * If this window loses the focus we destroy it in this callback
	 * function.
	 */
	g_signal_connect(G_OBJECT(calendar->window), "focus_out_event",
                    G_CALLBACK(calendar_focus_out_event), calendar);
	g_free(activation_area);
}

/*****************************************************************************
 * Instantiable classed types: CuiPopupMenu                                  *
 *                                                                           *
 *****************************************************************************/
static void
cui_popup_menu_set_property (
		GObject      *object, 
		guint         property_id, 
		const GValue *value, 
		GParamSpec   *pspec) {
}

static void
cui_popup_menu_get_property (
		GObject      *object, 
		guint         property_id, 
		GValue       *value, 
		GParamSpec   *pspec)
{
}

static void
cui_popup_menu_class_init(
		gpointer g_class, 
		gpointer g_class_data){

	GObjectClass *gobject_class = G_OBJECT_CLASS (g_class);
	CuiPopupMenuClass *klass = CUI_POPUP_MENU_CLASS(g_class);

	gobject_class->set_property = cui_popup_menu_set_property;
  	gobject_class->get_property = cui_popup_menu_get_property;

}

static void
cui_popup_menu_instance_init(
		GTypeInstance   *instance, 
		gpointer         g_class){
	CuiPopupMenu *popup_menu = CUI_POPUP_MENU(instance);
}

GType 
cui_popup_menu_get_type(void)
{
	static GType type = 0;
	
	if (type == 0) {
		static const GTypeInfo info = {
			sizeof(CuiPopupMenuClass),
			NULL,	
			NULL,
			(GClassInitFunc) cui_popup_menu_class_init,
			NULL,
			NULL,
			sizeof(CuiPopupMenu),
			0,
			(GInstanceInitFunc) cui_popup_menu_instance_init 
		};
		
		type = g_type_register_static(CUI_WIDGET_TYPE, 
				"CuiPopupMenu", 
				&info, 
				0);
	}
	return type;
}

/**
 * cui_popup_menu_new:
 * @gtk_widget: a GTK+ Widget used as a popup menu
 * @returns: a new #CuiPopupMenu.
 *
 * Creates a new #CuiPopupMenu.
 */
CuiWidget *
cui_popup_menu_new(GtkWidget *gtk_widget)
{
	CuiWidget *menu;
	menu = g_object_new(CUI_POPUP_MENU_TYPE, NULL);
	CUI_POPUP_MENU(menu)->menu = gtk_widget;
	return menu;
}

/**
 * cui_popup_menu_popup:
 * @cui_popup_menu: a #CuiPopupMenu to pop-up
 * @parent_widget: the parent widget of the menu
 * @event: the event that caused this opertaion
 *
 * Starts up a #CuiPopupMenu on the given location of the screen.
 */
void
cui_popup_menu_popup(CuiPopupMenu *cui_popup_menu,
		GtkWidget *parent_widget,
		GdkEventButton *event)
{
	Terminal  *terminal;
	g_return_if_fail(cui_popup_menu != NULL && 
			parent_widget != NULL && 
			event != NULL);
	g_return_if_fail(cui_popup_menu->menu != NULL);
	/*
	 * Setting some data for the callback function.
	 */
	terminal = cui_widget_get_terminal(CUI_WIDGET(cui_popup_menu));
	g_object_set_data(G_OBJECT(cui_popup_menu->menu), 
			"cui_widget", cui_popup_menu);
	g_object_set_data(G_OBJECT(cui_popup_menu->menu), 
			"terminal", terminal);
	g_object_set_data(G_OBJECT(cui_popup_menu->menu), 
			"main_widget", terminal->text_view);
	g_object_set_data(G_OBJECT(cui_popup_menu->menu), 
			"command_generator", terminal->command_generator);
	/*
	 * Starting up the popup menu.
	 */
	gtk_menu_popup(GTK_MENU(cui_popup_menu->menu), NULL, NULL, NULL, NULL, 
			event->button, event->time);
}

/*****************************************************************************
 * Instantiable classed types: CuiFileMenu                                   *
 *                                                                           *
 *****************************************************************************/
#define PROPERTY_FILE 1

static void
cui_file_menu_set_property(
		GObject      *object, 
		guint         property_id, 
		const GValue *value, 
		GParamSpec   *pspec) 
{
	CuiFileMenu *self = (CuiFileMenu *) object;
	
	switch(property_id){
		case PROPERTY_FILE:
			self->file = g_value_get_pointer(value);
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID(
					object, property_id, pspec);
	}		
}

static void
cui_file_menu_get_property(
		GObject      *object, 
		guint         property_id, 
		GValue       *value, 
		GParamSpec   *pspec)
{
	CuiFileMenu *self = (CuiFileMenu *) object;
	switch(property_id){
		case PROPERTY_FILE:
			g_value_set_pointer(value, self->file);
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID(
					object, property_id, pspec);
	}
}

static void
cui_file_menu_class_init(
		gpointer g_class, 
		gpointer g_class_data)
{
	GParamSpec *pspec;
	GObjectClass *gobject_class = G_OBJECT_CLASS (g_class);
	CuiFileMenuClass *klass = CUI_FILE_MENU_CLASS(g_class);	
	
	gobject_class->set_property = cui_file_menu_set_property;
  	gobject_class->get_property = cui_file_menu_get_property;
	klass->menu = create_file_popup_menu();
	
	pspec = g_param_spec_pointer(
			"filename", "filename-object",
                        "The CuiFilename type file name for the menu.",
                        G_PARAM_READWRITE);
	g_object_class_install_property(gobject_class, 
			PROPERTY_FILE, 
			pspec);
}

static void
cui_file_menu_instance_init(GTypeInstance   *instance, 
		gpointer         g_class)
{
	CuiFileMenu *self = (CuiFileMenu *)instance;
	CuiFileMenuClass *klass = (CuiFileMenuClass *)g_class;
	
	self->menu = klass->menu;
	self->file = NULL;
}

GType 
cui_file_menu_get_type(void)
{
	static GType type = 0;
	
	if (type == 0) {
		static const GTypeInfo info = {
			sizeof(CuiFileMenuClass),
			NULL,
			NULL,
			(GClassInitFunc) cui_file_menu_class_init,
			NULL,
			NULL,
			sizeof(CuiFileMenu),
			0,
			(GInstanceInitFunc) cui_file_menu_instance_init 
		};
		
		type = g_type_register_static(CUI_POPUP_MENU_TYPE, 
				"CuiFileMenu", 
				&info, 
				0);
	}
	return type;
}

/**
 * cui_file_menu_new:
 * @returns: a new #CuiFileMenu as #CuiWidget
 *
 * Creates a new #CuiFileMenu widget.
 */
CuiWidget*
cui_file_menu_new(void)
{
	CuiWidget *menu;
	menu = g_object_new(CUI_FILE_MENU_TYPE, NULL);
	return menu;
}

/**
 * cui_file_menu_set_filename:
 * @file_menu: a #CuiFileMenu
 * @filename: the #CuiFilename
 *
 * Sets the file for this #CuiFileMenu.
 */
void
cui_file_menu_set_filename(CuiFileMenu *file_menu, CuiFilename *filename)
{
	GValue val = {0,};
	g_return_if_fail(CUI_IS_FILE_MENU(file_menu));
	g_value_init(&val, G_TYPE_POINTER);
	g_value_set_pointer(&val, filename);
	g_object_set_property (G_OBJECT (file_menu), "filename", &val);
}

/**
 * cui_file_menu_get_filename:
 * @file_menu: a #CuiFileMenu
 * @returns: the #CuiFilename of the menu
 *
 * Returns the filename belongs to this filemenu.
 */
CuiFilename *
cui_file_menu_get_filename(CuiFileMenu *file_menu)
{
	GValue val = {0,};
	g_return_val_if_fail(CUI_IS_FILE_MENU(file_menu), NULL);
	g_value_init(&val, G_TYPE_POINTER);
	g_object_get_property (G_OBJECT (file_menu), "filename", &val);
	return CUI_FILENAME(g_value_get_pointer(&val));	
}

/**
 * cui_file_menu_popup:
 * @cui_file_menu: a #CuiFileMenu to pop-up
 * @parent_widget: the parent widget of the menu
 * @event: the event that caused this opertaion
 *
 * Starts up a #CuiFileMenu on the given location of the screen. This function
 * is very important since there are a lot of file names on the screen and the
 * file menu is a rather complicated and powerfull user interface element.
 */
void
cui_file_menu_popup(CuiFileMenu *cui_file_menu,
		GtkWidget *parent_widget,
		GdkEventButton *event)
{
	GtkWidget *menu_item;
	Terminal  *terminal;
	gboolean is;
	g_assert(cui_file_menu != NULL && 
			parent_widget != NULL && 
			event != NULL);
	g_assert(cui_file_menu->menu != NULL);
	/*
	 * We only uncompress the files we think are compressed, and we 
	 * compress only the files without compression.
	 */
	is = cui_filename_is_compressed(cui_file_menu->file);
	menu_item = lookup_widget(cui_file_menu->menu, "file_uncompress");
	gtk_widget_set_sensitive(menu_item, is);
	menu_item = lookup_widget(cui_file_menu->menu, "file_archive");
	gtk_widget_set_sensitive(menu_item, !is);
	/*
	 * We enable the "Paste" menu item only if it makes sense. That is, when
	 * there is something in the clipboard and this is a directory.
	 */
	menu_item = lookup_widget(cui_file_menu->menu, "file_paste");
	is = clipboard_get_paste() && cui_filename_is_dir(cui_file_menu->file);
	gtk_widget_set_sensitive(menu_item, is);
	/*
	 * Setting some data for the callback function.
	 */
	terminal = cui_widget_get_terminal(CUI_WIDGET(cui_file_menu));
	g_object_set_data(G_OBJECT(cui_file_menu->menu), 
			"cui_widget", cui_file_menu);
	g_object_set_data(G_OBJECT(cui_file_menu->menu), 
			"terminal", terminal);
	g_object_set_data(G_OBJECT(cui_file_menu->menu), 
			"command_generator", terminal->command_generator);
	/*
	 * Starting up the popup menu.
	 */
	gtk_menu_popup(GTK_MENU(cui_file_menu->menu), NULL, NULL, NULL, NULL, 
			event->button, event->time);
}

/*****************************************************************************
 * Instantiable classed types: CuiModeMenu                                   *
 *                                                                           *
 *****************************************************************************/
static void
cui_mode_menu_set_property(
		GObject      *object, 
		guint         property_id, 
		const GValue *value, 
		GParamSpec   *pspec) 
{
	CuiModeMenu *self = (CuiModeMenu *) object;
	
	switch(property_id){
		case PROPERTY_FILE:
			self->file = g_value_get_pointer(value);
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID(
					object, property_id, pspec);
	}		
}

static void
cui_mode_menu_get_property(
		GObject      *object, 
		guint         property_id, 
		GValue       *value, 
		GParamSpec   *pspec)
{
	CuiModeMenu *self = (CuiModeMenu *) object;
	switch(property_id){
		case PROPERTY_FILE:
			g_value_set_pointer(value, self->file);
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID(
					object, property_id, pspec);
	}
}

static void
cui_mode_menu_class_init(
		gpointer g_class, 
		gpointer g_class_data)
{
	GParamSpec *pspec;
	GObjectClass *gobject_class = G_OBJECT_CLASS (g_class);
	CuiModeMenuClass *klass = CUI_MODE_MENU_CLASS(g_class);	
	
	gobject_class->set_property = cui_mode_menu_set_property;
  	gobject_class->get_property = cui_mode_menu_get_property;
	klass->menu = create_mode_popup_menu();
	
	pspec = g_param_spec_pointer(
			"filename", "filename-object",
                        "The CuiFilename type file name for the menu.",
                        G_PARAM_READWRITE);
	g_object_class_install_property(gobject_class, 
			PROPERTY_FILE, 
			pspec);
}

static void
cui_mode_menu_instance_init(GTypeInstance   *instance, 
		gpointer         g_class)
{
	CuiModeMenu *self = (CuiModeMenu *)instance;
	CuiModeMenuClass *klass = (CuiModeMenuClass *)g_class;
	
	self->menu = klass->menu;
	self->file = NULL;
}

GType 
cui_mode_menu_get_type(void)
{
	static GType type = 0;
	
	if (type == 0) {
		static const GTypeInfo info = {
			sizeof(CuiModeMenuClass),
			NULL,
			NULL,
			(GClassInitFunc) cui_mode_menu_class_init,
			NULL,
			NULL,
			sizeof(CuiModeMenu),
			0,
			(GInstanceInitFunc) cui_mode_menu_instance_init 
		};
		
		type = g_type_register_static(CUI_POPUP_MENU_TYPE, 
				"CuiModeMenu", 
				&info, 
				0);
	}
	return type;
}

/**
 * cui_file_menu_new:
 *
 *
 */
CuiWidget*
cui_mode_menu_new(void)
{
	CuiWidget *menu;
	menu = g_object_new(CUI_MODE_MENU_TYPE, NULL);
	return menu;
}

void
cui_mode_menu_set_filename(CuiModeMenu *mode_menu, CuiFilename *filename)
{
	GValue val = {0,};
	g_return_if_fail(CUI_IS_MODE_MENU(mode_menu));
	g_value_init(&val, G_TYPE_POINTER);
	g_value_set_pointer(&val, filename);
	g_object_set_property (G_OBJECT(mode_menu), "filename", &val);
}

/**
 * cui_mode_menu_popup:
 * @cui_mode_menu: a #CuiModeMenu to pop-up
 * @parent_widget: the parent widget of the menu
 * @event: the event that caused this opertaion
 *
 * Starts up a #CuiFileMenu on the given location of the screen. This function
 * is very important since there are a lot of file names on the screen and the
 * file menu is a rather complicated and powerfull user interface element.
 */
void
cui_mode_menu_popup(CuiModeMenu *cui_mode_menu,
		GtkWidget *parent_widget,
		GdkEventButton *event)
{
	GtkWidget *menu_item;
	Terminal  *terminal;
	gboolean is;
	g_assert(cui_mode_menu != NULL && 
			parent_widget != NULL && 
			event != NULL);
	g_assert(cui_mode_menu->menu != NULL);
#if 0
	/*
	 * We only uncompress the files we think are compressed.
	 */
	is = cui_filename_is_compressed(cui_file_menu->file);
	menu_item = lookup_widget(cui_file_menu->menu, "file_uncompress");
	gtk_widget_set_sensitive(menu_item, is);
	/*
	 * We compress only the files without compression.
	 */
	menu_item = lookup_widget(cui_file_menu->menu, "file_archive");
	gtk_widget_set_sensitive(menu_item, !is);
#endif
	/*
	 * Setting some data for the callback function.
	 */
	terminal = cui_widget_get_terminal(CUI_WIDGET(cui_mode_menu));
	g_object_set_data(G_OBJECT(cui_mode_menu->menu), 
			"cui_widget", cui_mode_menu);
	g_object_set_data(G_OBJECT(cui_mode_menu->menu), 
			"terminal", terminal);
	g_object_set_data(G_OBJECT(cui_mode_menu->menu), 
			"command_generator", terminal->command_generator);
	/*
	 * Starting up the popup menu.
	 */
	gtk_menu_popup(GTK_MENU(cui_mode_menu->menu), NULL, NULL, NULL, NULL, 
			event->button, event->time);
}

/*****************************************************************************
 * Instantiable classed types: CuiUserMenu                                   *
 *                                                                           *
 *****************************************************************************/
#define PROPERTY_USERNAME 1

static void
cui_user_menu_set_property(
		GObject      *object, 
		guint         property_id, 
		const GValue *value, 
		GParamSpec   *pspec) 
{
	CuiUserMenu *self = (CuiUserMenu *)object;
	
	switch(property_id){
		case PROPERTY_USERNAME:
			self->username = g_value_dup_string(value);
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID(
					object, property_id, pspec);
	}		
}

static void
cui_user_menu_get_property(
		GObject      *object, 
		guint         property_id, 
		GValue       *value, 
		GParamSpec   *pspec)
{
	CuiUserMenu *self = (CuiUserMenu *) object;
	switch(property_id){
		case PROPERTY_USERNAME:
			g_value_set_pointer(value, self->username);
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID(
					object, property_id, pspec);
	}
}

static void
cui_user_menu_class_init(
		gpointer g_class, 
		gpointer g_class_data)
{
	GParamSpec *pspec;
	GObjectClass *gobject_class = G_OBJECT_CLASS (g_class);
	CuiUserMenuClass *klass = CUI_USER_MENU_CLASS(g_class);	
	
	gobject_class->set_property = cui_file_menu_set_property;
  	gobject_class->get_property = cui_file_menu_get_property;
	klass->menu = create_user_popup_menu();
	
	pspec = g_param_spec_pointer(
			"username", "login-name",
                        "The login name of the given user.",
                        G_PARAM_READWRITE);
	g_object_class_install_property(gobject_class, 
			PROPERTY_USERNAME, 
			pspec);
}

static void
cui_user_menu_instance_init(
		GTypeInstance   *instance, 
		gpointer         g_class)
{
	CuiUserMenu *self = (CuiUserMenu *)instance;
	CuiUserMenuClass *klass = (CuiUserMenuClass *)g_class;
	
	self->menu = klass->menu;
	self->username = NULL;
}

GType 
cui_user_menu_get_type(void)
{
	static GType type = 0;
	if (type == 0) {
		static const GTypeInfo info = {
			sizeof(CuiUserMenuClass),
			NULL,
			NULL,
			(GClassInitFunc) cui_user_menu_class_init,
			NULL,
			NULL,
			sizeof(CuiUserMenu),
			0,
			(GInstanceInitFunc) cui_user_menu_instance_init 
		};
		
		type = g_type_register_static(CUI_POPUP_MENU_TYPE, 
				"CuiUserMenu", 
				&info, 
				0);
	}
	return type;
}

/**
 * cui_file_menu_new:
 * @returns: a new #CuiUserMenu
 *
 * Creates a new use menu.
 *
 */
CuiWidget *
cui_user_menu_new(const gchar *username)
{
	CuiWidget *menu;
	menu = g_object_new(CUI_USER_MENU_TYPE, NULL);
	CUI_USER_MENU(menu)->username = g_strdup(username);
	g_object_notify(G_OBJECT(menu), "username");
	return menu;
}


/**
 * cui_user_menu_popup:
 * @cui_user_menu: a #CuiUserMenu to pop-up
 * @parent_widget: the parent widget of the menu
 * @event: the event that caused this opertaion
 *
 * Starts up a #CuiUserMenu on the given location of the screen. 
 */
void
cui_user_menu_popup(CuiUserMenu *cui_user_menu,
		GtkWidget *parent_widget,
		GdkEventButton *event)
{
	GtkWidget *menu_item;
	Terminal  *terminal;
	gboolean is;
	g_assert(cui_user_menu != NULL && 
			parent_widget != NULL && 
			event != NULL);
	g_assert(cui_user_menu->menu != NULL);
	/*
	 * Setting some data for the callback function.
	 */
	terminal = cui_widget_get_terminal(CUI_WIDGET(cui_user_menu));
	g_object_set_data(G_OBJECT(cui_user_menu->menu), 
			"cui_widget", cui_user_menu);
	g_object_set_data(G_OBJECT(cui_user_menu->menu), 
			"terminal", terminal);
	g_object_set_data(G_OBJECT(cui_user_menu->menu), 
			"command_generator", terminal->command_generator);
	g_object_set_data(G_OBJECT(cui_user_menu->menu), 
			"username", cui_user_menu->username); 
	/*
	 * Starting up the popup menu.
	 */
	gtk_menu_popup(GTK_MENU(cui_user_menu->menu), NULL, NULL, NULL, NULL, 
			event->button, event->time);
}


/*****************************************************************************
 * Instantiable classed types: CuiRectangle                                  *
 *                                                                           *
 *****************************************************************************/
#define PROPERTY_COLOR 1
#define PROPERTY_HCOLOR 2
#define PROPERTY_LCOLOR 3
#define PROPERTY_IS_DND_TARGET 4
#define PROPERTY_IS_DND_SOURCE 5
#define PROPERTY_IS_HIGHLIGHTED 6
#define PROPERTY_TARGETLIST 7
#define PROPERTY_TARGET_OBJECT 8

static void
cui_rectangle_set_property (
		GObject      *object, 
		guint         property_id, 
		const GValue *value, 
		GParamSpec   *pspec) 
{
	CuiRectangle *self = (CuiRectangle *) object;
	switch(property_id){
		case PROPERTY_COLOR:
			g_free(self->color);
			self->color = g_memdup(g_value_get_pointer(value),
					sizeof(GdkColor));
			break;
			
		case PROPERTY_HCOLOR:
			g_free(self->hcolor);
			self->hcolor = g_memdup(g_value_get_pointer(value),
					sizeof(GdkColor));
			break;
			
		case PROPERTY_LCOLOR:
			g_free(self->lcolor);
			self->lcolor = g_memdup(g_value_get_pointer(value),
					sizeof(GdkColor));
			break;
		
		case PROPERTY_TARGETLIST:
			self->target_list = g_value_get_pointer(value);
			break;
			
		case PROPERTY_TARGET_OBJECT:
			if (self->target_object != NULL)
				g_object_unref(self->target_object);
			self->target_object = g_value_get_pointer(value);
			g_object_ref(self->target_object);
			break;
			
		case PROPERTY_IS_DND_TARGET:
			self->is_dnd_target = g_value_get_boolean(value);
			break;
			
		case PROPERTY_IS_DND_SOURCE:
			self->is_dnd_source = g_value_get_boolean(value);
			break;
			
		case PROPERTY_IS_HIGHLIGHTED:
			self->is_highlighted = g_value_get_boolean(value);
			break;
			
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID(
					object, property_id, pspec);
	}
}

static void
cui_rectangle_get_property (
		GObject      *object, 
		guint         property_id, 
		GValue       *value, 
		GParamSpec   *pspec)
{
	CuiRectangle *self = (CuiRectangle *) object;
	switch(property_id){
		case PROPERTY_COLOR:
			g_value_set_pointer(value, self->color);
			break;
			
		case PROPERTY_HCOLOR:
			g_value_set_pointer(value, self->hcolor);
			break;
			
		case PROPERTY_LCOLOR:
			g_value_set_pointer(value, self->lcolor);
			break;
		
		case PROPERTY_TARGETLIST:
			g_value_set_pointer(value, self->target_list);
			break;
			
		case PROPERTY_TARGET_OBJECT:	
			g_value_set_pointer(value, self->target_object);
			break;
				
		case PROPERTY_IS_DND_TARGET:
			g_value_set_boolean(value, self->is_dnd_target);
			break;
		
		case PROPERTY_IS_DND_SOURCE:
			g_value_set_boolean(value, self->is_dnd_source);
			break;
			
		case PROPERTY_IS_HIGHLIGHTED:
			g_value_set_boolean(value, self->is_highlighted);
			break;
			
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID(
					object, property_id, pspec);
	}
}

/*
 * 
 */
static void 
cui_rectangle_data_received_for_file(
		CuiRectangle     *rectangle,
		GtkSelectionData *data,
		CuiFilename      *target)
{
	gchar        *name = gdk_atom_name(data->type);
	Terminal     *t = cui_widget_get_terminal(CUI_WIDGET(rectangle));
	gchar        *basename;
	gchar        *directory;
	CuiFilename  *source;
	CuiGenerator *generator;
	gchar        *wd;
	gchar        *newname;
	
	basename = g_path_get_basename(data->data);
	directory = g_path_get_dirname(data->data);
	source = cui_filename_new(basename, directory);
	
#if 0
	g_message("%s(): start", __func__);	
	g_message("%s(): Got a '%s'", __func__, name);
	g_message("%s(): Filename: '%s'", __func__, data->data);
	g_message("%s(): basename: '%s'", __func__, basename);	
	g_message("%s(): directory: '%s'", __func__, directory);	
#endif	
	
	g_assert(t != NULL);
	generator = t->command_generator;

	wd = (gchar *)terminal_get_working_directory(t);
	newname = cui_filename_get_canonical_filename(target, wd);
	
	if (strcmp(name, "file/name") == 0) {
		cui_generator_emit_file_command_with_arg(generator, FileMove, source, newname);
		cui_widget_activate(CUI_WIDGET(rectangle));
		goto end_of_job;
	}

	if (strcmp(name, "file/mode") == 0) {
		cui_generator_emit_file_command_with_arg(generator, FileCopyMode, source, newname);
		cui_widget_activate(CUI_WIDGET(rectangle));
		goto end_of_job;
	}
	
	if (strcmp(name, "file/owner") == 0) {
		cui_generator_emit_file_command_with_arg(generator, FileCopyOwner, source, newname);
		cui_widget_activate(CUI_WIDGET(rectangle));
		goto end_of_job;
	}
	
	if (strcmp(name, "file/group") == 0) {
		cui_generator_emit_file_command_with_arg(generator, FileCopyGroupOwner, source, newname);
		cui_widget_activate(CUI_WIDGET(rectangle));
		goto end_of_job;
	}
			
end_of_job:
	g_free(name);
}


/*
 * default signal handler for the "drop-data-received" signal. This signal sent
 * when a GDK_ACTION_COPY drop action has been detected for the given
 * CuiRectangle which is the target of the operation.
 *
 */
static void
cui_rectangle_drop_data_received_handler(
		CuiRectangle      *rectangle, 
		GtkSelectionData  *data, 
		gpointer user_data)
{
	g_return_if_fail(CUI_IS_RECTANGLE(rectangle));
	
	if (CUI_IS_FILENAME(rectangle->target_object))
		cui_rectangle_data_received_for_file(rectangle, data, CUI_FILENAME(rectangle->target_object));
	else 
		g_warning("%s(): Unknown DnD target object: %p", __func__, rectangle->target_object);
}

static gboolean
cui_rectangle_expose_handler(CuiRectangle *rec,
		gint arg1,
		gpointer user_data)
{
	Background *bg;
	GdkColor *color, *hcolor, *lcolor;
	GdkRectangle *position = NULL;
	gint border = 1;
	gint n;
	
	/*
	 * Collecting all needed information
	 */
	g_return_val_if_fail(rec != NULL, FALSE);
	bg = cui_widget_get_background(CUI_WIDGET(rec));
	g_return_val_if_fail(bg != NULL, FALSE);	
	position = cui_widget_get_area_window_coords(CUI_WIDGET(rec));
	g_return_val_if_fail(position != NULL, FALSE);
	
	if (position->width <= 0 || position->height <= 0)
		return FALSE;
	
	color = cui_rectangle_get_color(CUI_RECTANGLE(rec));
	hcolor = cui_rectangle_get_high_color(CUI_RECTANGLE(rec));
	lcolor = cui_rectangle_get_low_color(CUI_RECTANGLE(rec));

	/*
	 * The highlighted rectangle is something we use to show
	 * the focus.
	 */
	if (rec->is_highlighted){
		GtkStyle *style = gtk_widget_get_style(bg->widget); 
#if 0
		gtk_paint_flat_box(style,              /* style */
				bg->window,            /* window */
				GTK_STATE_SELECTED,    /* statetype */
				GTK_SHADOW_ETCHED_IN,  /* shadowtype */
				NULL,                  /* clip area */
				bg->widget,            /* widget */ 
				NULL,                  /* style detail */
                                position->x,           /* x */
				position->y,           /* y */
                                position->width - 1,   /* width */
                                position->height - 1); /* height */
#endif
		gtk_paint_focus(style,                 /* style */
				bg->window,            /* window */
				GTK_STATE_SELECTED,    /* statetype */
				NULL,                  /* clip area */
				bg->widget,            /* widget */ 
				NULL,                  /* style detail */
                                position->x,           /* x */
				position->y,           /* y */
                                position->width - 1,   /* width */
                                position->height - 1); /* height */
	}

	if (color == NULL || hcolor == NULL || lcolor == NULL)
		return FALSE;


	//if (CUI_IS_FRAME(rec)){
		gdk_gc_set_foreground(bg->gc, color);
		gdk_draw_rectangle(bg->window, bg->gc, FALSE, 
				position->x, position->y, 
				position->width, position->height);
		
		gdk_gc_set_foreground(bg->gc, lcolor);
		gdk_draw_rectangle(bg->window, bg->gc, FALSE, 
				position->x + 1, position->y + 1, 
				position->width - 2, position->height - 2);
		
		gdk_gc_set_foreground(bg->gc, hcolor);
		gdk_draw_rectangle(bg->window, bg->gc, FALSE, 
				position->x + 2, position->y + 2, 
				position->width - 4, position->height - 4);
	//} /* if this is a rectangle */
	
	/*
	 * The cui_widget_get_area_window_coords returning an allocated memory
	 * pointer, however, the cui_widget_get_area is not. Stupid...
	 */
	g_free(position);
	return TRUE;
}

static guint expose_id = 0;
static guint drop_id = 0;

/**
 * cui_rectangle_get_expose_signal_id:
 * @returns: the numerical id of the "expose" signal
 *
 * Returns the id of the "expose" signal. This function is for internal use
 * only.
 */
guint
cui_rectangle_get_expose_signal_id(void)
{
	return expose_id;
}

static void
cui_rectangle_base_class_init(CuiRectangleClass *klass)
{
	static gboolean initialized = FALSE;
	GClosure *expose_closure;
	GClosure *drop_closure;
	GType params[5];

	if (!initialized){
		/*
		 * Signals creation: drop-data-received
		 */
		params[0] = G_TYPE_POINTER;
		params[1] = G_TYPE_POINTER;
		drop_closure = g_cclosure_new(
			G_CALLBACK(cui_rectangle_drop_data_received_handler),
			NULL, /* user data */
			NULL  /* destroy data */
			);
		drop_id = g_signal_newv(
				"drop-data-received",
				CUI_RECTANGLE_TYPE,
				G_SIGNAL_RUN_LAST,
				drop_closure,
				NULL, /* Accumulator function */
				NULL, /* Accumulator data */
				g_cclosure_marshal_VOID__POINTER, 
				G_TYPE_NONE, /* return value */
				1, /* param number */
				params /*param type array */
				);
		/* 
		 * Signals creation: expose
		 */
		expose_closure = g_cclosure_new(
				G_CALLBACK(cui_rectangle_expose_handler),
				NULL, /* user data */
				NULL  /* destroy data */
				);
		params[0] = G_TYPE_INT;
		expose_id = g_signal_newv(
				"expose",
				CUI_RECTANGLE_TYPE,
				G_SIGNAL_RUN_LAST,
				expose_closure,
				NULL, /* Accumulator function */
				NULL, /* Accumulator data */
				g_cclosure_marshal_BOOLEAN__FLAGS, 
				G_TYPE_BOOLEAN, /* return value */
				1, /* param number */
				params /*param type array */
				);
		initialized = TRUE;
	}
}

static void
cui_rectangle_class_init(
		gpointer g_class, 
		gpointer g_class_data)
{
	GParamSpec *pspec;
	GObjectClass *gobject_class = G_OBJECT_CLASS (g_class);
	CuiRectangleClass *klass = CUI_RECTANGLE_CLASS(g_class);	

	gobject_class->set_property = cui_rectangle_set_property;
  	gobject_class->get_property = cui_rectangle_get_property;

	pspec = g_param_spec_pointer(
			"color", "color-pointer",
                        "The color of the rectangle.",
                        G_PARAM_READWRITE);
	g_object_class_install_property(gobject_class, 
			PROPERTY_COLOR, 
			pspec);
	
	pspec = g_param_spec_pointer(
			"high-color", "high-color-pointer",
                        "The color of the upper left part of the rectangle.",
                        G_PARAM_READWRITE);
	g_object_class_install_property(gobject_class, 
			PROPERTY_HCOLOR, 
			pspec);
	
	pspec = g_param_spec_pointer(
			"low-color", "low-color-pointer",
                        "The color of the upper right part of the rectangle.",
                        G_PARAM_READWRITE);
	g_object_class_install_property(gobject_class, 
			PROPERTY_LCOLOR, 
			pspec);
	
	pspec = g_param_spec_pointer(
			"target_list", "target-list",
                        "The GtkTargetList shows what can this rectangle "
			"accept as a DnD target.",
                        G_PARAM_READWRITE);
	g_object_class_install_property(gobject_class, 
			PROPERTY_TARGETLIST, 
			pspec);

	pspec = g_param_spec_boolean(
			"is_dnd_target", "is-dnd-target",
                        "If the rectangle can be a Drag&Drop target.",
			FALSE, G_PARAM_READWRITE);
	g_object_class_install_property(gobject_class, 
			PROPERTY_IS_DND_TARGET, 
			pspec);
	
	pspec = g_param_spec_boolean(
			"is_dnd_source", "is-dnd-source",
                        "If the rectangle can be a Drag&Drop source.",
			FALSE, G_PARAM_READWRITE);
	g_object_class_install_property(gobject_class, 
			PROPERTY_IS_DND_SOURCE, 
			pspec);
	
	pspec = g_param_spec_boolean(
			"is_highlighted", "is-highlighted",
                        "If the rectangle is highlighted (DnD motion).",
			FALSE, G_PARAM_READWRITE);
	g_object_class_install_property(gobject_class, 
			PROPERTY_IS_HIGHLIGHTED, 
			pspec);
	
	pspec = g_param_spec_pointer(
			"target_object", "target-object",
                        "The GObject we use as a DnD target.",
                        G_PARAM_READWRITE);
	g_object_class_install_property(gobject_class, 
			PROPERTY_TARGET_OBJECT, 
			pspec);
}


static void
cui_rectangle_instance_init(
		GTypeInstance   *instance, 
		gpointer         g_class)
{
	CuiRectangle *self = (CuiRectangle *)instance;	
}

GType 
cui_rectangle_get_type(void)
{
	static GType type = 0;
	
	if (type == 0) {
		static const GTypeInfo info = {
			sizeof(CuiRectangleClass),
			(GBaseInitFunc) cui_rectangle_base_class_init,
			NULL,	/* base_finalize */
			(GClassInitFunc) cui_rectangle_class_init,
			NULL,	/* class_finalize */
			NULL,	/* class_data */
			sizeof(CuiRectangle),
			0,	/* n_preallocs */
			/* instance_init */
			(GInstanceInitFunc) cui_rectangle_instance_init 
		};
		
		type = g_type_register_static(CUI_WIDGET_TYPE, 
				"CuiRectangle", 
				&info, 
				0);
	}
	return type;
}

/**
 * cui_rectangle_new:
 * @returns: a new #CuiRectangle as #CuiWidget
 *
 * Creates a new rectangle.
 */
CuiWidget *
cui_rectangle_new(void)
{
	CuiWidget *rectangle;
	rectangle = g_object_new(CUI_RECTANGLE_TYPE, NULL);
	return rectangle;
}

/**
 * cui_rectangle_set_is_dnd_source:
 * @rectangle: a #CuiRectangle
 * @is_dnd_source: the new value of the is_dnd_source property
 * 
 * Sets the "is_dnd_source" property which is TRUE if this rectangle can be a
 * Drag&Drop source.
 */
void 
cui_rectangle_set_is_dnd_source(CuiRectangle *rectangle, 
		gboolean is_dnd_source)
{
	GValue val = {0, };
	g_return_if_fail(CUI_IS_RECTANGLE(rectangle));
	g_value_init(&val, G_TYPE_BOOLEAN);
	g_value_set_boolean(&val, is_dnd_source);
	g_object_set_property(G_OBJECT(rectangle), "is_dnd_source", &val);
}

/**
 * cui_rectangle_set_dnd_source_data:
 * @rectangle: a #CuiRectangle
 * @source_type: a #TargetType type source type identifier
 * @source_data: the dragged data pointer
 * @source_size: the size of the source_data
 *
 * Sets the information needed to start a Drag&Drop operation. Every DnD source
 * is capable of initializing a one type DnD operation.
 */
void
cui_rectangle_set_dnd_source_data(
		CuiRectangle *rectangle,
		TargetType    source_type,
		gpointer      source_data,
		gsize         source_size)
{
	g_return_if_fail(CUI_IS_RECTANGLE(rectangle));
	rectangle->source_type = source_type;
	rectangle->source_data = source_data;
	rectangle->source_size = source_size;
}

/**
 * cui_rectangle_set_is_dnd_target:
 * @rectangle: a #CuiRectangle
 * @is_dnd_target: the new value of the is_dnd_target property
 * 
 * Sets the "is_dnd_target" property which is TRUE if this rectangle can be a
 * Drag&Drop target.
 */
void 
cui_rectangle_set_is_dnd_target(CuiRectangle *rectangle, 
		gboolean is_dnd_target)
{
	GValue val = {0, };
	g_return_if_fail(CUI_IS_RECTANGLE(rectangle));
	g_value_init(&val, G_TYPE_BOOLEAN);
	g_value_set_boolean(&val, is_dnd_target);
	g_object_set_property(G_OBJECT(rectangle), "is_dnd_target", &val);
}

/**
 * cui_rectangle_is_dnd_source:
 * @rectangle: a #CuiRectangle
 * @returns: TRUE if the rectangle can be a Drag&Drop source
 *
 * Returns TRUE if the rectangle can be a Drag&Drop source.
 */
gboolean
cui_rectangle_is_dnd_source(CuiRectangle *rectangle)
{
	gboolean retval;
	g_return_val_if_fail(CUI_IS_RECTANGLE(rectangle), FALSE);
	g_object_get(G_OBJECT(rectangle), "is_dnd_source", &retval, NULL);
	return retval;
}

/**
 * cui_rectangle_is_dnd_target:
 * @rectangle: a #CuiRectangle
 * @returns: TRUE if the rectangle can be a Drag&Drop target
 *
 * Returns TRUE if the rectangle can be a Drag&Drop target.
 */
gboolean
cui_rectangle_is_dnd_target(CuiRectangle *rectangle)
{
	gboolean retval;
	g_return_val_if_fail(CUI_IS_RECTANGLE(rectangle), FALSE);
	g_object_get(G_OBJECT(rectangle), "is_dnd_target", &retval, NULL);
	return retval;
}

/**
 * cui_rectangle_set_highlighted:
 * @rectangle: a #CuiRectangle
 * @highlighted: if the rectangle widget is highlighted
 * @returns: the original highlight status of the rectangle
 *
 * Sets the highlight status of the rectangle. The rectangle is highlighted
 * when the mouse is moving over it during a Drag&Drop operation. The function
 * redrawi the rectangle if necessary.
 */
gboolean
cui_rectangle_set_highlighted(CuiRectangle *rectangle, 
		gboolean highlighted)
{
	CuiWidget *widget = rectangle;
	gboolean retval;
	gboolean redraw;
	GValue val = {0, };
	
	g_return_val_if_fail(CUI_IS_RECTANGLE(rectangle), FALSE);
	g_object_get(G_OBJECT(rectangle), "is_highlighted", &retval, NULL);
	if (retval == highlighted)
		return retval;
	
	g_value_init(&val, G_TYPE_BOOLEAN);
	g_value_set_boolean(&val, highlighted);
	g_object_set_property(G_OBJECT(rectangle), "is_highlighted", &val);
	
	// Invalidating.
	if (widget->background != NULL && widget->area != NULL) {
		GdkRectangle *rec;
		rec = cui_widget_get_area_window_coords(widget);
		gdk_window_invalidate_rect(widget->background->window, rec, TRUE);
		g_free(rec);
	}
	
	g_signal_emit_by_name(rectangle, "expose", 1, &redraw);
	return retval;
}

/**
 * cui_rectangle_set_target_list:
 * @rec: a #CuiRectangle
 * @list: a new #GtkTargetList for this rectangle
 * 
 * Sets the target list of the rectangle. This list is storing the data types
 * this rectangle can handle as DnD target.
 */
void
cui_rectangle_set_target_list(CuiRectangle *rec, GtkTargetList *list)
{
	GValue val = {0,};
	
	g_return_if_fail(CUI_IS_RECTANGLE(rec));
	g_value_init(&val, G_TYPE_POINTER);
	g_value_set_pointer(&val, list);
	g_object_set_property(G_OBJECT(rec), "target_list", &val);
}

/**
 * cui_rectangle_get_target_list:
 * @rec: a #CuiRectangle
 * @returns: the target list of the rectangle
 *
 * Returns the #GtkTargetList of the rectangle. This list contains the DnD types
 * this rectangle can handle as a target.
 */
GtkTargetList *
cui_rectangle_get_target_list(CuiRectangle *rec)
{
	GtkTargetList *list;
	g_return_val_if_fail(CUI_IS_RECTANGLE(rec), NULL);
	g_object_get(G_OBJECT(rec), "target_list", &list, NULL);
	return list;
}

/**
 * cui_rectangle_set_color:
 * @rec: a #CuiRectangle widget
 * @color_name: the color for the widget
 * @high_color_name: the high color
 * @low_color_name: the low color
 *
 * Sets the three colors for the frame. The function also allocates these colors
 * for the #Background of the widget.
 */
void
cui_rectangle_set_color(CuiRectangle *rec, 
		const char *color_name,
		const char *high_color_name,
		const char *low_color_name)
{
	GValue val = {0,};
	Background *bg;
	GdkColor color;
	g_return_if_fail(CUI_IS_RECTANGLE(rec));

	bg = cui_widget_get_background(CUI_WIDGET(rec));
	if (bg == NULL){
		g_warning("%s(): background is NULL unable to alloc color", __func__);
		return;
	}

	g_value_init(&val, G_TYPE_POINTER);
	
	gdk_color_parse(color_name, &color);
	gdk_colormap_alloc_color(bg->colormap, &color, FALSE, TRUE);
	g_value_set_pointer(&val, &color);
	g_object_set_property(G_OBJECT(rec), "color", &val);

	gdk_color_parse(high_color_name, &color);
	gdk_colormap_alloc_color(bg->colormap, &color, FALSE, TRUE);
	g_value_set_pointer(&val, &color);
	g_object_set_property(G_OBJECT(rec), "high-color", &val);
	
	gdk_color_parse(low_color_name, &color);
	gdk_colormap_alloc_color(bg->colormap, &color, FALSE, TRUE);
	g_value_set_pointer(&val, &color);
	g_object_set_property(G_OBJECT(rec), "low-color", &val);
}

/**
 * cui_rectangle_get_color:
 * @rec: a #CuiRectangle
 * @returns: the color of the rectangle as a #GdkColor
 *
 * Returns the color of the rectangle.
 */
GdkColor *
cui_rectangle_get_color(CuiRectangle *rec)
{
	GdkColor *color;
	g_return_val_if_fail(CUI_IS_RECTANGLE(rec), NULL);
	g_object_get(G_OBJECT(rec), "color", &color, NULL);
	return color;
}

/**
 * cui_rectangle_get_high_color:
 * @rec: a #CuiRectangle
 * @returns: the color of the rectangle as a #GdkColor
 *
 * Returns the color of the rectangle.
 */
GdkColor *
cui_rectangle_get_high_color(CuiRectangle *rec)
{
	GdkColor *color;
	g_return_val_if_fail(CUI_IS_RECTANGLE(rec), NULL);
	g_object_get(G_OBJECT(rec), "high-color", &color, NULL);
	return color;
}

/**
 * cui_rectangle_get_low_color:
 * @rec: a #CuiRectangle
 * @returns: the color of the rectangle as a #GdkColor
 *
 * Returns the color of the rectangle.
 */
GdkColor *
cui_rectangle_get_low_color(CuiRectangle *rec)
{
	GdkColor *color;
	g_return_val_if_fail(CUI_IS_RECTANGLE(rec), NULL);
	g_object_get(G_OBJECT(rec), "low-color", &color, NULL);
	return color;
}

/**
 * cui_rectangle_set_target_object:
 * @rec: a #CuiRectangle
 * @object: the new target object
 * 
 * Sets the Drag&Drop target object for the rectangle. The rectangle will hold a
 * reference of the object until we change the target.
 */
void
cui_rectangle_set_target_object(CuiRectangle *rec, GObject *object)
{
	GValue val = {0,};
	
	g_return_if_fail(CUI_IS_RECTANGLE(rec));
	g_value_init(&val, G_TYPE_POINTER);
	g_value_set_pointer(&val, object);
	g_object_set_property(G_OBJECT(rec), "target_object", &val);
}

/**
 * cui_rectangle_redraw:
 * @rectangle: the #CuiRectangle to redraw
 * @debug: if this argument is not 0 the area will be highlighted
 * @returns: the return value of the signal handler
 *
 * Emits the expose event for the widget to redraw the screen.
 */
gboolean 
cui_rectangle_redraw(CuiRectangle *rectangle, gint debug)
{
	gboolean retval;
	g_return_val_if_fail(CUI_IS_RECTANGLE(rectangle), FALSE);
	g_signal_emit_by_name(rectangle, "expose", debug, &retval);
	return retval;	
}


/*****************************************************************************
 * Class: CuiRectangleEmboss: A simple rectangle with 3D look.               *
 *                                                                           *
 *****************************************************************************/
static void
cui_rectangle_emboss_set_property (
		GObject      *object, 
		guint         property_id, 
		const GValue *value, 
		GParamSpec   *pspec) 
{
}

static void
cui_rectangle_emboss_get_property (
		GObject      *object, 
		guint         property_id, 
		GValue       *value, 
		GParamSpec   *pspec)
{
	//CuiFrame *self = (CuiFrame *) object;
	//g_value_set_string (value, "value");
}


static gboolean
cui_rectangle_emboss_expose_handler(CuiRectangle *rec,
		gint arg1,
		gpointer user_data)
{
	Background *bg;
	GdkColor *color, *hcolor, *lcolor;
	GdkRectangle *position = NULL;
	gint border = 1;
	gint n;
	
	/*
	 * Collecting all needed information
	 */
	g_return_val_if_fail(rec != NULL, FALSE);
	bg = cui_widget_get_background(CUI_WIDGET(rec));
	g_return_val_if_fail(bg != NULL, FALSE);	
	position = cui_widget_get_area_window_coords(CUI_WIDGET(rec));
	g_return_val_if_fail(position != NULL, FALSE);
	/*
	 * If this widget is too far we don't have to bother.
	 */
	if (position->width <= 0 || position->height <= 0)
		return FALSE;
	
	color = cui_rectangle_get_color(CUI_RECTANGLE(rec));
	hcolor = cui_rectangle_get_high_color(CUI_RECTANGLE(rec));
	lcolor = cui_rectangle_get_low_color(CUI_RECTANGLE(rec));

	if (color == NULL || hcolor == NULL || lcolor == NULL)
		return FALSE;
	/*
	 * Now drawing. 
	 */

	gdk_gc_set_foreground(bg->gc, color);
	gdk_draw_rectangle(bg->window, bg->gc, TRUE, 
			position->x + border, 
			position->y + border, 
			position->width - 2 * border + 1, 
			position->height - 2 * border + 1);

	gdk_gc_set_foreground(bg->gc, hcolor);
	for (n = 0; n < border; ++n) {
		gdk_draw_line(bg->window, bg->gc, 
				position->x, 
				position->y + n, 
				position->x + position->width, 
				position->y + n);
		gdk_draw_line(bg->window, bg->gc, 
				position->x + n, 
				position->y, 
				position->x + n, 
				position->y + position->height);
	}

	gdk_gc_set_foreground(bg->gc, lcolor);
	for (n = 0; n < border; ++n) {
		gdk_draw_line(bg->window, bg->gc,
			      position->x + position->width - n,
			      position->y + n, 
			      position->x + position->width - n, 
			      position->y + position->height);
		gdk_draw_line(bg->window, bg->gc,
			      position->x + n,
			      position->y + position->height - n, 
			      position->x + position->width, 
			      position->y + position->height - n);
	}

	/*
	 * The cui_widget_get_area_window_coords returning an allocated memory
	 * pointer, however, the cui_widget_get_area is not. Stupid...
	 */
	g_free(position);
	return TRUE;
}


static void
cui_rectangle_emboss_class_init(
		gpointer g_class, 
		gpointer g_class_data)
{
	static gboolean   initialized = FALSE;
	GObjectClass     *gobject_class = G_OBJECT_CLASS (g_class);
	GClosure         *class_closure;
	GType             params[1];
	guint             id;
	
	if (!initialized){
		/* interface signals creation*/
		class_closure = g_cclosure_new(
				G_CALLBACK(cui_rectangle_emboss_expose_handler),
				NULL, /* user data */
				NULL  /* destroy data */
				);
		params[0] = G_TYPE_INT;
		g_signal_override_class_closure(
				cui_rectangle_get_expose_signal_id(),
				CUI_RECTANGLE_EMBOSS_TYPE,
                                class_closure);
		initialized = TRUE;
	}
}


static void
cui_rectangle_emboss_instance_init(
		GTypeInstance   *instance, 
		gpointer         g_class)
{
}


GType 
cui_rectangle_emboss_get_type(void)
{
	static GType type = 0;
	
	if (type == 0) {
		static const GTypeInfo info = {
			sizeof(CuiRectangleEmbossClass),
			(GBaseInitFunc) NULL, 
			NULL,
			(GClassInitFunc) cui_rectangle_emboss_class_init,
			NULL,
			NULL,
			sizeof(CuiRectangleEmboss),
			128,
			/* instance_init */
			(GInstanceInitFunc) cui_rectangle_emboss_instance_init 
		};
		
		type = g_type_register_static(CUI_RECTANGLE_TYPE, 
				"CuiRectangleEmboss", 
				&info, 
				0);
	}
	return type;
}

/**
 * cui_rectangle_emboss_new:
 * @returns: a new embossed rectangle
 *
 */
CuiWidget *
cui_rectangle_emboss_new(void)
{
	CuiWidget *rec;
	rec = g_object_new(CUI_RECTANGLE_EMBOSS_TYPE, NULL);
	//cui_widget_set_active(rec, FALSE);
	return rec;
}


/*****************************************************************************
 * Class: CuiSizeBar: A horizontal bar representing sizes with 3D look.      *
 *                                                                           *
 *****************************************************************************/
#define PROPERTY_DECOR     1
#define PROPERTY_PERCENT   2
#define PROPERTY_TURNPOINT 3

static void
_cui_size_bar_set_colors(CuiSizeBar *self)
{
	Background *bg;

	bg = cui_widget_get_background(CUI_WIDGET(self));
	if (bg == NULL){
		g_warning("%s(): background is NULL unable to alloc color", __func__);
		return;
	}

	if (self->turnpoint != -1.0) {
		if (self->percent < self->turnpoint){
			gdk_color_parse("White", &(self->colors[0]));
			gdk_colormap_alloc_color(bg->colormap, &(self->colors[0]), FALSE, TRUE);
			gdk_color_parse("Green", &(self->colors[1]));
			gdk_colormap_alloc_color(bg->colormap, &(self->colors[1]), FALSE, TRUE);
			cui_rectangle_set_color(CUI_RECTANGLE(self), "#87f387", "#42f142", "#bcf1bc"); 
		}else{
			gdk_color_parse("White", &(self->colors[0]));
			gdk_colormap_alloc_color(bg->colormap, &(self->colors[0]), FALSE, TRUE);
			gdk_color_parse("Red", &(self->colors[1]));
			gdk_colormap_alloc_color(bg->colormap, &(self->colors[1]), FALSE, TRUE);
			cui_rectangle_set_color(CUI_RECTANGLE(self), "#f3acac", "#f37777", "#f3cbcb"); 
		}
	} else {
		gdk_color_parse("White", &(self->colors[0]));
		gdk_colormap_alloc_color(bg->colormap, &(self->colors[0]), FALSE, TRUE);
		gdk_color_parse("Cyan", &(self->colors[1]));
		gdk_colormap_alloc_color(bg->colormap, &(self->colors[1]), FALSE, TRUE);
	   	cui_rectangle_set_color(CUI_RECTANGLE(self), "#8cfbef", "#45ebf9", "#b6fbf2"); 
	}
}

static void
cui_size_bar_set_property (
		GObject      *object, 
		guint         property_id, 
		const GValue *value, 
		GParamSpec   *pspec) 
{
	CuiSizeBar *self = (CuiSizeBar *) object;
	
	switch(property_id){
		case PROPERTY_NAME:
			self->decor = g_value_get_int(value);
			break;
			
		case PROPERTY_PERCENT:
			self->percent = g_value_get_double(value);
			_cui_size_bar_set_colors(self);
	   		cui_widget_set_width_by_percent(CUI_WIDGET(self), self->percent);
			break;
			
		case PROPERTY_TURNPOINT:
			self->turnpoint = g_value_get_double(value);
			_cui_size_bar_set_colors(self);
			break;
			
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID(object, property_id, pspec);
	}
}

static void
cui_size_bar_get_property (
		GObject      *object, 
		guint         property_id, 
		GValue       *value, 
		GParamSpec   *pspec)
{
	//CuiFrame *self = (CuiFrame *) object;
	//g_value_set_string (value, "value");
	g_warning("%s(): Unimplemented", __func__);
}


static gboolean
cui_size_bar_expose_handler(CuiSizeBar *bar,
		gint arg1,
		gpointer user_data)
{
	Background *bg;
	GdkColor *color, *hcolor, *lcolor;
	GdkRectangle *position = NULL;
	gint border = 1;
	gint n;
	gint dx;
	gint WIDTH = 10;
	gint HEIGHT = 5;
	gint middle;
	
	/*
	 * Collecting all needed information
	 */
	g_return_val_if_fail(bar != NULL, FALSE);
	bg = cui_widget_get_background(CUI_WIDGET(bar));
	g_return_val_if_fail(bg != NULL, FALSE);	
	position = cui_widget_get_area_window_coords(CUI_WIDGET(bar));
	g_return_val_if_fail(position != NULL, FALSE);
	/*
	 * If this widget is too far we don't have to bother.
	 */
	if (position->width <= 0 || position->height <= 0)
		return FALSE;
	
	color = cui_rectangle_get_color(CUI_RECTANGLE(bar));
	hcolor = cui_rectangle_get_high_color(CUI_RECTANGLE(bar));
	lcolor = cui_rectangle_get_low_color(CUI_RECTANGLE(bar));

	/*
	 * Now drawing. FIXME: there is a better method for this, see CuiWindow
	 * for an example.
	 */
	if (color == NULL || hcolor == NULL || lcolor == NULL)
		return FALSE;

	if (!background_get_view_flag(bg, Sizebars))
		goto finalize;
	
	// The rectangle.
	gdk_gc_set_foreground(bg->gc, color);
	gdk_draw_rectangle(bg->window, bg->gc, TRUE, 
			position->x + border, 
			position->y + border, 
			position->width - 2 * border + 1, 
			position->height - 2 * border + 1);

	// The high color lines.
	gdk_gc_set_foreground(bg->gc, hcolor);
	for (n = 0; n < border; ++n) {
		gdk_draw_line(bg->window, bg->gc, 
				position->x, 
				position->y + n, 
				position->x + position->width, 
				position->y + n);
		gdk_draw_line(bg->window, bg->gc, 
				position->x + n, 
				position->y, 
				position->x + n, 
				position->y + position->height);
	}
	
	// The low color lines.
	gdk_gc_set_foreground(bg->gc, lcolor);
	for (n = 0; n < border; ++n) {
		gdk_draw_line(bg->window, bg->gc,
			      position->x + position->width - n,
			      position->y + n, 
			      position->x + position->width - n, 
			      position->y + position->height);
		gdk_draw_line(bg->window, bg->gc,
			      position->x + n,
			      position->y + position->height - n, 
			      position->x + position->width, 
			      position->y + position->height - n);
	}
	
	/*
	 * If the bar is not wide enough we cant draw the decoration.
	 */
	if (position->width < 15) 
		goto finalize;
	
	// The decoration
	if (bar->decor & DecorIncrease) {
		middle = position->y + (position->height / 2);
		
		//gdk_gc_set_line_attributes(bg->gc, 2, GDK_LINE_SOLID, GDK_CAP_PROJECTING, GDK_JOIN_ROUND);
		
		for (dx = position->x + position->width - 2; dx > position->x + 4 + WIDTH; dx -= WIDTH + 6) {
		gdk_gc_set_foreground(bg->gc, &(bar->colors[1]));
		gdk_draw_line(bg->window, bg->gc,
				dx - WIDTH, 
				middle - HEIGHT,
				dx - WIDTH, 
				middle + HEIGHT
				);
		
		gdk_draw_line(bg->window, bg->gc,
				dx - WIDTH,
				middle - HEIGHT,
				dx,
				middle
				);
		
		gdk_gc_set_foreground(bg->gc, &(bar->colors[0]));
		gdk_draw_line(bg->window, bg->gc,
				dx - WIDTH,
				middle + HEIGHT,
				dx,
				middle
				);
		}
	}
	
	if (bar->decor & DecorDecrease) {
		middle = position->y + (position->height / 2);
		
		for (dx = position->x + position->width - 2; dx > position->x + 4 + WIDTH; dx -= WIDTH + 6) {
		gdk_gc_set_foreground(bg->gc, &(bar->colors[1]));
		gdk_draw_line(bg->window, bg->gc,
				dx, 
				middle - HEIGHT,
				dx, 
				middle + HEIGHT
				);
		
		gdk_draw_line(bg->window, bg->gc,
				dx,
				middle - HEIGHT,
				dx - WIDTH,
				middle
				);
		
		gdk_gc_set_foreground(bg->gc, &(bar->colors[0]));
		gdk_draw_line(bg->window, bg->gc,
				dx,
				middle + HEIGHT,
				dx - WIDTH,
				middle
				);
		}
	}
		
finalize:
	g_free(position);
	return TRUE;
}


static void
cui_size_bar_class_init(
		gpointer g_class, 
		gpointer g_class_data)
{
	static gboolean   initialized = FALSE;
	GObjectClass     *gobject_class = G_OBJECT_CLASS (g_class);
	CuiWidgetClass   *klass = CUI_WIDGET_CLASS(g_class);
	GParamSpec       *pspec;
	GClosure         *class_closure;
	GType             params[1];
	guint             id;
#ifdef DEBUG	
	g_message("%s(): start", __func__);
#endif

	if (!initialized){
		/*
		 * The initialization.
		 */
		gobject_class->set_property = cui_size_bar_set_property;
		gobject_class->get_property = cui_size_bar_get_property;
		
		pspec = g_param_spec_int("decoration", "decor",
				"The type of the complementary decoration.",
				G_MININT, G_MAXINT, 0, G_PARAM_READWRITE);
		g_object_class_install_property(gobject_class, PROPERTY_DECOR, pspec);
		
		pspec = g_param_spec_double("percent", "width-percent",
				"The width of the bar as percentage of the width of the window.",
				0.0, 1.0, 0.0, G_PARAM_READWRITE); 
		g_object_class_install_property(gobject_class, PROPERTY_PERCENT, pspec);
		
		pspec = g_param_spec_double("turnpoint", "turn-point",
				"The percent where we turn from green to red or -1.",
				-1.0, 1.0, -1.0, G_PARAM_READWRITE); 
		g_object_class_install_property(gobject_class, PROPERTY_TURNPOINT, pspec);
		/* 
		 * Interface signals creation
		 */
		class_closure = g_cclosure_new(
				G_CALLBACK(cui_size_bar_expose_handler),
				NULL, /* user data */
				NULL  /* destroy data */
				);
		params[0] = G_TYPE_INT;
		g_signal_override_class_closure(
				cui_rectangle_get_expose_signal_id(),
				CUI_SIZE_BAR_TYPE,
                                class_closure);
		initialized = TRUE;
	}
}


static void
cui_size_bar_instance_init(
		GTypeInstance   *instance, 
		gpointer         g_class)
{
	CuiSizeBar *self = (CuiSizeBar *)instance;

	self->decor = DecorNone;
	self->percent = 0.0;
	self->turnpoint = -1.0;
}


GType 
cui_size_bar_get_type(void)
{
	static GType type = 0;
	
	if (type == 0) {
		static const GTypeInfo info = {
			sizeof(CuiSizeBarClass),
			(GBaseInitFunc) NULL, 
			NULL,
			(GClassInitFunc) cui_size_bar_class_init,
			NULL,
			NULL,
			sizeof(CuiSizeBar),
			128,
			/* instance_init */
			(GInstanceInitFunc) cui_size_bar_instance_init 
		};
		
		type = g_type_register_static(CUI_RECTANGLE_TYPE, 
				"CuiSizeBar", 
				&info, 
				0);
	}
	return type;
}

/**
 * cui_size_bar_new:
 * @returns: a new size bar
 *
 */
CuiWidget *
cui_size_bar_new(void)
{
	CuiWidget *rec;
	rec = g_object_new(CUI_SIZE_BAR_TYPE, NULL);
	return rec;
}



/*****************************************************************************
 * Class: CuiFrame                                                           *
 *                                                                           *
 *****************************************************************************/
static void
cui_frame_set_property (
		GObject      *object, 
		guint         property_id, 
		const GValue *value, 
		GParamSpec   *pspec) 
{
}

static void
cui_frame_get_property (
		GObject      *object, 
		guint         property_id, 
		GValue       *value, 
		GParamSpec   *pspec)
{
	//CuiFrame *self = (CuiFrame *) object;
	g_value_set_string (value, "value");
}

static void
cui_frame_class_init(
		gpointer g_class, 
		gpointer g_class_data)
{
	GObjectClass *gobject_class = G_OBJECT_CLASS (g_class);

	gobject_class->set_property = cui_frame_set_property;
  	gobject_class->get_property = cui_frame_get_property;
}

static void
cui_frame_instance_init(
		GTypeInstance   *instance, 
		gpointer         g_class)
{
}

GType 
cui_frame_get_type(void)
{
	static GType type = 0;
	
	if (type == 0) {
		static const GTypeInfo info = {
			sizeof(CuiFrameClass),
			(GBaseInitFunc) NULL,
			NULL,
			(GClassInitFunc) cui_frame_class_init,
			NULL,
			NULL,
			sizeof(CuiFrame),
			128,
			(GInstanceInitFunc) cui_frame_instance_init 
		};
		
		type = g_type_register_static(CUI_RECTANGLE_TYPE, 
				"CuiFrame", 
				&info, 
				0);
	}
	return type;
}

/**
 * cui_frame_new:
 * @returns: a new #CuiFrame
 *
 * Creates a new #CuiFrame widget.
 */
CuiWidget *
cui_frame_new(void)
{
	CuiWidget *frame;
	frame = g_object_new(CUI_FRAME_TYPE, NULL);
	return frame;
}

